/*
** Module   : LocalSyncher.java
** Abstract : Synchronizes a foreign association based on changes made to the
**            referencing DMO
**
** Copyright (c) 2006-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20061113   @31139 Created initial version. Automatically
**                           synchronizes the keys of a foreign
**                           association based on changes made to legacy
**                           key properties in the referencing DMO.
** 002 ECF 20070629   @35299 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 003 ECF 20071219   @36579 Interim fix for retrieveForeignDMO(). This
**                           method previously was acquiring an exclusive
**                           lock on the foreign DMO, but no more than a
**                           share lock is necessary. Further modification
**                           to this method is necessary to prevent the
**                           introduction of unnecessary contention.
** 004 ECF 20071228   @31139 Fixed retrieveForeignDMO() again. Actually,
**                           retrieving the the foreign DMO with no lock
**                           is sufficient. However, the problem of the
**                           foreign record being deleted before the end
**                           of the current session's transaction is a
**                           problem that will have to be solved
**                           elsewhere. ForeignNuller will have to be
**                           aware of the uncommitted change, so that it
**                           can properly null out this orphaned foreign
**                           key.
** 005 GES 20090421   @41825 Matched utility class package change.
** 006 OM  20130829          Updated calls to match the new Persistence.load() signatures.
*/ 

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import com.goldencode.util.*;

/**
 * Automatically synchronizes both ends of a foreign key association, based
 * upon changes made to the values of DMO properties which participate in a
 * legacy foreign key.  This class drives synchronization when it is
 * triggered by changes to the DMO at the referencing, or "local" end of the
 * association.  In P2J, the association is actually bidirectional;  however,
 * in the database schema, the end of the association which stores the
 * foreign key determines which end is the referencing end and which is the
 * referent end.
 *
 * @see  AssociationSyncher
 * @see  ForeignResolver
 * @see  RelationInfo
 * @see  RecordBuffer
 */
final class LocalSyncher
extends AssociationSyncher
{
   /** Object which resolves legacy keys to a foreign DMO instance */
   private final ForeignResolver resolver;
   
   /** Method used to retrieve foreign DMO from a local DMO */
   private final Method foreignGetter;
   
   /**
    * Constructor.
    *
    * @param   buffer
    *          Record buffer containing the DMO whose property change(s)
    *          will trigger synchronization attempts.
    * @param   info
    *          Foreign relation descriptor object.
    */
   LocalSyncher(RecordBuffer buffer, RelationInfo info)
   {
      super(buffer, info, info.getForeignClass());
      
      resolver = new ForeignResolver(info, buffer, false);
      
      String ifaceName = info.getForeignInterfaceName();
      String method = null;
      try
      {
         // Store foreign record setter method.
         Class<?> localIface = info.getLocalInterface();
         StringBuilder buf = new StringBuilder("get");
         buf.append(ifaceName);
         buf.append("Record");
         method = buf.toString();
         foreignGetter = localIface.getDeclaredMethod(method);
      }
      catch (NoSuchMethodException exc)
      {
         throw new IllegalArgumentException(
           "Invalid DMO class:  missing foreign getter [" +
           info.getLocalInterfaceName() +
           "." +
           method +
           "]");
      }
   }
   
   /**
    * Retrieve the set of property names which represent the columns which
    * comprise the legacy key joining two records.  Changes to these
    * properties trigger synchronization of the two ends of a foreign key
    * association.  A set is returned because legacy keys are often
    * composites of multiple columns.
    *
    * @return  Set of property names comprising the legacy key.  Although
    *          they should generally be the same as those of the referent end
    *          of the association, this implementation returns the property
    *          names associated with the referencing end.
    */
   protected Set getLegacyKeys()
   {
      return resolver.getInfo().getLocalProperties();
   }
   
   /**
    * The local end of an association has changed.  Find the old foreign DMO
    * it referenced.  If different than the new foreign DMO provided,
    * dereference the local DMO from the old foreign DMO.
    *
    * @param   newLocalDMOs
    *          Set of zero or more local DMOs which are about to be updated
    *          with the new association.
    * @param   newForeignDMO
    *          Foreign DMO which is about to be updated with the new
    *          association.
    *
    * @throws  IllegalAccessException
    *          if an access error occurs while invoking a method using
    *          reflection.
    * @throws  InvocationTargetException
    *          if a method invoked using reflection throws an exception.
    */
   protected void preSynch(Set newLocalDMOs,
                           Persistable newForeignDMO)
   throws IllegalAccessException,
          InvocationTargetException
   {
      Persistable localDMO = buffer.getCurrentRecord();
      Persistable foreignDMO = (Persistable) foreignGetter.invoke(localDMO);
      if (foreignDMO == null || foreignDMO.equals(newForeignDMO))
      {
         return;
      }
      
      helper.remove(foreignDMO, localDMO);
   }
   
   /**
    * Retrieve the set of DMOs which comprise the local end of the foreign
    * association.  In this implemenation, this will contain only the current
    * record of the buffer whose changes triggered synchronization.  It is
    * assumed that record is already exclusively locked, otherwise there
    * could not have been a property change to trigger synchronization.
    *
    * @return  Set of newly resolved local DMOs.  The set may be empty, but
    *          it will not be <code>null</code>.
    *
    * @throws  PersistenceException
    *          if there are any errors retrieving DMOs from the database.
    */
   protected Set retrieveLocalDMOs()
   throws PersistenceException
   {
      Persistable dmo = buffer.getCurrentRecord();
      
      return (dmo == null ? (new EmptySet()) : Collections.singleton(dmo));
   }
   
   /**
    * Retrieve the DMO, if any, at the foreign end of the association.  The
    * record is resolved using the new property values of the underlying
    * buffer's current record.
    *
    * @return  Newly resolved foreign DMO, or <code>null</code> if none was
    *          available.
    *
    * @throws  PersistenceException
    *          if there are any errors retrieving the DMO from the database.
    */
   protected Persistable retrieveForeignDMO()
   throws PersistenceException
   {
      Serializable id = resolver.resolveForeign(buffer);
      Persistable dmo = null;
      
      if (id != null)
      {
         Class<?> foreignClass = resolver.getInfo().getForeignClass();
         Persistence persistence = buffer.getPersistence();
         dmo = persistence.load(lockTable,
                                foreignClass,
                                id,
                                LockType.NONE,
                                false,
                                false);
      }
      
      return dmo;
   }
}
