/*
** Module   : RecordBuffer.java
** Abstract : Encapsulates a data record and provides methods to operate on it
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20050816   @22479 Created initial version. Abstract base class
**                           of all DMO classes which provides integration
**                           with stateful, navigable queries.
** 002 ECF 20050916   @22934 Complete rewrite to use dynamic proxies. The
**                           buffer is now a container for a DMO reference
**                           which may change during the course of query
**                           execution. The buffer reference is registered
**                           with the transaction manager, possibly more
**                           than once, by client code.
** 003 ECF 20051028   @23244 First cut at full transaction support. This
**                           includes undo processing, transaction
**                           commit/rollback, automatic lock upgrade,
**                           downgrade, release, and record release. Added
**                           logging as well.
** 004 ECF 20051106   @23375 Added real-time update validation. Introduced
**                           use of DMOValidator class. Also made various
**                           improvements to transaction support and
**                           adjustments to accommodate use with compound
**                           and preselect queries.
** 005 ECF 20051129   @23566 Added support for batch assign mode. DMO
**                           updates must be bracketed by static method
**                           calls to startBatch and endBatch.
** 006 ECF 20051205   @23669 Integrated with ConnectionManager. Uses CM
**                           services at construction to determine logical
**                           and physical database names. Notifies CM of
**                           buffer scope open/close events.
** 007 ECF 20051216   @23748 Implemented various Progress language
**                           statement and built-in function replacements:
**                           copy, compare, ambiguous, new, and locked.
** 008 ECF 20060104   @23830 Added support for temporary records. These
**                           are used for client-side (i.e., non-database)
**                           record sorting.
** 009 ECF 20060117   @23951 Honor pending error status when invoking DMO
**                           setter methods. If in silent error mode and
**                           and error is pending, assignment of DMO
**                           properties is not permitted.
** 010 ECF 20060120   @24042 Replaced ThreadLocal with ContextLocal. Minor
**                           changes to logging.
** 011 ECF 20060201   @24350 Added temp table support. This required
**                           removing the final modifier from the class,
**                           making numerous private methods protected,
**                           and abstracting/adding several new virtual
**                           methods to make this class extensible.
** 012 ECF 20060310   @25031 Added support for natural joins. Added new
**                           inner class ForeignSyncher to perform reverse
**                           foreign record resolution as legacy foreign
**                           key properties are changed by application
**                           code.
** 013 ECF 20060503   @25961 Fixed incorrect handling of latestRecord.
**                           This variable should not be updated when a
**                           record is newly created.
** 014 ECF 20060504   @25981 Improved error reporting. Better handling of
**                           DMO getter/setter index out of bounds errors.
** 015 ECF 20060504   @26007 Fixed defects in buffer copy. Indexed fields
**                           were not being managed properly, leading to
**                           exceptions.
** 016 GES 20060517   @26200 Match new Finalizable interface requirement.
** 017 ECF 20060522   @26588 Fixed delete processing. In addition to
**                           deleting the current record, it is removed
**                           from the buffer.
** 018 ECF 20060531   @26852 Replaced EndConditionException with
**                           QueryOffEndException. The latter is a more
**                           specific version of the former, used to
**                           indicate a query has run off the end of its
**                           result set.
** 019 ECF 20060608   @27025 Fixed defect in record create. After creating
**                           a DMO instance, all records joined to the DMO
**                           by legacy keys must be set to their defaults.
** 020 ECF 20060610   @27103 Fixed defect with wasLocked() reporting.
**                           Underlying locked flag was being cleared too
**                           aggressively.
** 021 ECF 20060621   @27507 Flush newly created records to database more
**                           aggressively. These are now flushed upon any
**                           validation of a new record to ensure that a
**                           query made subsequent to a property being set
**                           on a newly created record finds the new
**                           record.
** 022 ECF 20060627   @27592 Refined validation and flushing. We now track
**                           dirty properties and only validate and flush
**                           if changes to the DMO's state actually have
**                           been made. As a result, validation is more
**                           targeted.
** 023 ECF 20060628   @27669 Track nested open scopes. Added method to
**                           report a reference count of open scopes:
**                           getOpenScopeCount().
** 024 ECF 20060630   @27790 More fixes to flushing/validating. Separated
**                           flushing from validation logic. Added a
**                           rollback capability for invalid buffers.
** 025 ECF 20060720   @28119 Major enhancements. Implemented the change
**                           broker architecture and a more aggressive
**                           eviction policy for DMOs. The former enables
**                           numerous performance optimizations. The
**                           latter reduces the working set and improves
**                           the execution time of most session flushes,
**                           though eviction itself is somewhat costly.
** 026 ECF 20060728   @28273 Moved DMO reference counting mechanism to
**                           BufferManager. incrementDMOUseCount() and
**                           decrementDMOUseCount() are now implemented
**                           there.
** 027 ECF 20060831   @29088 Changed implementation of getSchema().
** 028 ECF 20060906   @29302 Fixed validation defect. Validation and
**                           flushing of a newly created record must not
**                           be triggered by a proxy setter method call
**                           until all columns of a unique constraint have
**                           been updated. The first unique constraint to
**                           have all its columns updated triggers the
**                           flush.
** 029 ECF 20060911   @29462 Fixed memory leak. Setting the same current
**                           record more than once was incrementing the
**                           DMO use count with the BufferManager too
**                           often.
** 030 ECF 20060916   @29667 Major changes to undo architecture. We no
**                           longer rely on the TransactionManager undo
**                           processing for record changes. Rather, we
**                           track property changes and record-level
**                           actions as they occur, saving instructions on
**                           how to rollback changes at each block scope.
**                           Also made some minor fixes to lock handling
**                           to prevent locks from being released too
**                           early in some cases.
** 031 ECF 20060918   @29676 Fix to setCurrentRecord(). Don't flush brand
**                           new records; instead, give business logic a
**                           chance to modify them first.
** 032 ECF 20060919   @29698 Introduced in/decrementDMOUseCount() methods
**                           which delegate to worker methods in
**                           BufferManager. Previously, the BufferManager
**                           methods were called directly, but the new
**                           methods allow subclasses to override the
**                           implementations. Also made processChange()
**                           protected, so that subclasses could override
**                           its implementation.
** 033 ECF 20060926   @29959 Back out #031 (@29676). This was an improper
**                           fix, in that we are not flushing the newly
**                           set record, but rather the record it is
**                           replacing. The latter record MUST be flushed,
**                           or it is possible to lose new records.
** 034 ECF 20061002   @30072 Fixed invocation handler. A getter method
**                           call on an empty buffer must return unknown
**                           value when the error manager is in silent
**                           error mode. Previously it returned null.
** 035 ECF 20061006   @30198 Fixed snapshot logic. A snapshot is never
**                           taken for newly created records, only for
**                           records read from the database. Modified
**                           implementation of isNew() to retain the
**                           knowledge of a record's newly created status,
**                           even after it is made persistent.
** 036 ECF 20061010   @30311 Made various fixes and added foundational
**                           support for auto-commit temp-tables. Added
**                           safety code to prevent an error from being
**                           raised if batch assign mode is interrupted
**                           prematurely, then restarted at a later time.
** 037 ECF 20061012   @30379 Fixed iterate() method. Must release current
**                           record in preparation for next pass of loop.
** 038 ECF 20061018   @30485 Moved processing to clean up batch assign
**                           mode to BufferManager. This was interfering
**                           with end-of-scope processing in some cases.
** 039 ECF 20061018   @30531 Integrated SessionListener architecture.
**                           The invocation handler now receives
**                           sessionClosing notifications. This allows the
**                           handler to force the Persistence object to
**                           bind to a Hibernate session before attempting
**                           a method call on a DMO. This is required now
**                           because a DMO may be proxied by Hibernate,
**                           and we no longer pin sessions open
**                           artificially. Thus, it is necessary to force
**                           a session bind in some cases, to avoid a
**                           LazyInitializationException.
** 040 ECF 20061023   @30596 Fixed defect in invocation handler. Newly
**                           created DMOs were being validated too early,
**                           before all properties in a unique constraint
**                           had been updated.
** 041 ECF 20061027   @30756 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 042 ECF 20061103   @30947 Fixed reportChange(). Bind the persistence
**                           service to a backing session if necessary,
**                           before invoking DMO's deepCopy() method. This
**                           allows the DMO's internal associations to be
**                           lazily populated by Hibernate if this was not
**                           previously done.
** 043 ECF 20061105   @31095 Fixed delete(). This method was not properly
**                           handling inverse foreign key relations to the
**                           deleted record. These are now nulled out when
**                           a record is deleted. The ForeignNuller class
**                           is used to track down dependent records and
**                           break their foreign key references to the
**                           deleted record. Also fixed a defect whereby a
**                           deleted record was not being cleared properly
**                           from other buffers which held the record.
** 044 ECF 20061114   @31141 Removed ForeignSyncher inner class. This code
**                           was refactored and expanded to create the
**                           external AssociationSyncher class and its
**                           subclasses.
** 045 ECF 20061114   @31144 Disabled prohibitively expensive foreign
**                           association synchronization operations. This
**                           effectively removes all such synchronization
**                           except setting the foreign DMO in a local DMO
**                           when a legacy key property in that local DMO
**                           is changed.
** 046 ECF 20061114   @31160 Moved batch assign mode state variables to
**                           BufferManager. This change enables us to
**                           employ a more discriminating batch mode
**                           cleanup mechanism because the batch mode
**                           can be tied to a specific block scope.
** 047 ECF 20061115   @31188 Fixed off-end behavior in setCurrentRecord().
**                           Clear the snapshot record only if setting a
**                           null current record AND this would be
**                           considered an error condition by the calling
**                           query.
** 048 ECF 20061115   @31211 Fixed regression caused by #047 (@31188).
**                           Added parameter to setCurrentRecord() method
**                           to allow caller to determine under what
**                           circumstance to reset snapshot.
** 049 ECF 20061116   @31237 Fixed nested buffer scope handling. Upon the
**                           closure of a nested buffer scope, cleanup
**                           processing was releasing the current record
**                           unconditionally.  This must happen only upon
**                           closure of the outermost buffer scope.
** 050 ECF 20061116   @31242 Fixed error saving new records. Needed to
**                           register the need for a pending session flush
**                           with the ChangeBroker for the saved DMO type.
**                           Otherwise, Hibernate complains about possible
**                           nonthreadsafe access to the session.
** 051 ECF 20061120   @31327 Optimization. Do not create reversibles for
**                           create and delete events if buffer is not
**                           undoable.
** 052 ECF 20061207   @31620 Report record insert as a state change to the
**                           ChangeBroker. Previously, an insert caused a
**                           pending session flush but did not trigger a
**                           state change event to be sent to ChangeBroker
**                           listeners.
** 053 ECF 20061212   @31668 Fixed undoable processing for newly created
**                           records. In cases where the record's creation
**                           and its persistence to the database occurred
**                           in different scopes, undo was not working
**                           properly.
** 054 ECF 20070130   @32040 Fixed handling of no-undo temp-tables. Change
**                           undone with a database-level transaction
**                           rollback is re-rolled forward and committed,
**                           as if the rollback had not occurred. Moved
**                           Reversible inner class into a top level
**                           class. Also fixed copy implementation to
**                           flush newly created destination record at the
**                           appropriate time.
** 055 ECF 20070202   @32071 Changed signature for copy() worker method.
**                           It is necessary for this method to accept
**                           either a proxy or backing DMO object as the
**                           destination record. The former supports the
**                           buffer-copy implementation; the latter is
**                           used internally by TemporaryBuffer.
** 056 ECF 20070212   @32160 Rollback and flush fixes. Force session flush
**                           when DMOs are rolled back in memory. Fixed
**                           NPE when flushing a newly created record to
**                           database in some circumstances. Fixed
**                           database error when a newly created record
**                           was improperly flushed upon a buffer scope
**                           ending after a rollback.
** 057 ECF 20070221   @32229 Track reference count for DMOs managed by
**                           no-undo temp-tables. These had been excluded
**                           from the aggressive eviction policy.
** 058 ECF 20070306   @32288 Fixed change notification. Changes to
**                           multiple indexes in an extent field made in
**                           batch mode were not being processed and
**                           reported properly. Only the last change was
**                           being remembered.
** 059 ECF 20070307   @32336 No longer throw 'error not on file' when a
**                           null current record is set. This is now the
**                           responsibility of the caller, since different
**                           errors can be raised for the same condition,
**                           depending upon the calling context.
** 060 ECF 20070309   @32367 Fixed ClassCastException. Regression
**                           introduced with #058 (@32288).
** 061 ECF 20070318   @32444 Fixed defect persisting unmodified, new DMO
**                           instances. An unmodified, newly created DMO
**                           must not be flushed to the database.
** 062 ECF 20070322   @32546 Reworked flushing and validation. A record is
**                           flushed at the earlier of:  all columns in
**                           any index are updated, or another record (or
**                           null) is stored as the current buffer record,
**                           or a query is about to be executed, or the
**                           outermost open buffer scope ends.
** 063 ECF 20070323   @32563 More flushing cleanup. This version of the
**                           class implements an interim solution for
**                           dirty checks before flushing a transient
**                           record. When checking whether all columns of
**                           any index are dirty, we use all indexes for
**                           temp tables, but only unique indexes for
**                           permanent tables. This is because index info
**                           for temp tables is available via DMOIndex,
**                           which is not the case for permanent tables.
**                           Ultimately, we will want to use JDBC database
**                           metadata to query this index information for
**                           both types.
** 064 ECF 20070412   @32986 Rearchitected record locking implementation.
**                           The same lock manager is used, but record
**                           locks within a session are now coordinated by
**                           one RecordLockContext instance per database,
**                           per session. Removed the LockState inner
**                           class, as this logic is now managed by the
**                           RecordLockContext class for all buffers in a
**                           session (by database), rather than by each
**                           buffer independently.
** 065 ECF 20070415   @32994 Fixed record delete notification regression.
**                           Notification of a record delete was not
**                           being broadcast via the ChangeBroker. Refined
**                           rollback-based deletes as well. Added static
**                           helper method to compare DMOs by primary key.
** 066 ECF 20070416   @33029 Integrated user interrupt handling.
** 067 ECF 20070503   @33391 Added registration of new instance with the
**                           BufferManager.
** 068 ECF 20070509   @33461 Fixed validation processing during bulk copy.
**                           Fixed isIndexDirty() implementation to be
**                           more complete in its index checks.
** 069 ECF 20070515   @33672 Fixed DMO proxy invocation handler. When a
**                           setter method is invoked, the underlying
**                           property must be marked dirty regardless of
**                           whether the value of the property actually
**                           changed. This is necessary for the proper
**                           function of the isIndexDirty() method, so
**                           that newly created records are flushed to the
**                           database at the appropriate time.
** 070 ECF 20070517   @33676 More changes to fine tune flush timing for a
**                           newly created record. Records in tables with
**                           no indices are flushed immediately upon
**                           creation (effectively just temp tables, since
**                           permanent tables always have indices). When
**                           in batch assign mode, any setter invocation
**                           causes the buffer to be marked dirty, even if
**                           the invocation did not change the value.
** 071 ECF 20070523   @33728 Fixed defect related to deletions and saves.
**                           Deletion of a record was not properly
**                           cleaning up the validation context and the
**                           dirty read manager. Likewise, saves related
**                           to undo processing were not updating these
**                           mechanisms.
** 072 ECF 20070525   @33769 Fixed LightweightUndoable processing. Need to
**                           increment/decrement DMO use counts properly
**                           when TransactionManager creates Undoable
**                           backup sets and when scopes end. Also moved
**                           registration of LightweightUndoable from
**                           constructor to when first buffer scope is
**                           opened. These changes ensure that undo
**                           processing does not store a detached DMO into
**                           the buffer.
** 073 ECF 20070531   @33909 Corrected handling of bulk setter methods.
**                           DMOs with indexed properties have bulk setter
**                           methods which accept a scalar value or an
**                           array of values to set multiple indices of
**                           property in one call. The invocation handler
**                           had to be reworked to properly detect, report
**                           and process change in these cases.
** 074 ECF 20070601   @33917 Moved registration with ChangeBroker. Instead
**                           of registering upon construction, we now wait
**                           until the outermost scope is opened. This
**                           prevents a memory leak from registering at
**                           the wrong scope.
** 075 ECF 20070403   @34361 Retrofit for ForeignNuller change. Changed
**                           invocation of ForeignNuller.breakLinkages()
**                           to use simpler signature.
** 076 ECF 20070706   @34403 Added deregisteredSessionListener() to
**                           invocation handler. Required by change to
**                           SessionListener interface.
** 077 ECF 20070706   @34415 Minor optimizations. Cache ForeignNuller
**                           objects lazily instead of letting the static
**                           method ForeignNuller.breakLinkages() gather
**                           these instances repeatedly for each deleted
**                           record. Replaced StringBuffer with
**                           StringBuilder. Cache ChangeBroker instance.
** 078 ECF 20070709   @34423 Minor cleanup. Optimized commit() and
**                           rollback() slightly.
** 079 ECF 20070726   @34719 Refined release on iterate behavior. Allow
**                           release on iterate to be overridden. This is
**                           necessary for compound queries, so that the
**                           outer buffers in a client-side join do not
**                           lose their state while records are found for
**                           inner buffers.
** 080 ECF 20070815   @34853 Improved temporary record handling for nested
**                           client-side where expressions. Necessary to
**                           support client-side, converted CAN-FIND
**                           statements nested within a where clause.
** 081 ECF 20070821   @34902 Added prepare() and disableSilentError()
**                           methods. These additions enable errors during
**                           query substitution parameter processing to be
**                           suppressed using silent error mode, just like
**                           in Progress during where clause processing.
** 082 ECF 20070827   @34992 Added support for restricted bulk deletes.
**                           A new static delete() method variant accepts
**                           a where clause and query substitution parms
**                           and delegates the bulk delete service to the
**                           appropriate RecordBuffer instance. Currently,
**                           this service is not supported for permanent
**                           tables (future optimization), but the
**                           addition of this infrastructure allows
**                           subclasses to implement support for this
**                           service.
** 083 ECF 20070926   @35248 Fixed undo processing defect. Snapshot was
**                           incorrectly being nulled out in cases where
**                           a TM undo cycle tried to reset the current
**                           record to itself. We now backup the latest
**                           snapshot and restore it upon undo.
** 084 ECF 20070907   @35293 Fixed ConnectionManager notifications. Calls
**                           to open and close buffer scopes notified the
**                           connection manager unconditionally, but this
**                           is only appropriate for permanent tables.
**                           This functionality was abstracted into
**                           protected methods which can be overridden by
**                           subclasses. Replaced database name string
**                           database information object.
** 085 ECF 20071018   @35591 Integrated PersistenceFactory.
** 086 ECF 20071029   @35633 Changed in[de]crementDMOUseCount() signature.
**                           Matches change in BufferManager.
** 087 ECF 20071130   @36124 Fixed setRecord(). Removed resetSnapshot
**                           parameter. This method no longer clears the
**                           snapshot when a null current record is set.
**                           In this circumstance, the placeholder
**                           snapshot must not be lost. Integrated
**                           generics.
** 088 ECF 20071205   @36265 Track off-end status and sort index of last
**                           query updating the buffer. Added sortIndex
**                           and offEnd parameters to setRecord(); added
**                           getOffEnd() method. This information is used
**                           by a find query to ensure it starts its
**                           search at the proper placeholder record.
** 089 ECF 20071221   @36546 Refined rollback of auto-commit temp records.
**                           Do not try to re-delete a record already not
**                           present in the database.
** 090 SVL 20071231   @36627 Next primary key generation is carried out
**                           to the separate function.
** 091 ECF 20080129   @36935 Added debug/trace level logging. Log all
**                           create and flush events, as well as the
**                           details of any rollbacks.
** 092 ECF 20080310   @37483 Added getDialect() convenience method. Also
**                           integrated generics.
** 093 SVL 20080312   @37428 The inverse synchronization is enabled and
**                           performed for newly created DMOs.
** 094 SVL 20080418   @38038 Added support for the inverse sorting.
** 095 ECF 20080424   @38150 Integrated generics.
** 096 SVL 20080421   @38098 Association Synchers are disabled if foreign
**                           keys are disabled.
** 097 SVL 20080430   @38175 Reversible actions and updates are removed
**                           from the current scope into the commit()
**                           if we are at the master transaction level.
** 098 ECF 20080506   @38220 Fixed automatic lock upgrade logic. No
**                           upgrade is necessary if the datum being set
**                           in the DMO is the same as the existing value.
** 099 SVL 20080512   @38260 Added pinnedLockTypes map which maps primary
**                           keys of past and present currentRecords to
**                           their pinned LockTypes.
** 100 SVL 20080516   @38291 Memory consumption optimization.
** 101 SVL 20080526   @38397 pinnedLockTypes are cleared into finished(). 
** 102 CA  20080604   @38539 Fixed the rollback of nested CREATE/DELETE 
**                           actions, when they are applied for the same
**                           DMO. Fixed the rollback of nested UPDATE
**                           actions for the same DMO - the changes must
**                           be reported to the ChangeBroker. Ref: #38537
** 103 ECF 20080506   @38612 Integrated DirtyShareContext. Used to share
**                           uncommitted index changes across user
**                           sessions.
** 104 ECF 20080606   @38643 Removed ValidationContext references.
**                           Obsoleted by DirtyShareContext.
** 105 ECF 20080609   @38653 Fixed several regressions. LightweightUnoable
**                           was improperly flushing a transient record
**                           during undo processing. Refactored dirty
**                           index update sharing to permit temp tables to
**                           bypass this processing.
** 106 ECF 20080610   @38691 Fixed rollback of DMO property updates. Dirty
**                           share manager was not being notified of
**                           changes being rolled back.
** 107 CA  20080612   @38713 Fixed FIND statement record retrieval and a
**                           dirty share issue:the snapshot must be set to
**                           null when current block is finished and also
**                           the snapshot must save the new record on 
**                           create.
** 108 ECF 20080612   @38729 Fixed rollback() in ReversibleCreate and
**                           ReversibleDelete. Removed optimization to let
**                           database handle rollback on full transaction
**                           boundaries. Our internal state (and that of
**                           Hibernate) was not correct in this case.
** 109 CA  20080616   @38787 Fixed rollback() in ReversibleCreate: 
**                           reportChange(...) must be called on a record 
**                           insert rollback, so all the registered
**                           listeners for this buffer will be notified.
** 110 ECF 20080626   @38961 Fixed ReversibleCreate.rollback(). Current
**                           record must be nulled out unconditionally,
**                           whether or not the record was persisted. Also
**                           removed immediate rollback of DMO updates
**                           during failed validation. Instead, we rely on
**                           the normal, TM-driven rollback mechanism.
** 111 CA  20080605   @38568 Implemented rollback of bulk delete actions.
**                           Fixed rollback of reversible update actions.
** 112 ECF 20080628   @39070 More undo/rollback fixes. We no longer pin
**                           DMOs to the current Hibernate session when
**                           they are stored in LightweightUndoable.
**                           Added special handling for sharing dirty
**                           changes when newly created records have not
**                           yet been stored in the primary database.
** 113 CA  20080702   @39018 Fixed ReversibleUpdate.rollback() - the map
**                           unreportedChanges was not populated correct
**                           when there are changes for extent fields.
** 114 CA  20080716   @39124 Modifications due to DirtyShareContext API 
**                           changes. Fixed rollback - if there are dirty
**                           CREATE or DELETE actions, all the indexes for
**                           the entity will be locked before the chain of
**                           reversibles is processed.
** 115 CA  20080718   @39170 Fix ValidationException in not-transacted
**                           blocks.
** 116 ECF 20080723   @39171 Fixed flushing behavior for temp-tables with
**                           no indexes. Updates to such tables must be
**                           flushed to the database immediately.
** 117 CA  20080725   @39211 Fix rollback for when the reversibles 
**                           reference "stale" records.
** 118 ECF 20080806   @39328 Added rowid support. Reimplemented recordID()
**                           and added rowID().
** 119 CA  20080808   @39358 Implemented no-undo temp-table support: these
**                           tables require no rollback; on each CREATE,
**                           DELETE or UPDATE event, the BufferManager is
**                           added a Reversible containing the current
**                           record state. Fixed subtle flow in
**                           detectChange() related to diff processing.
** 120 ECF 20080815   @39557 Support global DMO change events.
** 121 CA  20080826   @39567 Added rollbackPending which is called by TM
**                           when UNDO targets a parent block. 
** 122 CA  20080815   @39466 Support API change in DBUtils. Register with
**                           the BufferManager at the end of the c'tor.
** 123 SIY 20080904   @39719 Inferred generics in buffer creation methods.
** 124 SVL 20080901   @39653 Key reclamation is performed into commit()
**                           and rollback().
** 125 ECF 20080925   @39948 Refactored Reversible concrete classes.
** 126 GES 20080926   @39974 Minor signature change.
** 127 ECF 20081001   @39998 Changed openScope() to use varargs. This
**                           allows the conversion to compact multiple
**                           openScope() calls in business logic to a
**                           single call. Also fixed NPE in
**                           ReversibleDelete.getId(). Added versions of
**                           certain static helpers which return primitive
**                           boolean instead of logical.
** 128 ECF 20081007   @40049 Removed deprecated available() method.
** 129 ECF 20081020   @40209 Fixed rollback at full transaction. Cannot
**                           create/update/delete DMO at full transaction
**                           boundary just before database rollback.
** 130 ECF 20081031   @40286 Fixed delete of DMOs with extent fields. In
**                           some cases, DMOs with uninitialized, lazy,
**                           persistent collections were being stored in
**                           ReversibleDelete objects. Upon rollback, the
**                           missing collection data was unavailable.
** 131 ECF 20081106   @40326 Fixed errorNotOnFile(). Use DMO alias instead
**                           of DMO interface name.
** 132 ECF 20081114   @40453 Fixed record lock leak. Record locks must be
**                           relinquished in iterate() and retry() if we
**                           are at the outermost open scope.
** 133 ECF 20081118   @40550 Fixed ReversibleDelete. RecordBuffer's
**                           current record DMO must be used for rollback
**                           if it represents the same record as the
**                           ReversibleDelete's internal DMO. Removed
**                           processing in RecordBuffer.rollback() which
**                           was intended to handle this case, but was
**                           incorrect.
** 134 ECF 20081201   @40733 Added support for key reclamation for temp
**                           tables. Uses the same hooks as we had for key
**                           reclamation in permanent tables (commit() and
**                           rollback() methods). A new reclaimKeys(List)
**                           method allows subclasses to override the
**                           original implementation.
** 135 SVL 20081204   @40794 Added undoables map cleanup into rollback().
** 136 ECF 20081211   @40870 Better debug information in toString().
** 137 ECF 20081215   @40918 Modified transient record flushing. Removed
**                           flush(Method, Object[]) variant. Simplified
**                           flush() variant.
** 138 ECF 20090107   @41015 Minor update to stateChanged().
** 139 ECF 20090112   @41163 Fixed undo-related locking behavior. Locks
**                           were being downgraded rather than released in
**                           certain transaction-boundary cases.
** 140 ECF 20090128   @41240 Added stale flag and isStale(), setStale().
**                           A buffer is stale when the contained DMO's
**                           state is known to be out of sync with the
**                           corresponding record in the database. This
**                           can occur after a database-level rollback.
** 141 ECF 20090202   @41249 Fixed setCurrentRecord(). Release/downgrade
**                           of lock on outgoing record was not properly
**                           handled in transaction cases.
** 142 ECF 20090206   @41264 Major changes to invocation handler and auto-
**                           commit processing. Batch mode was modified in
**                           order to enable an auto-commit transaction to
**                           begin at the first update and end in the
**                           endBatch() method. The copy() method was
**                           modified to not begin a new transaction if
**                           already in an auto-commit batch. The proxy
**                           invocation handler now attempts to associate
**                           a detached record with the current Hibernate
**                           session before invoking an indexed setter or
**                           getter method, to avoid a lazy initialization
**                           exception.
** 143 ECF 20090218   @41332 Memory improvements. Avoid creating
**                           unnecessary Reversible maps. Set reduced
**                           default sizes on certain collections.
**                           Leverage cached collections of getter methods
**                           in PropertyHelper.
** 144 ECF 20090306   @41437 Fixed load(). Added code to prevent releasing
**                           a newly created (still transient) record when
**                           trying to re-load the same record.
** 145 ECF 20090312   @41585 Fixed regression caused by #142 (@41264).
**                           Detached record association logic inside
**                           invocation handler could try to reassociate a
**                           null record.
** 146 ECF 20090315   @41607 Fixed ReversibleDelete.getDMO(). In the event
**                           target DMO is currentRecord, ensure lazily
**                           initialized extent field collections have
**                           been realized before returning DMO. Minor
**                           optimization to load() to avoid reloading the
**                           same temp table record unnecessarily during
**                           undo processing.
** 147 GES 20090424   @41942 Import change.
** 148 ECF 20090508   @42127 Rewrote validation processing. Validation is now
**                           much more incremental: only those indexes and
**                           properties which have been modified are validated
**                           immediately for new records. Remaining validation
**                           occurs when the buffer is flushed.
** 149 ECF 20090512   @42155 Fixed regressions caused by #148 (@42127).
**                           Integrated BufferFlushQueue.
** 150 ECF 20090513   @42167 Added isPropertyIndexed(). Needed for validation.
** 151 SVL 20090225   @41364 API change. recordID() returns recid instead of
**                           integer.
** 152 ECF 20090610   @42633 Added evictDMOIfUnused(). Evicts DMO from
**                           Hibernate session if no resource references it.
** 153 ECF 20090611   @42660 Optimized load() method and ValidationHelper.
**                           We now avoid expensive database operations and
**                           validation, respectively, when possible.
** 154 ECF 20090612   @42666 Fixed regressions in #153 (@42660). Partially
**                           rolled back changes to load() method.
** 155 ECF 20090623   @42963 Made changes to use updated BufferManager API.
**                           Some methods to retrieve undoable/reversible
**                           actions were changed.
** 156 SVL 20090706   @43054 Into copy() catch and rethrow all
**                           ConditionExceptions without rolling back
**                           transaction.
** 157 ECF 20090707   @43104 Fixed validation corner case. A newly created,
**                           permanent record with dirty properties but no
**                           dirty indexes was being validated and flushed,
**                           but not shared with other sessions. Later, an
**                           attempt to rollback dirty changes failed.
** 158 SVL 20090706   @43062 Into endBatch() rollback transaction if
**                           validation has failed.
** 159 CA  20090709   @43126 When batch mode ends, if validation fails, the
**                           reversible info must be collected (to allow
**                           rollback for the changed values). On rollback,
**                           a ReversibleUpdate must actually rollback the
**                           changed property, regardless of the mode (re-roll
**                           forward or not full transaction block).
** 160 ECF 20090716   @43222 Refactored initialization logic. Since it is
**                           possible to define a buffer and even open a scope
**                           on it without the backing database being
**                           connected, initialization of many critical
**                           instance variables cannot be set up during
**                           construction. This logic was moved to the new
**                           initialize() method, which is invoked lazily, as
**                           late as possible, within and external to this
**                           class.
** 161 ECF 20090717   @43229 Fixed regressions in #159 and #160. Backed out
**                           changes to ReversibleUpdate.rollbackWorker, which
**                           can cause unique constraint violations. Buffer
**                           must be initialized when invocation handler is
**                           called for a DMO method.
** 162 ECF 20090720   @43306 Improved ValidationHelper. Changed prevalidate
**                           method to prevent unnecessary, future validation
**                           in some cases.
** 163 CA  20090724   @43382 Rolledback H140 (#41240), to fix the rollback 
**                           mechanism. Also, on rollback, the DMO's which
**                           have reversible actions will be processed in 
**                           reverse order instead the order they were 
**                           registered.
** 164 CA  20090724   @43395 On query off end, the undoables registered with
**                           this record buffer are not undone - the buffer 
**                           must not reference any records after a query
**                           which uses the buffer goes off-end.
** 165 ECF 20090725   @43422 Fixed validation and flushing of temp tables in
**                           invocation handler. Added getMasterBuffer()
**                           method to support proper rollback handling of
**                           shared temp tables. Prevent prevalidation in some
**                           cases.
** 166 ECF 20090729   @43449 Implemented Commitable.validate() and fixed
**                           flush(). The former is needed to validate and
**                           flush transient records when field assignments do
**                           not trigger flushing. The persistence and change
**                           reporting within flush() is now bracketed with an
**                           explicit transaction to prevent auto-commit temp
**                           table records from leaving behind a pending flush
**                           entry in the ChangeBroker. This was causing flush
**                           problems and downstream data loss.
** 167 SVL 20090803   @43477 Raise stop condition into initialize() if the
**                           database isn't connnected.
** 168 CA  20090804   @43508 Added implementation for reset() method in
**                           QueryOffEndListener - this will set the buffer to
**                           reference no record.
** 169 CA  20090807   @43557 Reversibles must be kept by database and table
**                           name, not by RecordBuffer instance - to ensure
**                           proper reversible order when rolling back records
**                           in a table with multiple buffers defined for it.
**                           This fixes @43542.
** 170 SVL 20090810   @43578 Into Handler.invoke() log
**                           IndexOutOfBoundsException stacktrace only as 
**                           debug output.
** 171 ECF 20090810   @43581 Fixed getDatabaseAndTable(). This method must not
**                           throw IllegalStateException if the buffer is not
**                           yet initialized. Instead, it must return null.
** 172 ECF 20090810   @43590 Changed to accommodate SessionListener API
**                           change. sessionClosing() method was replaced with
**                           sessionEvent().
** 173 ECF 20090812   @43603 Replaced getDatabaseAndTable() with
**                           getBufferType(). More information was required by
**                           callers than was provided by the database and
**                           table string, so the BufferType class is used
**                           instead.
** 174 ECF 20090813   @43615 Fixed validation problem which caused a flushing
**                           error downstream. When validating dirty props, we
**                           now undo them immediately upon failure, so they
**                           are never flushed to the database nor recorded
**                           in Reversibles. Failure to do this was causing
**                           unique constraint failures in some cases.
** 175 ECF 20090815   @43634 Fixed invocation handler range checking for
**                           extent field access. Element index is now checked
**                           for unknown value and out of bounds conditions
**                           before backing method is invoked.
** 176 ECF 20090819   @43691 Fixed ReversibleUpdate.rollbackWorker(). This
**                           method was making an assumption that the buffer
**                           still contained a snapshot of the record being
**                           rolled back. It now uses a different API into the
**                           dirty share context.
** 177 CA  20091001   @44059 Fixed leak caused by bulk delete duplication: do 
**                           not duplicate it if the record was deleted 
**                           explicitly (via a DELETE operation) or implicitly
**                           (via a bulk delete).
** 178 CA  20091001   @44081 Avoid NPE in rollback(). In certain bulk delete
**                           cases, the Reversible will not return a DMO, so
**                           we cannot make a deep copy for the snapshot.
** 179 SVL 20091019   @44144 Added support of assign and delete database
**                           triggers.
** 180 SVL 20091019   @44159 Pass a duplicate of the old value into assign
**                           triggers in order to ensure that changes into
**                           this value will not affect undo processing.
** 181 ECF 20091026   @44192 Fixed compare() method. Need to use getter method
**                           specific to each DMO's buffer, in case the types
**                           are different. We were using the getter method
**                           for only the first DMO, which made an implicit
**                           (and incorrect) assumption that both DMOs were of
**                           the same type.
** 182 CA  20091102   @44300 Fixed retry processing - on retry, the openScope 
**                           call must not be executed.
** 183 SVL 20091117   @44402 Notify the BufferManager when a new record was 
**                           loaded (or the old record was reloaded) into the
**                           buffer.
** 184 CA  20091202   @44467 Reworked the rollback mechanism, to maintain 
**                           proper reversible order and thus to solve some
**                           unique-constraint violations during rollback of
**                           some bulk delete cases and mixed record operations.
**                           The ReversibleChain is no longer used, instead a 
**                           linked set with all operations which affect a 
**                           certain table is used.
** 185 ECF 20100128   @44553 Fixed NPE in rollback(). Added safety code to
**                           handle certain bulk delete case.
** 186 SIY 20100429   @44837 Fixed double rollback in some conditions.
** 187 CA  20100527   @44880 Added support for CREATE and "no record is 
**                           available triggers". The CREATE, DELETE and 
**                           ASSIGN triggers will execute too, but the passed
**                           DMO will not be the buffer proxy; instead, it
**                           will be the actual record instance.
** 188 CS  20121031          Removed client specific references
** 189 CA  20130117          Added posibility to retreive the DMO interface
**                           from the proxy's invocation handler.
** 190 ECF 20130119          Changed the public API to this class; a number of
**                           static methods have been removed and replaced
**                           by the Buffer interface API. This enables more
**                           natural, direct method calls on DMO proxy objects
**                           for standard services. This entailed a change to
**                           the DMO proxy implementation from the dynamic
**                           proxy support in java.lang.reflect to that in
**                           com.goldencode.proxy (we needed to inherit our
**                           proxies from a functional base class).
** 191 CA  20130126          Renamed getDMOInterfaceForProxy to 
**                           getDMOInterfaceForObject and fixed possible 
**                           ClassCastException and made it to receive any 
**                           kind of object. 
** 192 OM  20121106          Added CURRENT-CHANGED built-in function implementation.
** 193 OM  20130212          Added CREATE-BUFFER conversion support.
** 194 ECF 20130224          Added limited support for embedded SQL.
** 195 OM  20130304          Added missing create() (CREATE-BUFFER) signatures and ordered the
**                           existing ones.
** 196 VMN 20130428          Implemented runtime support for handle-based methods BUFFER-COMPARE
**                           and BUFFER-COPY, reusing code with corresponding statements. Stub
**                           added for compare() method containing FieldReference as last
**                           parameter. Added CASE-SENSITIVE option conversion and runtime
**                           support for the statement version of BUFFER-COMPARE
** 197 SVL 20130331          Upgraded to Hibernate 4.
** 198 OM  20130416          Added reminder for integer (maybe other) validation fails.
**                           Scanned InvocationTargetException in invoke(Object, Method, Object[])
**                           for ErrorConditionException and act as P4GL.
**                           Added support for int64 expression indexes.
** 199 VMN 20130711          Fixed properties comparing with extent: isAssignableTypes().
** 200 OM  20130702          Fixed 'unmappable character for encoding ASCII' error
** 201 VMN 20130722          Added aggregated functions support for embedded SQL.
** 202 OM  20130829          Finished/fixed Current-changed implementation.
*/

package com.goldencode.p2j.persist;

import java.beans.Introspector;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.HibernateException;
import org.hibernate.LockMode;
import org.hibernate.Session;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.id.IdentityManager;
import com.goldencode.p2j.persist.trigger.*;
import com.goldencode.p2j.util.*;
import com.goldencode.proxy.*;
import com.goldencode.util.*;

/**
 * A container for a Data Model Object (DMO) instance which represents a
 * single database record (or the lack thereof). A record buffer provides an
 * interface for the persistence framework to deal with many different DMO
 * types in a consistent manner. It exposes several static methods which
 * permit client code to define a new buffer and to manipulate the data
 * record which backs that buffer.
 * <p>
 * Client code does not manipulate DMO implementation class instances
 * directly, nor does it manage instances of <code>RecordBuffer</code>
 * directly. Rather, DMOs are exposed to client code by proxy. Each DMO
 * defines a business-domain interface, which is the means by which client
 * code manages data records. A new buffer is defined using the {@link
 * #define} factory method. This method returns a dynamic proxy to a
 * particular DMO interface; however, the proxy initially is not backed by
 * any DMO implementation instance. A backing instance is only created by
 * either invoking the {@link #create} method, or by passing the proxy to
 * a query and successfully retrieving a data record. Prior to the creation
 * of a backing DMO implementation instance in a buffer by one of these
 * means, any invocation of the DMO interface methods against the proxy will
 * result in an exception. Likewise, if any attributes of the buffer which
 * require a backing record are manipulated before such a record is available,
 * an exception is thrown.
 * <p>
 * The dynamic proxy created at buffer definition actually implements more
 * than just the DMO-specific interface. It also implements the {@link
 * BufferReference} interface. This package private interface allows other
 * persistence framework classes to access the record buffer associated with
 * a DMO proxy, when the proxy is passed in as a parameter from client code.
 * The proxy {@link RecordBuffer.Handler invocation handler} will detect
 * method calls destined for this interface and will retrieve the record
 * buffer object reference.
 * <p>
 * The primary responsibility of the invocation handler, though, is to
 * dispatch DMO interface getter and setter method calls invoked against the
 * proxy, to the correct backing DMO. This decoupled architecture is
 * leveraged to provide support for the Progress early constraint checking
 * semantic, whereby not-null and uniqueness constraints are checked at
 * data value assignment time, rather than at record commit time.  We also
 * take advantage of this hook to perform reverse foreign record resolution.
 * <p>
 * Reverse foreign record resolution is the process whereby a foreign record
 * associated with the current record DMO must be updated at runtime, as
 * application code modifies the properties which represent the database
 * fields upon which Progress relied as a foreign key.  During conversion,
 * natural joins between tables are detected and any legacy foreign keys are
 * replaced with a foreign key which refers to the foreign table's surrogate
 * primary key.  Since this explicit foreign key reference did not exist in
 * the pre-conversion Progress code, we must maintain relational integrity by
 * manually updating this foreign key as the legacy key values change.  This
 * is the job of {@link AssociationSyncher} and its subclasses.  The DMO
 * proxy invocation handler is used to detect when the value of such a legacy
 * key is modified, and the foreign key is reset immediately, unless the
 * current context is running within an assignment batch (see below where
 * batch processing is described with respect to validation).  In the case of
 * an active batch, the foreign record resolution is deferred until the batch
 * ends.
 * <p>
 * In some cases, client code must make multiple updates to a DMO instance,
 * and validation must be deferred until the last update is made, where it
 * would not be meaningful to validate while the DMO is in an intermediate
 * state.  Batch assign mode is designed to handle this requirement.  Once
 * batch assign mode is {@link #startBatch started}, validation for all
 * buffers in the current context is deferred until batch mode is {@link
 * #endBatch ended}.  At this point, all buffers which were updated during
 * batch assign mode are validated, in the order in which they were updated.
 * <p>
 * Each new instance registers a lightweight implementation of the
 * <code>Undoable</code> interface to enable the {@link
 * com.goldencode.p2j.util.TransactionManager TransactionManager} to preserve
 * the state of the buffer, such that it can be restored in an undo situation.
 * <p>
 * This class implements the <code>Finalizable</code> interface, so that it
 * can release locks and records when a buffer scope closes.  The actual time
 * of release also depends upon the current state of the transaction manager.
 * <p>
 * This class implements the <code>Commitable</code> interface to enable the
 * {@link com.goldencode.p2j.util.TransactionManager TransactionManager} to
 * commit database transactions and subtransactions.
 * <p>
 * Upon instantiation, a buffer is associated with a logical database. From
 * that point on, it can only be used with that database instance.
 * <p>
 * In certain circumstances (for outer join queries, particularly), a record
 * buffer may not have a database record loaded, but must still be able to
 * service proxied, DMO method calls without failing.  <i>Unknown mode</i> is
 * intended to support this need.  A record buffer can temporarily be placed
 * into this mode, such that getter methods invoked on the associated proxy
 * return unknown value, instead of raising an error condition.  This mode
 * only is used internally by this package.
 * <p>
 * A similar feature is <i>temporary record mode</i>.  Not to be confused
 * with temp-table support (see {@link TemporaryBuffer}), this mode is used
 * in certain cases where the buffer must temporarily hold some record data,
 * in order to allow evaluation of an expression involving the associated
 * DMO.  This is done, for instance, when presorting preselected query
 * results, and when filtering query results against a client-side where
 * clause filter.  This mode only is used internally by this package.  The
 * query invokes this mode by invoking the {@link #setTempRecord} method first
 * with a non-<code>null</code> value, then executing the target expression
 * possibly repeating this sequence multiple times, and finally clears the
 * mode by invoking the <code>setTempRecord</code> method again, this time
 * with a <code>null</code> argument.  Setting the buffer's temp record does
 * not trigger all of the locking and other baggage normally associated with
 * setting the buffer's current record.
 * <p>
 * TODO:  combine many boolean flags into a bit field.
 * <p>
 * TODO:  currently, when there are multiple buffers defined for the same 
 *        table, the BufferManager keeps together all reversibles related 
 *        to the same table belonging to the same database. This ensures 
 *        proper reversible processing on rollback.
 *        <p>
 *        But, the rollback related part in {@link Commitable} can be reworked 
 *        to allow multiple buffers for a certain table in a database. This 
 *        will eliminate the small overhead of rolling back a buffer which
 *        has nothing to do (rollback was performed by one of its siblings).
 *        <p>
 *        Note: a case which needs to be taken into consideration and checked 
 *        is when the foreign-keys are enabled. In this case, if the child is 
 *        rolled back first in a Child.Parent relation, there might be some
 *        problems.
 * 
 * @see     CompoundQuery
 * @see     PreselectQuery
 * @see     PresortQuery
 * @see     RandomAccessQuery
 */
public class RecordBuffer
implements BufferReference,
           RecordChangeListener,
           Commitable,
           Finalizable
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(RecordBuffer.class);
   
   /** Array of methods not to be proxied */
   private static final Method[] doNotProxy = Buffer.class.getMethods();


   /** Possible return code from invocation handler index range checks */
   private static enum RangeCode
   {
      /** Index value is normal */
      NORMAL,
      
      /** Index value is out of bounds */
      OUT_OF_BOUNDS,
      
      /** Index value is unknown */
      UNKNOWN_VALUE,
   };

   /** Possible buffer operation types */
   protected static enum OperationType
   {
      /** Compare operation */
      COMPARE,

      /** Copy operation */
      COPY
   }

   /** Possible method types */
   protected static enum MethodType
   {
      /** Statement */
      STATEMENT,

      /** Method */
      METHOD
   }

   /** Context-local ChangeBroker */
   protected final ChangeBroker changeBroker = ChangeBroker.get();
   
   /** Object which manages this and all buffers in the current context */
   protected final BufferManager bufferManager = BufferManager.get();
   
   /** Logical database or database alias associated with this buffer */
   private final String ldbOrAlias;
   
   /** Invocation handler for DMO proxy associated with this buffer */
   private final Handler handler = new Handler();
   
   /** Object which notifies buffers of new record flush events */
   private final BufferFlushQueue flushQueue = BufferFlushQueue.get();
   
   /** Grouping DMO and Buffer interface associated with this buffer */
   private final Class<?> dmoBufInterface;
   
   /** Interface which defines DMO's API */
   private final Class<?> dmoInterface;
   
   /** Transaction manager's lookup key for this buffer */
   private final String variable;
   
   /** Map of setter methods of the current DMO interface to property names */
   private final Map<Method, String> propsBySetter;
   
   /** Map of getter methods of the current DMO interface to property names */
   private final Map<Method, String> propsByGetter;
   
   /** Map of property names to extent values */
   private final Map<String, Integer> extentsByProperty;
   
   /** Set of getter methods of the current DMO interface */
   private final Set<Method> getters;
   
   /** Map of DMO properties to pending update values */
   private final Map<String, Object> dirtyProps =
      new HashMap<String, Object>(8);
   
   /** Property changes pending submission to the ChangeBroker */
   private final Map<String, Object> unreportedChanges =
      new HashMap<String, Object>(8);
   
   /** Record locking context for the current user session */
   private RecordLockContext lockContext = null;
   
   /** Object which manages dirty read sharing across sessions */
   private DirtyShareContext dirtyContext = null;
   
   /** Validation helper object */
   private ValidationHelper validationHelper = null;
   
   /** Name of database table associated with this buffer */
   private String table = null;
   
   /** Buffer type (key consisting of database, table, and undoable flag) */
   private BufferType bufferType = null;
   
   /** Class which defines DMO's implementation */
   private Class<?> dmoClass = null;
   
   /** Persistence service object */
   private Persistence persistence = null;
   
   /** Logical name of database associated with this buffer */
   private String logicalDatabase = null;
   
   /** Foreign association synchronization helper objects */
   private ArrayList<AssociationSyncher> assocSynchers = null;
   
   /** Names of properties which represent legacy foreign keys */
   private Set<String> legacyForeignKeys = null;
   
   /** Legacy key properties whose setters have been invoked */
   private Set<String> dirtyLegacyKeyProps = null;
   
   /** Object which identifies modified indexes */
   private IndexHelper indexHelper = null;
   
   /** Properties which participate in any indexes of the backing table */
   private Set<String> indexedProperties = null;
   
   /** Properties which participate in unique indexes of the backing table */
   private Set<String> uniqueProperties = null;
   
   /**
    * Maps primary keys of past and present currentRecords to their pinned
    * LockTypes
    */
   private ScopedDictionary<Serializable, LockType> pinnedLockTypes = null;
   
   /** Convenience reference to associated DMO proxy */
   private BufferReference dmoProxy = null;
   
   /** Map of property names to setter methods for all DMO properties */
   private Map<String, Method> setterMap = null;
   
   /** Map of property names to getter methods for all DMO properties */
   private Map<String, Method> getterMap = null;
   
   /** Map of property names to size methods for indexed DMO properties */
   private Map<String, Method> sizerMap = null;
   
   /** Getter methods to trigger loading of lazy collections for extents */
   private List<Method> lazyExtentTriggers = null;
   
   /** Is this buffer registered with the BufferFlushQueue? */
   private boolean registeredForFlush = false;
   
   /** Current block scope depth */
   private int blockDepth = 0;
   
   /** Block depth at which outermost scope was opened */
   private int activeScopeDepth = 0;
   
   /** Number of open scopes for this buffer (open scopes can be nested) */
   private int openScopeCount = 0;
   
   /** DMO data record which currently backs this buffer */
   private Persistable currentRecord = null;
   
   /** Latest DMO data record to back this buffer */
   private Persistable snapshot = null;
   
   /** Temporary record stack used for purposes such as sorting */
   private List<Persistable> tempRecords = null;
   
   /** Current <code>ReversibleCreate</code>, if any */
   private ReversibleCreate reversibleCreate = null;
   
   /** Helpers which null out foreign key references to deleted DMOs */
   private List<ForeignNuller> foreignNullers = null;
   
   /** Scope depth at which transient record was created */
   private int createScope = -1;
   
   /** Is current record newly created (not read from the database)? */
   private boolean newlyCreated = false;
   
   /** Safe mode where record is null but DMO getters return unknown value */
   private boolean unknownMode = false;
   
   /** Indicates if last find failed because record could not be locked */
   private boolean locked = false;
   
   /** Indicates if last find failed because criteria was ambiguous */
   private boolean ambiguous = false;
   
   /** Is a transaction commit necessary upon ending a batch assign? */
   private boolean commitPending = false;
   
   /** Does Persistence service need to be bound to a backing session? */
   private boolean bindNeeded = false;
   
   /** Is buffer currently the destination of a bulk copy operation? */
   private boolean bulkCopy = false;
   
   /** Scope at which iterate should not release the current record */
   private int iterateReleaseOverrideScope = -1;
   
   /** Sort index last used by a query updating the contents of this buffer */
   private SortIndex sortIndex = null;
   
   /** Off-end status of query which last updated this buffer */
   private OffEnd offEnd = OffEnd.FRONT;
   
   /** Is current record actually a dirty copy from the dirty database? */
   private boolean dirtyCopy = false;
   
   /** Flag indicating the rollback is caused by a off-end query */
   private boolean offEndRollback = false;
   
   /** 
    * Flag indicating that the block to which the buffer is scoped is 
    * retrying 
    */
   private boolean retrying = false;

   /**
    * Delegate object used for invocation of database triggers for this
    * buffer.
    */
   private DatabaseTriggerDelegate triggerDelegate = null;

   /** The query off-end listener associated with this record buffer. */
   private QueryOffEndListener offEndListener = new QueryOffEndListener() 
   {   
      public void initialize() { offEndRollback  = true; }
      public void finish() { offEndRollback = false; }
      public void reset() { setCurrentRecord(null, false, false, false); }
   };
   
   /** Track whether a rollback action is pending */
   private boolean pendingRollbackProcessed = false;

   /**
    * Flag to mark this buffer as changed by another user on a multi-user 
    * environment. Its value is will be returned by currentChanged() method.
    */
   private boolean recordChanged = false;

   /**
    * Temporary copy of the <code>currentRecord</code> used to detect if its value has changed
    * as a result of refreshing from db (FIND CURRENT / GET CURRENT).
    * This field is normally null except for armCurrentChanged/updateCurrentChanged bracketing.
    */
   private Persistable shadowCopyRecord = null;
   
   /**
    * Constructor. This class is only instantiated via the {@link #define}
    * factory method.
    *
    * @param   dmoBufIface
    *          Interface which defines DMO's public API plus all {@link
    *          Buffer} methods.
    * @param   ldbOrAlias
    *          Logical name of database with which this buffer will be
    *          associated permanently, or an alias previously created for the
    *          target database.
    * @param   variable
    *          Name of the variable with which the record buffer is associated
    *          for the purposes of differentiating multiple buffers backed by
    *          the same DMO type.
    */
   protected RecordBuffer(Class<?> dmoBufIface,
                          String ldbOrAlias,
                          String variable)
   {
      this.dmoBufInterface = dmoBufIface;
      
      // extract only the DMO schema-specific interface (i.e., exclude the
      // Buffer interface)
      Class<?> dmoIface = null;
      Class<?>[] superIfaces = dmoBufIface.getInterfaces();
      int len = superIfaces.length;
      for (int i = 0; i < len; i++)
      {
         Class<?> next = superIfaces[i];
         if (DataModelObject.class.isAssignableFrom(next))
         {
            dmoIface = next;
            break;
         }
      }
      
      if (dmoIface == null)
      {
         throw new IllegalArgumentException(
            "Invalid DMO interface used to create RecordBuffer");
      }
      
      this.dmoInterface = dmoIface;
      this.ldbOrAlias = ldbOrAlias;
      this.variable = variable;
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() + ":  constructing record buffer");
      }
      
      // Store maps of getter/setter methods to their property names.
      propsBySetter = PropertyHelper.propertiesBySetter(dmoInterface);
      propsByGetter = PropertyHelper.propertiesByGetter(dmoInterface);
      
      // Store map of property names to extent values (may not have any or all
      // properties immediately available -- see PropertyHelper method).
      extentsByProperty = PropertyHelper.extentsByProperty(dmoInterface);
      
      // Store the set of getter methods.
      getters = PropertyHelper.gettersForInterface(dmoInterface);
      
      initPinnedLockTypes();
      
      // the c'tor will be invoked ALWAYS from the program's default c'tor. 
      // so, there will be no opened scope in P2J - and the registration with  
      // the BufferManager will be leaked to the block from the parent program
      // which executed this program.
      // so, the registration is delayed until the next scope is entered.
      // if any exceptions are thrown during initialization, the registration 
      // must not happen - so this was moved at the end of the c'tor.
      bufferManager.registerPending(this);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the specified database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Record type.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API plus all {@link
    *          Buffer} methods.
    * @param   database
    *          Name of the database which will be associated with the new
    *          record buffer.
    * @param   variable
    *          Name of the variable with which the record buffer is associated
    *          for the purposes of differentiating multiple buffers backed by
    *          the same DMO type.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    */
   @SuppressWarnings("unchecked")
   public static <T> T define(Class<T> dmoBufIface,
                              String database,
                              String variable)
   {
      RecordBuffer buffer = new RecordBuffer(dmoBufIface, database, variable);
      
      return createProxy(dmoBufIface, buffer);
   }
   
   /**
    * Create a proxy object which client code will use to interact with the
    * backing DMO.  Create an invocation handler for the proxy which
    * intercepts method requests made on the proxy and redirects them as
    * necessary.
    *
    * @param   <T> 
    *          Record type.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API plus all {@link
    *          Buffer} methods.
    * @param   buffer
    *          Record buffer instance associated with the created proxy.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    */
   @SuppressWarnings("unchecked")
   protected static <T> T createProxy(Class<T> dmoBufIface,
                                      RecordBuffer buffer)
   {
      Class<?>[] interfaces =
      {
         dmoBufIface,
         BufferReference.class
      };
      
      T dmoProxy = (T) ProxyFactory.getProxy(BufferImpl.class,
                                             interfaces,
                                             false,
                                             doNotProxy,
                                             buffer.handler);
      
      buffer.setDMOProxy((BufferReference) dmoProxy);
      
      return dmoProxy;
   }
   
   /**
    * BUFFER-COPY statement runtime.
    *
    * Perform a bulk copy of data from the current source record (if any) to
    * the current destination record.  Data is copied between properties of
    * like name.  Other properties of the source and destination DMOs are
    * ignored.  If the destination buffer does not contain a current record,
    * one is automatically created.
    * <p>
    * The destination buffer is optionally validated after the copy.
    *
    * @param   srcDMO
    *          Source DMO instance returned by a previous call to {@link
    *          #define}.
    * @param   dstDMO
    *          Source DMO instance returned by a previous call to {@link
    *          #define}.
    * @param   validate
    *          <code>true</code> to validate the destination buffer after the
    *          copy.
    *
    * @throws  ErrorConditionException
    *          if destination buffer is validated and fails;  if there is no
    *          record in the source buffer.
    */
   public static void copy(DataModelObject srcDMO,
                           DataModelObject dstDMO,
                           boolean validate)
   {
      copy(srcDMO, null, false, dstDMO, validate);
   }
   
   /**
    * BUFFER-COPY statement runtime.
    *
    * Perform a bulk copy of data from the current source record (if any) to
    * the current destination record.  Data is copied between properties of
    * like name.  Other properties of the source and destination DMOs are
    * ignored.  If the destination buffer does not contain a current record,
    * one is automatically created.
    * <p>
    * If a list of source properties is provided, the copy behaves
    * differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       copy;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties copied.
    * </ul>.
    * <p>
    * The destination buffer is optionally validated after the copy.
    *
    * @param   srcDMO
    *          Source DMO instance returned by a previous call to {@link
    *          #define}.
    * @param   srcProps
    *          Names of those source properties to be included or excluded
    *          from the copy, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>srcProps</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dstDMO
    *          Destination DMO instance returned by a previous call to {@link
    *          #define}.
    * @param   validate
    *          <code>true</code> to validate the destination buffer after the
    *          copy.
    *
    * @throws  ErrorConditionException
    *          if destination buffer is validated and fails;  if there is no
    *          record in the source buffer.
    */
   public static void copy(DataModelObject srcDMO,
                           String[] srcProps,
                           boolean exclusive,
                           DataModelObject dstDMO,
                           boolean validate)
   {
      BufferReference dstProxy = (BufferReference) dstDMO;
      RecordBuffer srcBuf = ((BufferReference) srcDMO).buffer();
      RecordBuffer dstBuf = dstProxy.buffer();
      
      srcBuf.initialize();
      dstBuf.initialize();
      
      // Get source record.  It is an error if there is none.
      Persistable srcRec = srcBuf.getCurrentRecord();
      if (srcRec == null)
      {
         String msg = "Source element of a BUFFER-COPY statement has no record";
         ErrorManager.recordOrThrowError(5365, msg, false);
         return;
      }
      
      // Get destination record.  Create one if necessary.
      Persistable dstRec = dstBuf.getCurrentRecord();
      if (dstRec == null)
      {
         dstBuf.create();
         dstRec = dstBuf.getCurrentRecord();
      }
      
      // Get maps of source getter methods and destination setter methods,
      // each keyed by property name.
      Map<String, Method> getters = srcBuf.getGetterMap();
      Map<String, Method> setters = dstBuf.getSetterMap();

      // Establish the set of properties being copied.
      Set<String> props = getCommonProperties(getters.keySet(),
                                              setters.keySet(),
                                              srcProps,
                                              exclusive);
      copy(srcBuf,
           dstBuf,
           createSimpleFieldsMap(props, null, null, OperationType.COPY),
           validate);

   }
   
   /**
    * Case insensitive compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data type.  Other
    * properties are ignored for comparison purposes.  The result of the comparison is stored in
    * a logical variable.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to <code>true</code> if all common properties compared as
    *          equal;  else set to <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              DataModelObject dmo2,
                              logical result)
   {
      compare(dmo1, null, false, dmo2, false, result);
   }

   /**
    * Compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data
    * type.  Other properties are ignored for comparison purposes.  The result
    * of the comparison is stored in a logical variable.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   caseSensitive
    *          if <code>true</code>, directs to perform a case-sensitive comparison.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to <code>true</code> if all common properties compared as
    *          equal;  else set to <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              DataModelObject dmo2,
                              boolean caseSensitive,
                              logical result)
   {
      compare(dmo1, null, false, dmo2, caseSensitive, result);
   }

   /**
    * Compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data
    * type.  Other properties are ignored for comparison purposes.  The result
    * of the comparison is stored in a character variable, in the form of a
    * list of legacy field names which failed to compare as equal (NOTE:  the
    * current implementation uses DMO property names instead of legacy names;
    * TODO:  this must be corrected).
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   caseSensitive
    *          if <code>true</code>, directs to perform a case-sensitive comparison.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to the comma-separated list of legacy field names which
    *          failed to compare as equal.  If this variable contains the
    *          empty string after invocation of this method, this indicates
    *          that all common properties compared as equal.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              DataModelObject dmo2,
                              boolean caseSensitive,
                              character result)
   {
      compare(dmo1, null, false, dmo2, caseSensitive, result);
   }

   /**
    * Case insensitive compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data type.  Other
    * properties are ignored for comparison purposes.  The result of the comparison is stored in
    * a character variable, in the form of a list of legacy field names which failed to compare
    * as equal (NOTE:  the current implementation uses DMO property names instead of legacy
    * names;
    * TODO:  this must be corrected).
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to the comma-separated list of legacy field names which
    *          failed to compare as equal.  If this variable contains the
    *          empty string after invocation of this method, this indicates
    *          that all common properties compared as equal.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              DataModelObject dmo2,
                              character result)
   {
      compare(dmo1, dmo2, false, result);
   }

   /**
    * BUFFER-COMPARE statement runtime.
    *
    * Case-insensitive compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data type.  Other
    * properties are ignored for comparison purposes.  The result of the comparison is stored in
    * a logical variable.
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to <code>true</code> if all common properties compared as
    *          equal;  else set to <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              String[] propList,
                              boolean exclusive,
                              DataModelObject dmo2,
                              logical result)
   {
      compare(dmo1, propList, exclusive, dmo2, false, result);
   }

   /**
    * BUFFER-COMPARE statement runtime.
    *
    * Compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data
    * type.  Other properties are ignored for comparison purposes.  The result
    * of the comparison is stored in a logical variable.
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   caseSensitive
    *          if <code>true</code>, directs to perform a case-sensitive comparison.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to <code>true</code> if all common properties compared as
    *          equal;  else set to <code>false</code>.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              String[] propList,
                              boolean exclusive,
                              DataModelObject dmo2,
                              boolean caseSensitive,
                              logical result)
   {
      List<String> diffs = compare(dmo1, dmo2, propList, caseSensitive, exclusive, false);
      if (diffs == null)
      {
         result.setUnknown();
      }
      else
      {
         result.setValue(diffs.isEmpty());
      }
   }

   /**
    * BUFFER-COMPARE statement runtime.
    *
    * Compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data
    * type.  Other properties are ignored for comparison purposes.  The result
    * of the comparison is stored in a character variable, in the form of a
    * list of legacy field names which failed to compare as equal (NOTE:  the
    * current implementation uses DMO property names instead of legacy names;
    * TODO:  this must be corrected).
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   caseSensitive
    *          if <code>true</code>, directs to perform a case-sensitive comparison.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to the comma-separated list of legacy field names which
    *          failed to compare as equal.  If this variable contains the
    *          empty string after invocation of this method, this indicates
    *          that all common properties compared as equal.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              String[] propList,
                              boolean exclusive,
                              DataModelObject dmo2,
                              boolean caseSensitive,
                              character result)
   {
      List<String> diffs = compare(dmo1, dmo2, propList, caseSensitive, exclusive, true);

      if (diffs == null)
      {
         result.setUnknown();
      }
      else
      {
         // TODO:  confirm formatting;  use legacy field names.
         result.setValue(diffs.isEmpty() ? "" : diffs.toString());
      }
   }

   /**
    * BUFFER-COMPARE statement runtime.
    *
    * Case-insensitive compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data type.  Other
    * properties are ignored for comparison purposes.  The result of the comparison is stored in
    * a character variable, in the form of a list of legacy field names which failed to compare
    * as equal (NOTE:  the current implementation uses DMO property names instead of legacy
    * names;
    * TODO:  this must be corrected).
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   result
    *          A variable into which the result of the comparison is saved.
    *          Set to the comma-separated list of legacy field names which
    *          failed to compare as equal.  If this variable contains the
    *          empty string after invocation of this method, this indicates
    *          that all common properties compared as equal.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              String[] propList,
                              boolean exclusive,
                              DataModelObject dmo2,
                              character result)
   {
      compare(dmo1, propList, exclusive, dmo2, false, result);
   }

   /**
    * Case-insensitive compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data type.  Other
    * properties are ignored for comparison purposes.  The result of the comparison is stored in
    * a character variable, in the form of a list of legacy field names which failed to compare
    * as equal (NOTE:  the current implementation uses DMO property names instead of legacy
    * names;
    * TODO:  this must be corrected).
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   fieldReference
    *          Reference to a buffer field, used for "SAVE RESULT IN" result-field
   .
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              String[] propList,
                              boolean exclusive,
                              DataModelObject dmo2,
                              FieldReference fieldReference)
   {
      compare(dmo1, propList, exclusive, dmo2, false, fieldReference);
   }

   /**
    * Compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data
    * type.  Other properties are ignored for comparison purposes.  The result
    * of the comparison is stored in a character variable, in the form of a
    * list of legacy field names which failed to compare as equal (NOTE:  the
    * current implementation uses DMO property names instead of legacy names;
    * TODO:  this must be corrected).
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   caseSensitive
    *          if <code>true</code>, directs to perform a case-sensitive comparison.
    * @param   fieldReference
    *          Reference to a buffer field, used for "SAVE RESULT IN" result-field
   .
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   public static void compare(DataModelObject dmo1,
                              String[] propList,
                              boolean exclusive,
                              DataModelObject dmo2,
                              boolean caseSensitive,
                              FieldReference fieldReference)
   {
      // TODO: this method needs to be implemented (see #2080, notes 111-112) for support
      //       SAVE RESULT IN in case where result-field is not CHARACTER or LOGICAL variable,
      //       but field

      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by name, data type, and
    * extent-matching, between the source buffer and the target buffer. The resulting logical
    * value is either TRUE or FALSE as a whole. A single field that does not compare causes the
    * entire buffer to return FALSE. If there are fields in one buffer that do not exist in the
    * other, they are ignored.
    *
    * @param   srcBuf
    *          Source record buffer.
    * @param   dstBuf
    *          Destination record buffer.
    * @param   mode
    *          If mode-exp is given, it must evaluate to either "binary" or "case-sensitive"
    *          to provide that type of comparison. BUFFER-COMPARE( ) method supports binary
    *          and case-sensitive comparisons between CLOB as well as CHARACTER fields.
    * @param   except
    *          A character expression that evaluates to a comma-separated list of field names
    *          to be excluded from the compare.
    * @param   pairs
    *          A character expression that evaluates to a comma-delimited list of field-name
    *          pairs to be compared.
    *
    * @return  <code>true</code> if all fields values are equal, otherwise <code>false</code>.
    */
   public static logical compare(RecordBuffer srcBuf,
                                 RecordBuffer dstBuf,
                                 character mode,
                                 character except,
                                 character pairs)
   {
      if (!areBuffersExist(srcBuf, dstBuf, MethodType.METHOD))
      {
         return new logical(false);
      }

      // Get maps of source getter methods and destination setter methods,
      // each keyed by property name.
      Map<String, Method> srcGetters = srcBuf.getGetterMap();
      Map<String, Method> dstGetters = dstBuf.getGetterMap();
      Map<String, String> fieldsMap = getFieldsMap(srcGetters,
                                                   dstGetters,
                                                   except,
                                                   pairs,
                                                   OperationType.COMPARE);
      if (fieldsMap == null)
      {
         return new logical(false);
      }
      else
      {
         List<String> diffs = compare(srcBuf, dstBuf, fieldsMap, mode, false);
         return new logical(diffs != null && diffs.size() == 0);
      }
   }

   /**
    * Open a new buffer scope for the record buffer which backs the specified
    * DMO.  This ensures that the {@link BufferManager} is tracking this
    * buffer for block scope entries and exits within transactions.
    *
    * @param   dmos
    *          Variable length list of DMO instances returned by previous
    *          calls to {@link #define}.
    */
   public static void openScope(DataModelObject... dmos)
   {
      for (DataModelObject dmo : dmos)
      {
         get(dmo).openScope();
      }
   }
   
   /**
    * Enter batch assign mode in the current context.  Batch assign mode
    * allows multiple DMO property assignments without incurring validation.
    * Validation for all modified buffers in the current context is performed
    * when batch mode is ended for the current context.
    *
    * @see     #endBatch
    *
    * @throws  IllegalStateException
    *          if the current context is already in batch mode when this
    *          method is invoked.
    */
   public static void startBatch()
   {
      if (LOG.isDebugEnabled())
      {
         LOG.debug(Utils.describeContext() + ":  entering batch assign mode");
      }
      
      BufferManager bufMan = BufferManager.get();
      
      if (bufMan.isBatchAssignMode())
      {
         throw new IllegalStateException("Cannot nest batch assign modes");
      }
      
      bufMan.startBatchAssignMode();
   }
   
   /**
    * Exit batch assign mode in the current context, triggering pending
    * validation on all buffers which were modified while in this mode.
    * <p>
    * Implementation Note:  as a simplification, dirty buffers are validated
    * in their entirety, not only the fields that were modified during the
    * most recent batch.  However, this should be safe (though possibly less
    * efficient), since the buffers must already have been valid upon
    * entering batch mode.
    *
    * @see     #startBatch
    *
    * @throws  IllegalStateException
    *          if the current context is not in batch mode when this method
    *          is invoked.
    * @throws  ErrorConditionException
    *          if validation fails on any buffer which was modified in batch
    *          mode;  if there is any error accessing the database.
    */
   public static void endBatch()
   {
      BufferManager bufMan = BufferManager.get();
      
      Set<RecordBuffer> bufs = bufMan.endBatchAssignMode();
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(Utils.describeContext() +
                   ":  exiting batch assign mode;  dirty buffers:  " + bufs);
      }
      
      if (bufs == null)
      {
         throw new IllegalStateException(
            "Cannot end batch assign mode;  not currently in batch mode");
      }
      
      try
      {
         Iterator<RecordBuffer> iter = bufs.iterator();
         while (iter.hasNext())
         {
            RecordBuffer buffer = iter.next();
            ValidationHelper vHelper = buffer.validationHelper;
            boolean commitPending = buffer.isCommitPending();
            Persistence persistence = buffer.getPersistence();
            boolean exception = false;
            
            try
            {
               vHelper.validateDirty();

               // invoke database triggers
               if (buffer.triggerDelegate != null)
               {
                  for (Map.Entry<String, Object> entry :
                                          buffer.unreportedChanges.entrySet())
                  {
                     String property = entry.getKey();
                     Object value    = entry.getValue();

                     // triggers are executed only for non-extent fields
                     if (value instanceof Object[])
                     {
                        BaseDataType oldValue =
                                         (BaseDataType) ((Object[]) value)[0];
                        buffer.triggerDelegate.
                              executeAssignTriggers(buffer.getDMOProxy(),
                                                    property,
                                                    oldValue.duplicate(),
                                                    false);
                     }
                  }
               }
               
               buffer.synchForeignRelations(false);
               
               buffer.processChange();

               // this flush is needed as the processChange() call may leave
               // the ChangeBroker knowing a flush is pending, which is false,
               // as the validateDirty() call already flushed the changes.
               persistence.flush();
               
               if (buffer.dirtyLegacyKeyProps != null)
               {
                  buffer.dirtyLegacyKeyProps.clear();
               }
            }
            catch (ValidationException exc)
            {
               buffer.unreportedChanges.clear();
               exception = true;
               reportValidationException(exc);
            }
            catch (PersistenceException exc)
            {
               exception = true;
               throw new RuntimeException(exc);
            }
            catch (ConditionException exc)
            {
               // ConditionException may occur into database triggers
               buffer.unreportedChanges.clear();
               exception = true;
               throw exc;
            }
            finally
            {
               if (commitPending)
               {
                  buffer.setCommitPending(false);
                  
                  try
                  {
                     if (!exception)
                     {
                        persistence.commit();
                     }
                     else
                     {
                        persistence.rollback();
                     }
                  }
                  catch (PersistenceException exc)
                  {
                     throw new RuntimeException(exc);
                  }
               }
               
               iter.remove();
            }
         }
      }
      finally
      {
         cleanupBatchMode(bufMan);
      }
   }
   
   /**
    * Prepare for a query to process its substitution parameters.  Silent
    * error mode is temporarily enabled to suppress errors during parameter
    * processing.  This mimics Progress' lenient behavior with errors during
    * where clause processing.
    * <p>
    * This method is invoked by business logic just before construction of a
    * query whose substitution parameters could trigger an error condition.
    * It is the responsibility of the query to disable this temporary silent
    * error mode as soon as possible after its substitution parameters have
    * been processed, during query construction if at all possible.
    * 
    * @param   dmo
    *          Data model object.  This object is not used except as a quick
    *          link to the <code>BufferManager</code>, to avoid the overhead
    *          of a context-local lookup.
    *          
    * @see     #disableSilentError()
    */
   public static void prepare(DataModelObject dmo)
   {
      RecordBuffer buffer = get(dmo);
      buffer.bufferManager.setSilentError(true);
   }
   
   /**
    * Get the backing DMO interface for the given object. If the object is not 
    * a {@link BufferReference} or is null, return null.
    * 
    * @param     obj
    *            The assumed DMO for which the interface is needed.
    *            
    * @return    The DMO's interface.
    */
   public static Class<?> getDMOInterfaceForObject(Object obj)
   {
      if (obj == null || !(obj instanceof BufferReference))
      {
         return null;
      }
      
      return ((BufferReference) obj).buffer().getDMOInterface();
   }
   
   /**
    * Given two sets of property names, an additional list of names
    * (optional), and a flag indicating whether the additional list is
    * inclusive or exclusive, find the appropriate intersection of property
    * names.
    * <p>
    * The resulting set will contain at most the intersection of
    * <code>set1</code> and <code>set2</code>.  However, this set may be
    * further restricted if a non-<code>null</code> <code>propList</code> is
    * provided, as follows:
    * <ul>
    *   <li>if <code>exclusive</code> is <code>true</code>, all elements of
    *       <code>propList</code> are excluded from the intersection of
    *       <code>set1</code> and <code>set2</code>.
    *   <li>if <code>exclusive</code> is <code>false</code>, only those
    *       elements present in <code>set1</code> and in <code>set2</code>
    *       and in <code>propList</code> are included in the final result.
    * </ul>
    *
    * @param   set1
    *          A set of DMO property names.
    * @param   set2
    *          A set of DMO property names.
    * @param   propList
    *          An optional array of property names;  may be <code>null</code>.
    * @param   exclusive
    *          if <code>true</code>, <code>propList</code> is an exclusive
    *          list of property names;  if <code>false</code>,
    *          <code>propList</code> is an inclusive list.  Ignored if
    *          <code>propList</code> is <code>null</code>.
    *
    * @return  The set of property names resulting from following the rules
    *          described above.
    */
   protected static Set<String> getCommonProperties(Set<String> set1,
                                                    Set<String> set2,
                                                    String[] propList,
                                                    boolean exclusive)
   {
      Set<String> props = null;
      
      // Do we have an explicit property list?
      if (propList != null)
      {
         Set<String> explicit = new HashSet<String>(Arrays.asList(propList));
         if (exclusive)
         {
            // Find intersection of property sets.
            props = new HashSet<String>(set1);
            props.retainAll(set2);
            
            // Exclude properties explicitly listed.
            props.removeAll(explicit);
         }
         else
         {
            // Include only properties explicitly listed.
            props = explicit;
            
            // Find intersection with property sets.
            props.retainAll(set1);
            props.retainAll(set2);
         }
      }
      else
      {
         // Find simple intersection of property sets.
         props = new HashSet<String>(set1);
         props.retainAll(set2);
      }
      
      return props;
   }

   /**
    * This method establishes the map of properties being compared or copied.
    *
    * @param   srcMethods
    *          Map of source DMO methods.
    * @param   dstMethods
    *          Map of destination DMO methods.
    * @param   except
    *          A character expression that evaluates to a comma-separated list of field names
    *          to be excluded from the compare.
    * @param   pairs
    *          A character expression that evaluates to a comma-delimited list of field-name
    *          pairs to be compared.
    * @param   operationType
    *          Operation type (COMPARE/COPY).
    *
    * @return  <code>null</code> if error is happened during mapping or map of field name pairs
    *          for actual compare or copy.
    */
   protected static Map<String, String> getFieldsMap(Map<String, Method> srcMethods,
                                                     Map<String, Method> dstMethods,
                                                     character except,
                                                     character pairs,
                                                     OperationType operationType)
   {
      Set<String> srcSet = srcMethods.keySet();
      Set<String> dstSet = dstMethods.keySet();

      // Find intersection of property sets.
      Set<String> props = new HashSet<String>(srcSet);
      props.retainAll(dstSet);
      if (except != null)
      {
         List<String> exceptFields =
            StringHelper.parseTokens(except.getValue(), ",", true, true, false);
         props.removeAll(exceptFields);
      }
      Map<String, String> fieldsMap = createSimpleFieldsMap(props,
                                                            srcMethods,
                                                            dstMethods,
                                                            operationType);
      if (pairs == null)
      {
         return fieldsMap;
      }
      String pairsValue = pairs.getValue();
      if (pairsValue != null && pairsValue.length() > 0)
      {
         List<String> pairsFields =
            StringHelper.parseTokens(pairsValue, ",", true, true, false);
         for (int i = 0; i < pairsFields.size(); i++)
         {
            String dstField = pairsFields.get(i);
            i++;
            if (i < pairsFields.size())
            {
               if (dstSet.contains(dstField))
               {
                  String srcField = pairsFields.get(i);
                  if (srcSet.contains(srcField))
                  {
                     if (isAssignableTypes(srcMethods.get(srcField),
                                           dstMethods.get(dstField),
                                           operationType))
                     {
                        // TODO: we need to map the field names in the except and pairs lists to
                        //       their post-conversion, DMO property name equivalents before we
                        //       process them. These parameters represent lists of legacy field
                        //       names, but we currently are treating them as post-conversion
                        //       DMO property names.

                        fieldsMap.put(srcField, dstField);
                     }
                     else
                     {
                        StringBuilder msg = new StringBuilder("Datatypes in pairlist");
                        msg.append(" argument for ATTACH-DATA-SOURCE, BUFFER-COPY or");
                        msg.append(" BUFFER-COMPARE methods must match for fields ");
                        msg.append(srcField).append(" and ").append(dstField);
                        ErrorManager.recordOrThrowError(12860, msg.toString(), false);
                     }
                  }
                  else
                  {
                     StringBuilder msg = new StringBuilder("Pair with ");
                     msg.append(dstField).append(" failed to have a member in each table");
                     if (operationType == OperationType.COPY)
                     {
                        ErrorManager.recordOrThrowError(9034, msg.toString(), false);
                     }
                     else
                     {
                        ErrorManager.recordOrShowError(9034, msg.toString(), false);
                     }
                     return null;
                  }
               }
               else
               {
                  StringBuilder msg = new StringBuilder("Pair with ");
                  msg.append(dstField);
                  msg.append(" failed to have a member in each table");
                  if (operationType == OperationType.COPY)
                  {
                     ErrorManager.recordOrThrowError(9034, msg.toString(), false);
                  }
                  else
                  {
                     ErrorManager.recordOrShowError(9034, msg.toString(), false);
                  }
                  return null;
               }
            }
            else
            {
               String msg = "Pairs list had too many pairs or odd number of entries";
               ErrorManager.recordOrThrowError(9035, msg, false);
               return null;
            }
         }
      }
      return fieldsMap;
   }

   /**
    * Perform a bulk copy of data from a source record to a destination
    * record.  Data is copied for the properties listed in <code>props</code>.
    * Other properties of the source and destination DMOs are ignored.
    *
    * @param   srcRec
    *          Source DMO record (an actual record, not a proxy)
    * @param   dstDMO
    *          Destination DMO, which may be a dynamic proxy or the backing
    *          data model object.  A proxy should be used in cases where it
    *          is necessary to take advantage of flushing/validating logic
    *          implemented in the proxy invocation handler.  The backing DMO
    *          instance should be used if a raw copy is all that is needed.
    *          In the latter case, calling code is responsible for creating
    *          the record, optionally validating it, and persisting it to the
    *          database at the appropriate time.
    * @param   fieldsMap
    *          Map of properties in source record buffer to properties in destination
    *          record buffer
    * @param   getters
    *          Map of property names to getter methods for the source record.
    * @param   setters
    *          Map of property names to setter methods for the destination
    *          record.
    * @param   sizers
    *          Map of property names to sizer methods for the source record.
    *
    * @return  <code>true</code> if copy is successfull, otherwise <code>false</code>.
    *
    * @throws  IllegalAccessException
    *          if a getter or setter method cannot be invoked because it is not accessible.
    * @throws  InvocationTargetException
    *          if an invoked getter or setter method throws an exception.
    */
   protected static logical copy(Persistable srcRec,
                                 DataModelObject dstDMO,
                                 Map<String, String> fieldsMap,
                                 Map<String, Method> getters,
                                 Map<String, Method> setters,
                                 Map<String, Method> sizers)
   throws IllegalAccessException,
          InvocationTargetException
   {
      for (String field1 : fieldsMap.keySet())
      {
         Method getter = getters.get(field1);
         Method setter = setters.get(field1);
         Method sizer = sizers.get(field1);

         if (sizer == null)
         {
            // Simple getter and setter invocations.
            Object result = getter.invoke(srcRec);
            setter.invoke(dstDMO, result);
         }
         else
         {
            // Indexed getter and setter invocations.
            int size = ((int64) sizer.invoke(srcRec)).intValue();
            for (int i = 0; i < size; i++)
            {
               Object result = getter.invoke(srcRec, i);
               setter.invoke(dstDMO, i, result);
            }
         }
      }
      return new logical(true);
   }

   /**
    * BUFFER-COPY() handle-based method runtime.
    *
    * This method copies any common fields, determined by name, data type, and extent-matching,
    * from the source buffer to the receiving buffer. If there are fields in one buffer that do
    * not exist in the other, they are ignored. This method is used to accommodate temp-tables
    * of joins.
    *
    * @param   dstBuf
    *          Destination record buffer.
    * @param   srcBuf
    *          Source record buffer.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A character expression that evaluates to a comma-separated list of field-name
    *          pairs to be copied.
    *
    * @return  <code>true</code> if copy is successfull, otherwise <code>false</code>.
    */
   static logical copy(RecordBuffer dstBuf,
                       RecordBuffer srcBuf,
                       character except,
                       character pairs)
   {
      Persistable dstRec = dstBuf.currentRecord;
      if (dstRec == null)
      {
         // 4GL returns true independently on srcRec value
         return new logical(true);
      }
      else
      {
         Persistable srcRec = srcBuf.currentRecord;
         if (srcRec == null)
         {
            // 4GL message contains namely "statement" although error is related to handle-based
            // method
            String msg = "Source element of a BUFFER-COPY statement has no record";
            ErrorManager.recordOrThrowError(5365, msg, false);
            return new logical(false);
         }
         Map<String, Method> srcGetters = srcBuf.getGetterMap();
         Map<String, Method> dstSetters = dstBuf.getSetterMap();
         Map<String, String> fieldsMap = getFieldsMap(srcGetters,
                                                      dstSetters,
                                                      except,
                                                      pairs,
                                                      OperationType.COPY);
         if (fieldsMap == null)
         {
            return new logical(false);
         }
         else
         {
            return new logical(copy(srcBuf, dstBuf, fieldsMap, false));
         }
      }
   }

   /**
    * Creates simple properties map with mapping properties with the same names.
    *
    * @param   props
    *          Property names.
    * @param   srcMethods
    *          Map of source DMO methods.
    * @param   dstMethods
    *          Map of destination DMO methods.
    * @param   operationType
    *          Operation type (COMPARE/COPY).
    *
    * @return  Properties map with mapping property to the property with the same name.
    */
   protected static Map<String, String> createSimpleFieldsMap(Set<String> props,
                                                              Map<String, Method> srcMethods,
                                                              Map<String, Method> dstMethods,
                                                              OperationType operationType)
   {
      Map<String, String> fieldsMap = new HashMap<String, String>();
      for (String prop : props)
      {
         if (srcMethods == null ||
             dstMethods == null ||
             isAssignableTypes(srcMethods.get(prop), dstMethods.get(prop), operationType))
         {
            fieldsMap.put(prop, prop);
         }
      }
      return fieldsMap;
   }

   /**
    * Get the buffer instance which corresponds with the given DMO proxy.
    *
    * @param   dmo
    *          DMO instance returned by a previous call to {@link #define}.
    *
    * @return  Buffer associated with <code>dmo</code>.
    */
   static RecordBuffer get(DataModelObject dmo)
   {
      return ((BufferReference) dmo).buffer();
   }
   
   /**
    * Helper method to report and possibly log a validation exception.  Since
    * validation exceptions generally are recoverable errors, they are logged
    * only at debug level and finer.
    * 
    * @param   exc
    *          The validation exception.
    */
   protected static void reportValidationException(ValidationException exc)
   {
      if (LOG.isDebugEnabled())
      {
         LOG.debug("{" +  Utils.describeContext() + "} validation error",
                   exc);
      }
      
      ErrorManager.recordOrThrowError(exc.getNumber(), exc.getMessage());
   }
   
   /**
    * Reset to default values the state variables related to batch assign
    * mode.  If any dirty buffers have commits pending, these are rolled back
    * now.
    * 
    * @param   bufferManager
    *          Context-local <code>BufferManager</code> instance.
    */
   static void cleanupBatchMode(BufferManager bufferManager)
   {
      Set<RecordBuffer> bufs = bufferManager.endBatchAssignMode();
      
      if (bufs == null)
      {
         return;
      }
      
      try
      {
         Iterator<RecordBuffer> iter = bufs.iterator();
         while (iter.hasNext())
         {
            RecordBuffer buffer = iter.next();
            if (buffer.isCommitPending())
            {
               buffer.setCommitPending(false);
               try
               {
                  buffer.getPersistence().rollback();
               }
               catch (PersistenceException exc)
               {
                  ErrorManager.recordOrThrowError(exc);
               }
            }
         }
      }
      finally
      {
         bufs.clear();
      }
   }
   
   /**
    * Disable silent error mode which has been temporarily enabled with a
    * call from business logic to {@link #prepare(DataModelObject)}.  This
    * method is invoked from queries as early as possible, during or just
    * after query construction.
    * <p>
    * Silent error mode will only be disabled if it was previously enabled
    * via the <code>RecordBuffer</code> class.
    * 
    * @see     #prepare(DataModelObject)
    */
   static void disableSilentError()
   {
      BufferManager.get().setSilentError(false);
   }
   
   /**
    * BUFFER-COMPARE statement runtime.
    *
    * Compare the contents of the properties which two DMO instances have in
    * common.  Common properties are defined as having the same name and data
    * type.  Other properties are ignored for comparison purposes.
    * <p>
    * If a list of additional property names is provided, the comparison
    * behaves differently, depending upon the value of <code>exclusive</code>:
    * <ul>
    *   <li>if <code>true</code>, the listed properties are excluded from the
    *       compare;
    *   <li>if <code>false</code>, the listed properties are the only
    *       properties compared.
    * </ul>.
    * <p>
    * <b>Implementation Note:</b>  this is not a full replacement for the
    * Progress BUFFER-COMPARE statement.  In particular, it does not handle
    * the EXPLICIT COMPARES option.  Also, it currently stores the converted,
    * Java property name in the returned list of failed comparisons, whereas
    * dependent code will expect to find legacy field names in this list.
    *
    * @param   dmo1
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   dmo2
    *          DMO instance returned by a previous call to {@link #define}.
    * @param   propList
    *          Names of those properties to be included or excluded from the
    *          comparison, depending upon <code>exclusive</code>.  May be
    *          <code>null</code>.
    * @param   caseSensitive
    *          if <code>true</code>, directs to perform a case-sensitive comparison.
    * @param   exclusive
    *          <code>true</code> if <code>propsList</code> represents an
    *          exclusive list;  <code>false</code> if it represents an
    *          inclusive list.
    * @param   exhaustive
    *          if <code>true</code>, all common properties are compared;  if
    *          <code>false</code>, the comparison ends immediately after the
    *          first time a pair of properties fails to compare as equal.
    *
    * @return  A list of names of those properties which failed to compare as
    *          equal.  If empty, the two DMOs compared as equal overall;  if
    *          not empty, the comparison determined one or more differences
    *          between them.  Will return <code>null</code> only if one or
    *          both buffers are empty, indicating a comparison could not be
    *          performed.
    *
    * @throws  ErrorConditionException
    *          if one or both buffers are empty, such that a comparison could
    *          not be performed;  if there is a problem extracting data from
    *          a buffer.
    */
   private static List<String> compare(DataModelObject dmo1,
                                       DataModelObject dmo2,
                                       String[] propList,
                                       boolean caseSensitive,
                                       boolean exclusive,
                                       boolean exhaustive)
   {
      RecordBuffer buf1 = ((BufferReference) dmo1).buffer();
      RecordBuffer buf2 = ((BufferReference) dmo2).buffer();
      if (!areBuffersExist(buf1, buf2, MethodType.STATEMENT))
      {
         return null;
      }

      // Get maps of source getter methods and destination setter methods,
      // each keyed by property name.
      Map<String, Method> getters1 = buf1.getGetterMap();
      Map<String, Method> getters2 = buf2.getGetterMap();

      // Establish the set of properties being copied.
      Set<String> props = getCommonProperties(getters1.keySet(),
                                              getters2.keySet(),
                                              propList,
                                              exclusive);
      return compare(buf1,
                     buf2,
                     createSimpleFieldsMap(props, null, null, OperationType.COMPARE),
                     caseSensitive ? new character("case-sensitive") : null,
                     exhaustive);
   }

   /**
    * This method checks that two getter or setter methods have compatible corresponding return
    * or parameter types for given operation type.
    *
    * @param   srcMethod
    *          Source method for given operation.
    * @param   dstMethod
    *          Destination method for given operation.
    * @param   operationType
    *          Operation type (COMPARE/COPY).
    *
    * @return  <code>true</code> if methods are compatible, otherwise <code>false</code>.
    */
   private static boolean isAssignableTypes(Method srcMethod,
                                            Method dstMethod,
                                            OperationType operationType)
   {
      Class[] srcParameterTypes = srcMethod.getParameterTypes();
      Class[] dstParameterTypes = dstMethod.getParameterTypes();
      if (operationType == OperationType.COMPARE)
      {
         // values for srcParameterTypes.length and dstParameterTypes.length (both getters)
         // are 0, 1:
         // getF1() and getF2() or
         // getF1(int index) and getF2(int index)
         if (srcParameterTypes.length != dstParameterTypes.length)
         {
            return false;
         }
      }
      else
      {
         // values for srcParameterTypes.length (getter) are 0, 1, and for
         // dstParameterTypes.length (setter) are 1, 2:
         // getF1() and setF2(String f1) or
         // getF1(int index) and setF2(int index, String f2)
         if (srcParameterTypes.length != dstParameterTypes.length - 1)
         {
            return false;
         }
      }
      Class srcMethodType = DBUtils.getGetterSetterType(srcMethod);
      Class dstMethodType = DBUtils.getGetterSetterType(dstMethod);
      if (operationType == OperationType.COMPARE)
      {
         return dstMethodType.equals(srcMethodType);
      }
      else
      {
         return dstMethodType.isAssignableFrom(srcMethodType);
      }
   }

   /**
    * This method checks that two buffer correspond not null records and displayOrRecordError
    * otherwise.
    *
    * @param   srcBuf
    *          Source record buffer.
    * @param   dstBuf
    *          Destination record buffer.
    *
    * @return  <code>true</code> if both records are not nulls, otherwise <code>false</code>.
    */
   private static boolean areBuffersExist(RecordBuffer srcBuf,
                                          RecordBuffer dstBuf,
                                          MethodType methodType)
   {
      srcBuf.initialize();
      dstBuf.initialize();
      if (srcBuf.currentRecord != null && dstBuf.currentRecord != null)
      {
         return true;
      }
      if (methodType == MethodType.METHOD)
      {
         String msg = "BUFFER-COMPARE must have records in both target and source buffers";
         ErrorManager.recordOrShowError(9039, msg, false);
      }
      else
      {
         if (srcBuf.currentRecord == null)
         {
            String msg = "Source element of a BUFFER-COMPARE statement has no record";
            ErrorManager.recordOrShowError(5369, msg, false);
         }
         else
         {
            String msg = "Target element of a BUFFER-COMPARE statement has no record";
            ErrorManager.recordOrShowError(5370, msg, false);
         }
      }
      return false;
   }

   /**
    * Sets compare mode to data type instance for BUFFER-COPY and BUFFER-COMPARE methods.
    *
    * #2080, note 93: ... when you get the value for a DMO property (via getter.invoke), you
    * get a copy, not the actual reference which is saved at the DMO. Thus, is enough to use
    * setCaseSensitive() for that Text instance, without having to worry to restore the
    * previous state (as we work with a copy which will get discarded once its job is done).
    * About the threading issue: Text.caseSens is a field which is set per each
    * character/longchar var instance. Currently, all communication is done sync (with some few
    * CTRL-C exceptions), so no two threads should be able to change the Text.caseSens field
    * for the same instance.
    *
    * @param   baseDataType
    *          Data type instance.
    * @param   mode
    *          "case-sensitive" or "binary". Mode is only acceptible for CLOBs (not supported
    *          so far) and Text (it seems there is no difference between "case-sensitive"
    *          and "binary" for this type).
    */
   private static void setCompareMode(BaseDataType baseDataType, character mode)
   {
      if (mode != null)
      {
         if (mode.compareTo("case-sensitive") == 0 || mode.compareTo("binary") == 0)
         {
            if (baseDataType instanceof Text)
            {
               ((Text)baseDataType).setCaseSensitive(true);
            }
         }
      }
   }

   /**
    * This method compares given fields between the source buffer and the target buffer.
    *
    * @param   srcBuf
    *          Source record buffer.
    * @param   dstBuf
    *          Destination record buffer.
    * @param   fieldsMap
    *          Map of properties in source record buffer to properties in destination
    *          record buffer
    * @param   mode
    *          If mode-exp is given, it must evaluate to either "binary" or "case-sensitive"
    *          to provide that type of comparison.BUFFER-COMPARE( ) method supports binary
    *          and case-sensitive comparisons between CLOB as well as CHARACTER fields.
    * @param   exhaustive
    *          if <code>true</code>, all common properties are compared; if <code>false</code>,
    *          the comparison ends immediately after the first time a pair of properties fails
    *          to compare as equal.
    *
    * @return  A list of fields which failed to compare as equal.  If empty, the two buffers
    *          compared as equal overall; if not empty, the comparison determined one or more
    *          differences between them.  Will return <code>null</code> indicating a comparison
    *          could not be performed.
    */
   private static List<String> compare(RecordBuffer srcBuf,
                                       RecordBuffer dstBuf,
                                       Map<String, String> fieldsMap,
                                       character mode,
                                       boolean exhaustive)
   {
      if (fieldsMap == null)
      {
         return null;
      }
      Map<String, Method> srcGetters = srcBuf.getGetterMap();
      Map<String, Method> dstGetters = dstBuf.getGetterMap();
      Persistable srcRec = srcBuf.currentRecord;
      Persistable dstRec = dstBuf.currentRecord;
      try
      {
         Map<Method, String> propertiesByGetter =
            PropertyHelper.propertiesByGetter(srcBuf.dmoInterface);
         Method[] methodArrays = srcBuf.dmoInterface.getDeclaredMethods();

         // TODO: Temporary solution: order should be corrected. See #2080, note 121 ( "We
         //       can't rely on these always being in the correct order... Once we properly
         //       support metadata, we will get the order from the _Field table
         List<String> diffs = new ArrayList<String>();
         ArrayList<Method> methods = new ArrayList<Method>();
         methods.addAll(Arrays.asList(methodArrays));
         methods.retainAll(srcGetters.values());
         for (Method method : methods)
         {
            String field1 = propertiesByGetter.get(method);
            if (!fieldsMap.containsKey(field1))
            {
               continue;
            }
            String field2 = fieldsMap.get(field1);
            Method srcGetter = srcGetters.get(field1);
            Method dstGetter = dstGetters.get(field2);

            int parametersAmount = srcGetter.getParameterTypes().length;
            if (dstGetter.getParameterTypes().length != parametersAmount)
            {
               continue;
            }
            if (parametersAmount == 0)
            {
               // Simple getter invocations.
               BaseDataType d1 = (BaseDataType) srcGetter.invoke(srcRec);
               BaseDataType d2 = (BaseDataType) dstGetter.invoke(dstRec);

               setCompareMode(d1, mode);
               setCompareMode(d2, mode);
               if (!d1.equals(d2))
               {
                  // TODO: we must populate the list with legacy field names instead of
                  //       post-conversion DMO property names, once metadata support is
                  //       available.
                  diffs.add(field1);
                  if (!exhaustive)
                  {
                     return diffs;
                  }
               }
            }
            else
            {
               // Indexed getter and setter invocations.
               try
               {
                  // We don't know in advance the extent size, so we must
                  // keep going until we go off-end.
                  for (int i = 0; true; i++)
                  {
                     Object[] args = new Object[] {i};
                     BaseDataType d1 = (BaseDataType)srcGetter.invoke(srcRec, args);
                     BaseDataType d2 = (BaseDataType)dstGetter.invoke(dstRec, args);
                     setCompareMode(d1, mode);
                     setCompareMode(d2, mode);
                     if (!d1.equals(d2))
                     {
                        // TODO: we must populate the list with legacy field names instead of
                        //       post-conversion DMO property names, once metadata support is
                        //       available.
                        diffs.add(field1);
                        if (!exhaustive)
                        {
                           return diffs;
                        }
                     }
                  }
               }
               catch (InvocationTargetException exc)
               {
                  if (!(exc.getCause() instanceof IndexOutOfBoundsException))
                  {
                     throw exc;
                  }
               }
               catch (IndexOutOfBoundsException exc)
               {
                  // This is the normal exit route when we are done;  ignore.
               }
            }
         }
         return diffs;
      }
      catch (ConditionException exc)
      {
         DBUtils.handleException(srcBuf.getDatabase(), exc);
         DBUtils.handleException(dstBuf.getDatabase(), exc);
         if (!ErrorManager.isSilentError())
         {
            throw exc;
         }
      }
      catch (Exception exc)
      {
         DBUtils.handleException(srcBuf.getDatabase(), exc);
         DBUtils.handleException(dstBuf.getDatabase(), exc);
         String msg = "Error performing buffer compare";
         LOG.error(msg, exc);
         // TODO:  use proper error.
         ErrorManager.recordOrThrowError(-1, msg);
      }
      return null;
   }

   /**
    * This method copies given fields from the source buffer to the receiving buffer.
    *
    * @param   srcBuf
    *          Source record buffer.
    * @param   dstBuf
    *          Destination record buffer.
    * @param   fieldsMap
    *          Map of properties in source record buffer to properties in destination
    *          record buffer
    * @param   validate
    *          <code>true</code> to validate the destination buffer after the copy.
    *
    * @return  <code>true</code> if copy is successfull, otherwise <code>false</code>.
    */
   private static logical copy(RecordBuffer srcBuf,
                               RecordBuffer dstBuf,
                               Map<String, String> fieldsMap,
                               boolean validate)
   {
      Persistable srcRec = srcBuf.getCurrentRecord();
      BufferReference dstProxy = dstBuf.getDMOProxy();
      Map<String, Method> getters = srcBuf.getGetterMap();
      Map<String, Method> setters = dstBuf.getSetterMap();
      Map<String, Method> sizers = srcBuf.getSizerMap();

      // Perform the copy.  The record, if newly created, will automatically
      // be flushed to the database once all properties in any index have
      // been updated, or once the copy is complete if validation is
      // requested.
      boolean inBatch = dstBuf.bufferManager.isBatchAssignMode();
      boolean autoTx = dstBuf.isCommitPending() && !inBatch;
      try
      {
         Persistence p = dstBuf.persistence;
         if (validate && !inBatch)
         {
            startBatch();
         }

         // Perform the copy.
         dstBuf.bulkCopy = true;
         logical result = copy(srcRec, dstProxy, fieldsMap, getters, setters, sizers);

         if (validate && !inBatch)
         {
            endBatch();
         }
         if (autoTx)
         {
            try
            {
               p.commit();
            }
            finally
            {
               dstBuf.setCommitPending(false);
            }
         }
         return result;
      }
      catch (ConditionException exc)
      {
         throw exc;
      }
      catch (PersistenceException exc)
      {
         String msg = "Error performing buffer copy";
         LOG.error(msg, exc);

         // TODO:  use proper error.
         ErrorManager.recordOrThrowError(-1, msg);
      }
      catch (Exception exc)
      {
         DBUtils.handleException(srcBuf.getDatabase(), exc);
         DBUtils.handleException(dstBuf.getDatabase(), exc);

         if (autoTx)
         {
            try
            {
               dstBuf.persistence.rollback();
            }
            catch (PersistenceException pe)
            {
               LOG.error("Error rolling back during buffer copy", pe);
            }
         }

         String msg = "Error performing buffer copy";
         LOG.error(msg, exc);

         // TODO:  use proper error.
         ErrorManager.recordOrThrowError(-1, msg);
      }
      finally
      {
         dstBuf.bulkCopy = false;
      }
      return new logical(false);
   }

   /**
    * Indicate whether two DMO instances represent the same record, as
    * determined by a comparison of their primary key identifiers.  To be
    * considered the same, neither record may be <code>null</code>, nor have
    * a <code>null</code> primary key identifier.
    *
    * @param   dmo1
    *          First record to compare.
    * @param   dmo2
    *          Second record to compare.
    *
    * @return  <code>true</code> if both records are non-<code>null</code>
    *          and have the same non-<code>null</code> primary keys.
    */
   private static boolean areDMOsSame(Persistable dmo1, Persistable dmo2)
   {
      if (dmo1 == null || dmo2 == null)
      {
         return false;
      }
      
      Serializable id1 = dmo1.getId();
      Serializable id2 = dmo2.getId();
      
      if (id1 == null || id2 == null)
      {
         return false;
      }
      
      return id1.equals(id2);
   }
   
   /**
    * Report the record buffers for whose changes this object is interested
    * in listening.  This listener will be registered with the {@link
    * ChangeBroker} to receive notifications of any change to DMOs whose
    * interface types match those returned by the {@link
    * RecordBuffer#getDMOInterface} methods of the returned buffers.
    *
    * @return  An iterator on the record buffers of interest.  In this
    *          implementation, the iterator will only return this buffer
    *          instance.
    */
   public Iterator<RecordBuffer> recordBuffers()
   {
      return Arrays.asList(new RecordBuffer[] { this }).iterator();
   }
   
   /**
    * Respond to a record change event.  This implementation stores the
    * event's DMO snapshot as its own DMO snapshot if the following
    * conditions are met:
    * <ol>
    *   <li>this buffer's snapshot has not already been stored previously;
    *   <li>the event's <i>post-modification</i> DMO instance is the same
    *       object instance as this buffer's current record;
    *   <li>the current record was not newly created.
    * </ol>
    * <p>
    * If the event indicates a record delete operation, we check if the
    * deleted record is the same as our current record.  If so, we clear it
    * from the buffer, since that record should no longer be available in any
    * buffer, once it has been deleted.
    *
    * @param   event
    *          Event which describes the DMO state change.
    *
    * @throws  PersistenceException
    *          not thrown;  required by the interface.
    */
   public void stateChanged(RecordChangeEvent event)
   throws PersistenceException
   {
      if (!isActive())
      {
         return;
      }
      
      Persistable dmo = event.getDMO();
      
      if (snapshot == null && dmo == currentRecord && !newlyCreated)
      {
         snapshot = event.getSnapshot();
      }
      
      if (event.isDelete() && dmo == currentRecord && currentRecord != null)
      {
         release();
      }
   }
   
   /**
    * Get the entity name associated with this buffer, which is the fully
    * qualified name of the DMO implementation class.
    * 
    * @return  DMO implementation class name.
    */
   public String getEntityName()
   {
      checkActive();
      
      return dmoClass.getName();
   }

   /**
    * Get <code>IndexHelper</code> instance used by this buffer.
    * 
    * @return  <code>IndexHelper</code> instance.
    */
   public IndexHelper getIndexHelper()
   {
      checkActive();
      
      return indexHelper;
   }
   
   /**
    * Get a representation of the latest DMO to be held in the buffer.  This
    * may be the same instance as current record;  it may be a copy.  In
    * cases where the current record has been deleted or released, this
    * represents the most recent record to back the buffer <i>before</i> such
    * deletion or release.
    * <p>
    * This is used to enable a query to keep its place when navigating to a
    * record, based upon the data in the last record held in the buffer,
    * even if that record is no longer available.
    *
    * @return  Snapshot of the latest DMO record, or <code>null</code> if the
    *          buffer never held a current record.
    */
   public Persistable getSnapshot()
   {
      return (snapshot != null ? snapshot : currentRecord);
   }
   
   /**
    * Notifies that a block is about to be normally exited or normally
    * iterated and that any deferred validation processing should be executed.
    * A normal exit or iteration is one that occurs naturally by hitting the
    * end of the block (or loop). A <code>LEAVE</code>, <code>NEXT</code> and 
    * <code>RETURN</code> are also considered normal exits or iterations so
    * long as they were not preceded by an <code>UNDO</code>. Likewise, a
    * <code>QUIT</code> is a normal exit so long as it is not inside a block
    * that has an <code>ON QUIT UNDO, &lt;action&gt;</code>.
    * <p>
    * If the current record is transient, i.e., nothing has triggered a flush
    * by the time the creating block is about to exit/iterate, we validate and
    * flush the record.
    *
    * @param   transaction
    *          <code>true</code> if this is a full transaction and
    *          <code>false</code> if this is only a sub-transaction (a
    *          nested scope with transaction support).
    *
    * @throws  ErrorConditionException
    *          If validation fails.
    */
   public void validate(boolean transaction)
   {
      if (isTransient() && createScope == blockDepth)
      {
         try
         {
            flush();
         }
         catch (ValidationException exc)
         {
            ErrorManager.displayError(exc.getNumber(), exc.getMessage());
            throw new ErrorConditionException(exc);
         }
      }
   }
   
   /**
    * Commit the changes made within a subtransaction.  This entails:
    * <ul>
    *   <li>copying all reversible, record-level actions associated with the
    *       current scope to the next higher scope, so rollback capability
    *       for these actions is not lost when the scope is popped;
    *   <li>copying all undoable, property-level changes associated with the
    *       current scope to the next higher scope, so rollback capability
    *       for these actions is not lost when the scope is popped;
    * </ul>
    * <p>
    * Nothing is done if this is a full transaction commit, because no
    * rollback will be possible at the parent scope once the database-level
    * transaction is committed.
    *
    * @param   transaction
    *          <code>true</code> if this commit represents the master
    *          transaction commit;  <code>false</code> if it represents a
    *          sub-transaction commit.
    */
   public void commit(boolean transaction)
   {
      if (!isActive())
      {
         return;
      }
      
      popPinnedLockScope(transaction, false);
      pushPinnedLockScope();
      
      if (transaction)
      {
         // If we are at the master transaction level, we should remove
         // reversible actions and updates from the current scope.
         Set<Reversible> r0 = bufferManager.getReversibles(this, 0, false);
         
         if (isUndoable() && r0 != null && !r0.isEmpty())
         {
            List<Serializable> freeIds = new ArrayList<Serializable>();
            for (Reversible r : r0)
            {
               // Check whether reversible contains delete action.
               if (r instanceof ReversibleDelete)
               {
                  freeIds.add(r.getId());
               }
            }
   
            if (!freeIds.isEmpty())
            {
               // Reclaim keys from committed deletes.
               reclaimKeys(freeIds);
            }
            
            r0.clear();
         }
         
         Map<Serializable, Reversible> u0 =
            bufferManager.getUndoables(this, 0, false);
         if (u0 != null)
         {
            u0.clear();
         }
         
         return;
      }
      
      mergeReversibles();
   }
   
   /**
    * When the rollback is pending to be processed by a parent block, we must
    * rollup (i.e. call {@link #mergeReversibles()}) the reversibles up the
    * parent stack.
    */
   public void rollbackPending()
   {
      if (isActive() && !TransactionManager.isFullTransaction())
      {
         mergeReversibles();
         pendingRollbackProcessed = true;
      }
   }
   
   /**
    * Restores this buffer's state to that which it had at the beginning of
    * the current block.  In particular, this rolls back all the reversible
    * items for the current scope and removes these entries from the rollback
    * dictionaries managed by the <code>BufferManager</code>.
    * <p>
    * Auto-commit buffers are never rolled back.
    *
    * @param   transaction
    *          <code>true</code> if this commit represents the master
    *          transaction rollback;  <code>false</code> if it represents a
    *          sub-transaction rollback.
    */
   public void rollback(boolean transaction)
   {
      if (!isActive())
      {
         return;
      }
      
      popPinnedLockScope(transaction, true);
      pushPinnedLockScope();
      
      if (!isUndoable())
      {
         // move the reversibles to the parent scope, if sub-transaction
         if (!transaction)
         {
            mergeReversibles();
         }
         else
         {
            // special processing for no-undo tables: if validation fails
            // for the current record, remove all registered reversibles for
            // this record
            if (currentRecord != null)
            {
               boolean invalid = false;
               
               // at this point, before validation is performed, there might 
               // be Update operations which weren't flushed (as there are 
               // other pending operations). flush is needed so validation is 
               // done using the latest versions
               try
               {
                  persistence.maybeFlush();
                  validationHelper.validateFull(currentRecord, false);
               }
               catch (PersistenceException e) 
               {
                  invalid = true;
               }
               catch (ValidationException exc)
               {
                  invalid = true;
               }
               
               if (invalid)
               {
                  Long currentID = currentRecord.getId();
                  Set<Reversible> r0 = bufferManager.getReversibles(this, 
                                                                    0, 
                                                                    false);
                  if (r0 != null)
                  {
                     
                     Iterator<Reversible> iter = r0.iterator();
                     while (iter.hasNext())
                     {
                        Reversible rev = iter.next();
                        
                        try
                        {
                           Serializable id = rev.getId();
                           
                           if (currentID.equals(id))
                           {
                              iter.remove();
                           }
                        }
                        catch (UnsupportedOperationException e) 
                        {
                           // ignore
                        }
                     }
                  }

                  Map<Serializable, List<Reversible>> m0 = 
                     bufferManager.getReversiblesMap(this, 0, false);
                  if (m0 != null)
                  {
                     m0.remove(currentID);
                  }

                  Map<Serializable, Reversible> u0 = 
                     bufferManager.getUndoables(this, 0, false);
                  if (u0 != null)
                  {
                     u0.remove(currentID);
                  }
               }
            }
            
            // ensure the record is set back to the last used record when 
            // rollback completes
            bufferManager.registerBufferUndoable(this, 
                  new LightweightUndoable());
            
            // after transaction rollback, the current referenced record may  
            // not exist or may be in an incorrect state, so we get rid of it 
            // (allow it to be detached from the session)
            setCurrentRecord(null, true, false, false);
         }
         return;
      }
      
      if (blockDepth == createScope)
      {
         markPersisted();
      }
      
      Reversible rev = null;
      List<PersistenceException> errors = null;
      
      String entity = getEntityName();
      boolean forceFlush = false;
      boolean debug = LOG.isDebugEnabled();
      boolean lockAllIndexes = false;
      
      Set<Reversible> r0 = bufferManager.getReversibles(this, 0, false);

      if (debug)
      {
         String s = (r0 != null ? r0.toString() : "(none)");
         LOG.debug(describe() +
                   ":  rolling back reversible actions:  " + s +
                   ";  transaction: " + TransactionManager.isTransaction() +
                   ";  full transaction:  " + transaction);
      }
      
      // Store current snapshot;  we'll need to restore this after
      // Reversibles are rolled back below.
      Persistable oldSnap = snapshot;
      
      try
      {
         if (r0 != null && !r0.isEmpty())
         {
            Iterator<Reversible> iter;
            
            // determine if we lock all the indexes or not
            if (!isTemporary())
            {
               iter = r0.iterator();
               while (iter.hasNext() && !lockAllIndexes)
               {
                  rev = iter.next();
                  if (rev instanceof ReversibleCreate || 
                      rev instanceof ReversibleDelete)
                  {
                     lockAllIndexes = true;
                     break;
                  }
               }
            }
            
            if (lockAllIndexes)
            {
               dirtyContext.lockAllIndexes(entity, LockType.EXCLUSIVE);
            }
            
            List<Serializable> freeIds = new ArrayList<Serializable>();
            for (Reversible r : r0)
            {
               // Check whether reversible contains create action.
               if (r instanceof ReversibleCreate)
               {
                  freeIds.add(r.getId());
               }
            }
            
            if (!freeIds.isEmpty())
            {
               // Reclaim keys from rolled back inserts.
               reclaimKeys(freeIds);
            }
            
            // get the block which is a full transaction and encloses all
            // other transaction blocks
            int txBlock = TransactionManager.getFullTransactionBlock();
            int nestingLevel = TransactionManager.getNestingLevel();
            
            // this set contains all DMO's which are newly created in this
            // block or in a parent block which is full TX too
            Set<Serializable> newDMOs = new HashSet<Serializable>();
            
            // starting from the most outer transaction block, we must collect
            // all records which are newly created
            for (int i = txBlock; i < nestingLevel; i++)
            {
               // compute parent idx relative to current block
               int idx = nestingLevel - i - 1;
               
               // get the reversibles
               Set<Reversible> reversibles = 
                  bufferManager.getReversibles(this, idx, false);
               
               if (reversibles != null)
               {
                  for (Reversible r : reversibles)
                  {
                     if (r instanceof ReversibleCreate)
                     {
                        newDMOs.add(r.getId());
                     }
                  }
               }
            }
            
            ArrayDeque<Reversible> revStack = 
               new ArrayDeque<Reversible>(r0.size());
            
            // Prepare the records for rollback and build a stack of the 
            // reversibles
            Iterator<Reversible> iters = r0.iterator();
            while (iters.hasNext())
            {
               rev = iters.next();
               
               rev.prepareRollback(null);
               revStack.push(rev);
               
               iters.remove();
            }

            // Allow the records to rollback naturally. The rolled back DMOs 
            // must be processed in reverse order, to avoid index collisions
            while (!revStack.isEmpty())
            {
               rev = revStack.pop();
               
               try
               {
                  // the snapshot is used when the dirty share actions are
                  // rolled back
                  Persistable dmo = (Persistable) rev.getDMO();
                  
                  // prevent NPE's - for certain cases when Bulk delete is
                  // involved, there will be no DMO available
                  boolean noDMO = (dmo == null);
                  
                  snapshot = (Persistable) (noDMO ? null : dmo.deepCopy());
                  
                  boolean newDMO =
                     (noDMO ? false : newDMOs.contains(dmo.getId()));
                  
                  rev.setNewlyCreated(newDMO);
               }
               catch (UnsupportedOperationException exc)
               {
                  // ignore, thrown only by bulk delete reversibles,
                  // when getDMO is called
               }
               
               rev.rollback(false, transaction);
               forceFlush = true;
            }
         }
      }
      catch (PersistenceException exc)
      {
         if (LOG.isErrorEnabled())
         {
            LOG.error(describe() +
                      ":  error rolling back reversible action " +
                      rev,
                      exc);
         }
         
         if (errors == null)
         {
            errors = new ArrayList<PersistenceException>();
         }
         errors.add(exc);
      }
      finally
      {
         snapshot = oldSnap;
         
         if (lockAllIndexes)
         {
            try
            {
               dirtyContext.lockAllIndexes(entity, LockType.NONE);
            } 
            catch (PersistenceException exc)
            {
               if (LOG.isErrorEnabled())
               {
                  LOG.error(describe() +
                            ":  error releasing locks while rolling back",
                            exc);
               }
               
               if (errors == null)
               {
                  errors = new ArrayList<PersistenceException>();
               }
               errors.add(exc);
            }
         }
         
         // Clean up undoables.
         Map<Serializable, Reversible> u0 =
            bufferManager.getUndoables(this, 0, false);
         if (u0 != null)
         {
            u0.clear();
         }
      }
      
      // Flush in-memory rollbacks to database.
      if (forceFlush)
      {
         try
         {
            persistence.flush();
         }
         catch (PersistenceException exc)
         {
            if (errors == null)
            {
               errors = new ArrayList<PersistenceException>();
            }
            errors.add(exc);
         }
      }
      
      // Report or eat errors.
      if (errors != null)
      {
         Iterator<PersistenceException> iter = errors.iterator();
         while (iter.hasNext())
         {
            PersistenceException exc = iter.next();
            ErrorManager.recordOrThrowError(exc);
         }
      }
   }
   
   /**
    * Called by the transaction manager just before a buffer scope closes.
    * This is our hook to update the state of locks and record availability.
    * The current record is made unavailable and locks acquired during this
    * scope are either:
    * <ul>
    *   <li>released immediately, if this buffer scope never intersected a
    *       transaction's scope;
    *   <li>downgraded, if a transaction occurred within this buffer scope,
    *       but ended previous to this call;  or
    *   <li>marked for deferred release, if we are still within the scope of
    *       an active transaction;  the locks will be released at the end of
    *       the transaction.
    * </ul>
    */
   public void finished()
   {
      // Decrement reference count of open scopes for this buffer.
      openScopeCount--;
      
      boolean debug = LOG.isDebugEnabled();
      
      if (debug)
      {
         LOG.debug(describe() + ":  finished block scope " + blockDepth);
      }
      
      boolean active = isActive();
      
      if (active)
      {
         // Reset iterate release override scope.
         iterateReleaseOverrideScope = -1;
         
         // Perform cleanup if this is the outermost scope for this buffer
         // (buffers can be shared and open scopes can be nested).
         if (openScopeCount == 0)
         {
            activeScopeDepth = 0;
            
            // Mark current record unavailable.
            setCurrentRecord(null, false, false, false);
            
            // The snapshot is no longer available
            snapshot = null;
            
            // Deregister with the buffer flush queue.
            if (registeredForFlush)
            {
               flushQueue.removeListener(this);
               registeredForFlush = false;
            }
            
            // Attempt to release all locks held by this buffer.
            if (pinnedLockTypes != null)
            {
               pinnedLockTypes.clear();
            }
            lockContext.relinquishBufferLocks(this);
            
            // Notify connection manager that this buffer is done with the
            // connection.
            closeConnectionManagerScope();
         }
      }
         
      if (debug)
      {
         LOG.debug(describe() +
                   ":  closed " +
                   (active ? "" : "in") +
                   "active buffer scope " +
                   bufferManager.getCurrentScope());
      }
   }
   
   /**
    * Called by the transaction manager after each iteration of a repeating
    * block.  Releases the current record.  Relinquishes locks pinned by this
    * buffer if we are at the outermost scope boundary.
    */
   public final void iterate()
   {
      boolean active = isActive();
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() +
                   ":  iterating " +
                   (active ? "" : "in") +
                   "active buffer scope");
      }
      
      if (!active)
      {
         return;
      }
      
      // Release the current record if this behavior has not been overridden
      // in the current scope.
      // JPRM @34715
      if (blockDepth != iterateReleaseOverrideScope)
      {
         // JPRM @30382.
         release();
      }
      
      // We are about to make another pass through an iterating block.
      // Relinquish any record locks this buffer is pinning, if we are at the
      // outermost open scope boundary.
      if (blockDepth == activeScopeDepth)
      {
         lockContext.relinquishBufferLocks(this);
      }
      
      resetPendingRollbackProcessed();
   }
   
   /**
    * Called by the transaction manager before a block is retried after an
    * error.  Relinquishes locks pinned by this buffer if we are at the
    * outermost scope boundary.
    */
   public final void retry()
   {
      // on retry, set a flag so the next openScope call will act accordigly
      retrying = true;

      boolean active = isActive();
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() +
                   ":  retryting " +
                   (active ? "" : "in") +
                   "active buffer scope");
      }
      
      if (!active)
      {
         return;
      }
      
      // We are about to retry a block.  Relinquish any record locks this
      // buffer is pinning, if we are at the outermost open scope boundary.
      if (blockDepth == activeScopeDepth)
      {
         lockContext.relinquishBufferLocks(this);
      }
   }
   
   /**
    * Return a reference to this object instance.  This is implemented to
    * satisfy the <code>BufferReference</code> interface, and provides a
    * means for the persistence layer to retrieve a reference to the backing
    * <code>RecordBuffer</code> instance for a DMO proxy object provided by
    * client code.  It is invoked by the enclosed invocation handler
    * implementation, which services proxied method invocations on DMO proxy
    * objects.
    *
    * @return  Current record buffer (i.e., <code>this</code>).
    */
   public final RecordBuffer buffer()
   {
      return this;
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, String table, String bufName, String widgetPool)
   {
      // TODO implement method body in runtime

      // just a suggestion for implementation, add the widgetPool param, too:
      // Class dmo = null; // dynamically obtain class
      // String database = null;
      // if (bufName == null || "".equals(bufName)
      // {
      // bufName = generateUniqueName();
      // }
      //
      // DataModelObject buf = RecordBuffer.define(dmo, database, bufName);
      // RecordBuffer.openScope(bufName);
      // h.assign(bufName);
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, String table, String bufName, character widgetPool)
   {
      RecordBuffer.create(h, table, bufName, widgetPool.getValue());
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, String table, character bufName, String widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table, bufName.getValue(), widgetPool);
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, String table, character bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table, bufName.getValue(), widgetPool.getValue());
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, character table, String bufName, String widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table.getValue(), bufName, widgetPool);
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, character table, String bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table.getValue(), bufName, widgetPool.getValue());
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, character table, character bufName, String widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table.getValue(), bufName.getValue(), widgetPool);
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * The new new buffer will be created in the specified widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, character table, character bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table.getValue(), bufName.getValue(), widgetPool.getValue());
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, String table, String bufName)
   {
      RecordBuffer.create(h, table, bufName, "");
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, String table, character bufName)
   {
      String buffStr = bufName.isUnknown() ? "" : bufName.getValue();
      RecordBuffer.create(h, table, buffStr, "");
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, character table, String bufName)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table.getValue(), bufName, "");
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, character table, character bufName)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, table.getValue(), bufName.getValue(), "");
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    */
   public static void create(handle h, String table)
   {
      RecordBuffer.create(h, table, "", "");
   }

   /**
    * Creates a dynamic buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses the table-name.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   table
    *          The name of the table to create the buffer for.
    */
   public static void create(handle h, character table)
   {
      if (table.isUnknown())
      {
         RecordBuffer.create(h, table.getValue(), "", "");
      }
      // TODO: else throw some exception ?
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    */
   public static void create(handle h, Temporary tt)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, tt, "", "");
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, Temporary tt, String bufName)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, tt, bufName, "");
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, Temporary tt, character bufName)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, tt, bufName.getValue(), "");
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, Temporary tt, String bufName, String widgetPool)
   {
      // TODO write here the implementation for all overloads of
      // RecordBuffer.create(handle, handle, ...)
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, Temporary tt, String bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, tt, bufName, widgetPool.getValue());
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, Temporary tt, character bufName, String widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, tt, bufName.getValue(), widgetPool);
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses 
    * a temporary table reference. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   tt
    *          The temporary table to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, Temporary tt, character bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, tt, bufName.getValue(), widgetPool.getValue());
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    */
   public static void create(handle h, handle hbuf)
   {
      RecordBuffer.create(h, hbuf, "", "");
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, handle hbuf, String bufName)
   {
      RecordBuffer.create(h, hbuf, bufName, "");
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    */
   public static void create(handle h, handle hbuf, character bufName)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, hbuf, bufName.getValue(), "");
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, handle hbuf, String bufName, String widgetPool)
   {
      // TODO write here the implementation for all overloads of
      // RecordBuffer.create(handle, handle, ...)
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, handle hbuf, String bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, hbuf, bufName, widgetPool.getValue());
   }
   
   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, handle hbuf, character bufName, String widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, hbuf, bufName.getValue(), widgetPool);
   }

   /**
    * Dynamically creates a new buffer and assign it to handle h.
    * This is the correspondent of <code>CREATE-BUFFER</code> statement that uses a another 
    * buffer already assigned to a handle. The new new buffer will be created in the specified 
    * widgwet pool.
    * 
    * @param   h
    *          The handle to which the created buffer is assigned to.
    * @param   hbuf
    *          The handle to other table buffer to create the buffer for.
    * @param   bufName
    *          The name for the buffer.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic buffer.
    */
   public static void create(handle h, handle hbuf, character bufName, character widgetPool)
   {
      // TODO: check for unknown values
      RecordBuffer.create(h, hbuf, bufName.getValue(), widgetPool.getValue());
   }
   
   /**
    * Get a string representation of the internal state of this object,
    * primarily for debugging purposes.
    *
    * @return  String representing this object's state.
    */
   public String toString()
   {
      Persistable tempRecord = getTempRecord();
      if (tempRecord != null)
      {
         StringBuilder bld = new StringBuilder("[TEMP RECORD ACTIVE]\n");
         bld.append(toString(tempRecord));
         
         return bld.toString();
      }
      
      return toString(currentRecord);
   }
   
   /**
    * Initialize the core data on which this buffer relies.  This should be
    * called after the buffer's scope is first opened, but before the buffer
    * is used for a query, or interrogated for any information besides its DMO
    * interface or variable name.
    * 
    * @return  <code>true</code> if the buffer was initialized with this call;
    *          <code>false</code> if the buffer had been initialized with a
    *          previous call to this method.
    *
    * @throws  IllegalArgumentException
    *          if no implementation class is found for the given database and
    *          interface;  if setup for an invalid foreign relation fails.
    * @throws  StopConditionException
    *          if index information cannot be obtained for backing table or
    *          if database isn't connected.
    */
   protected boolean initialize()
   {
      if (persistence != null)
      {
         return false;
      }
      
      // Retrieve logical database name.
      logicalDatabase = getLDBName(ldbOrAlias);

      // Database hasn't been connected.
      if (logicalDatabase == null)
      {
         int[] num     = new int[] {855, 725};
         String[] text = new String[2];
         text[0] = "Unknown database name " + ldbOrAlias;
         text[1] = "Unknown or ambiguous table " + getDMOName();

         ErrorManager.showErrorAndAbend(num, text);
         return false;
      }
      
      // Look up physical database associated with logical name.
      Database database = getPDB(logicalDatabase);
      
      // Get IndexHelper for this database.
      indexHelper = IndexHelper.get(database);
      
      // Get Persistence instance used by this buffer.
      persistence = PersistenceFactory.getInstance(database);
      
      // Get lock context used by this buffer.
      lockContext = persistence.getRecordLockContext();
      
      // Get dirty share context used by this buffer.
      dirtyContext = DirtyShareFactory.getContextInstance(database);
      
      dmoClass = DMOIndex.getImplementingClass(getSchema(), dmoInterface);
      table = DatabaseManager.getBackingTableName(database, dmoClass);
      
      // Generate BufferType, which will be used for a hash key with
      // Reversibles.
      bufferType = new BufferType(database, table, isUndoable());
      
      // Get set of indexed properties for the DMO backing this buffer.
      indexedProperties = getIndexedProperties(false);
      uniqueProperties = getIndexedProperties(true);
      
      // Initialize the validation helper.
      validationHelper = initValidationHelper();

      // Get the delegate for invocation of database triggers.
      DatabaseTriggerManager triggerManager = persistence.getTriggerManager();
      if (triggerManager != null)
      {
         triggerDelegate = triggerManager.getDelegateForDMO(dmoInterface);
      }

      // Register the invocation handler with Persistence as a session
      // listener.
      handler.registerSessionListener();
      
      // Notify the connection manager that this buffer is using the backing,
      // physical database.
      openConnectionManagerScope();
      
      // Store synchers for all foreign key associations from this buffer's
      // table to other tables, and the inverse associations from other
      // tables to this buffer's table.
      if (Persistence.isForeignKeysEnabled())
      {
         String schema = getSchema();
         Iterator<RelationInfo> iter1 =
            DMOIndex.inverseRelations(schema, dmoInterface);
         Iterator<RelationInfo> iter2 =
            DMOIndex.relations(schema, dmoInterface);
         assocSynchers = (iter1.hasNext() || iter2.hasNext()
                          ? new ArrayList<AssociationSyncher>()
                          : null);
         
         if (assocSynchers != null)
         {
            // Create inverse foreign synchers.
            while (iter1.hasNext())
            {
               RelationInfo next = iter1.next();
               assocSynchers.add(new InverseSyncher(this, next));
            }
            
            // Create primary foreign synchers.
            while (iter2.hasNext())
            {
               RelationInfo next = iter2.next();
               assocSynchers.add(new LocalSyncher(this, next));
            }
            
            assocSynchers.trimToSize();
         }
         
         legacyForeignKeys = new HashSet<String>();
         dirtyLegacyKeyProps = new HashSet<String>();
         
         /*
         // Store synchers for all foreign key associations from this buffer's
         // table to other tables.
         String schema = getSchema();
         Iterator<RelationInfo> iter = DMOIndex.relations(schema, dmoInterface);
         assocSynchers = (iter.hasNext()
                          ? new ArrayList<AssociationSyncher>()
                          : null);
         
         if (assocSynchers != null)
         {
            // Create primary association synchers.
            while (iter.hasNext())
            {
               RelationInfo next = iter.next();
               assocSynchers.add(new LocalSyncher(this, next));
            }
            
            assocSynchers.trimToSize();
         }
         */
      }
      else
      {
         assocSynchers = null;
         legacyForeignKeys = null;
         dirtyLegacyKeyProps = null;
      }
      
      if (TransactionManager.isTransaction())
      {
         bufferManager.bufferActiveInTransaction(this, true);
      }
      
      return true;
   }
   
   /**
    * Indicate whether changes made to records managed by this buffer are
    * automatically committed at the earliest opportunity.  This should occur
    * if we are a temp-table buffer and not within an existing transaction.
    *
    * @return  <code>true</code> if auto-commit, else <code>false</code>.
    */
   protected final boolean isAutoCommit()
   {
      return (isTemporary() && !TransactionManager.isTransaction());
   }
   
   /**
    * Indicate whether changes made to records managed by this buffer can be
    * undone during a rollback.  This implementation always returns
    * <code>true</code>.
    *
    * @return  <code>true</code>.
    */
   protected boolean isUndoable()
   {
      return true;
   }
   
   /**
    * Convenience method to indicate whether this buffer is backed by a
    * temporary table.
    * 
    * @return  <code>false</code>.
    */
   protected boolean isTemporary()
   {
      return false;
   }
   
   /**
    * Indicate whether the table backing this buffer is known to contain no
    * rows.
    * 
    * @return  <code>false</code> to indicate it is not definitively known
    *          whether the backing table might have rows.
    */
   protected boolean isTableDefinitelyEmpty()
   {
      return false;
   }
   
   /**
    * Get the logical database name associated with the given parameter.  If
    * the parameter already represents a logical database name, this name is
    * returned.  If it represents a valid alias, the underlying, logical name
    * is returned.
    *
    * @param   ldbOrAlias
    *          A logical database name or database alias.
    *
    * @return  Logical database name.
    */
   protected String getLDBName(String ldbOrAlias)
   {
      ConnectionManager cm = ConnectionManager.get();
      
      // Retrieve logical database name in case this is an alias.
      return cm.getLDBName(ldbOrAlias);
   }
   
   /**
    * Get the physical database name associated with the given logical name.
    *
    * @param   ldbName
    *          A logical database name.
    *
    * @return  Physical database name.
    */
   protected String getPDBName(String ldbName)
   {
      ConnectionManager cm = ConnectionManager.get();
      
      // Convert logical database name to a physical database name.
      return cm.getPDBName(ldbName, true);
   }
   
   /**
    * Get the database information object associated with the given logical
    * database name.
    *
    * @param   ldbName
    *          A logical database name.
    *
    * @return  Standard, temp table database.
    */
   protected Database getPDB(String ldbName)
   {
      ConnectionManager cm = ConnectionManager.get();
      
      // Look up physical database by logical database name.
      return cm.getDatabase(ldbName);
   }
   
   /**
    * Register this buffer with the connection manager.
    */
   protected void openConnectionManagerScope()
   {
      ConnectionManager.get().openBufferScope(this);
   }
   
   /**
    * Deregister this buffer with the connection manager.
    */
   protected void closeConnectionManagerScope()
   {
      ConnectionManager.get().closeBufferScope(this);
   }
   
   /**
    * Create a new instance of the DMO implementation class associated with
    * this buffer.  All default data values are assigned, but the new instance
    * is not yet associated with the current database session.
    *
    * @return  A transient DMO instance.
    *
    * @throws  InstantiationException
    *          if the class cannot be instantiated.
    * @throws  IllegalAccessException
    *          if the DMO class' target constructor cannot be accessed.
    * @throws  InvocationTargetException
    *          if the underlying constructor throws an exception.
    */
   protected Persistable instantiateDMO()
   throws InstantiationException,
          IllegalAccessException,
          InvocationTargetException
   {
      return (Persistable) getDMOImplementationClass().newInstance();
   }
   
   /**
    * Open a new buffer scope for this record buffer.  This ensures that the
    * {@link BufferManager} is tracking this buffer for block scope entries
    * and exits within transactions, such that this buffer can be registered
    * with the <code>TransactionManager</code> for commit/rollback
    * notifications.  The buffer is also registered for finish processing so
    * that the appropriate record and lock transition processing can occur
    * when the buffer scope closes.
    */
   protected void openScope()
   {
      if (retrying)
      {
         // reset the flag
         retrying = false;

         // this is a retry, nothing to do here
         return;
      }
      
      boolean outermost = (openScopeCount == 0);
      
      // Register this scope with the buffer manager.
      bufferManager.openScope(this, !outermost);
      
      if (outermost)
      {
         // If this is the outermost scope open for this buffer, remember the
         // scope depth at which it was opened.  This is tracked so that we
         // know when it is safe to release resources.
         activeScopeDepth = blockDepth;
         
         // Listen for changes driven by this or other buffers for the same
         // DMO type.
         changeBroker.addListener(this);
         
         // Register lightweight undoable to reset current record for undo
         // and retry; this is only for undoable tables.
         if (isUndoable())
         {
            TransactionManager.registerUndo(false, new LightweightUndoable());
         }
      }
      
      boolean active = isActive();
      
      if (active)
      {
         openConnectionManagerScope();
      }
      
      // Register this buffer as a Finalizable with the TM, so that its
      // finished method is called when this buffer scope is about to close.
      TransactionManager.registerFinalizable(this, false);
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() +
                   ":  opened " +
                   (active ? "" : "in") +
                   "active buffer scope " +
                   bufferManager.getCurrentScope());
      }
      
      // Increment reference count of open scopes for this buffer.
      openScopeCount++;
   }
   
   /**
    * Get the multiplex ID of this buffer.  This default implementation
    * always returns <code>null</code>, but subclasses may return a
    * non-<code>null</code> value.
    *
    * @return  <code>null</code>.
    */
   protected Integer getMultiplexID()
   {
      return null;
   }
   
   /**
    * Indicate whether the backing table for this buffer is multiplexed;  that
    * is, whether the backing table represents multiple, virtual tables by
    * using a multiplex ID to differentiate table segments.
    * <p>
    * This default implementation always returns <code>false</code> and should
    * be overridden by subclasses which require a different return value.
    *  
    * @return  <code>false</code>.
    */
   protected boolean isMultiplexed()
   {
      return false;
   }
   
   /**
    * Indicate whether this buffer implementation allows bulk actions.
    *  
    * @return  <code>false</code>.
    */
   protected boolean allowsBulkActions()
   {
      return false;
   }
   
   /**
    * Retrieve the master record buffer for this buffer.
    * <p>
    * This only is meaningful for shared temp tables and should be overridden
    * by subclasses.
    * 
    * @return  This buffer.
    */
   protected RecordBuffer getMasterBuffer()
   {
      return this;
   }
   
   /**
    * Delete the current record from the database.  Set the current record to
    * null if the deletion is successful.
    *
    * @throws  PersistenceException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   protected void delete()
   throws PersistenceException
   {
      if (currentRecord == null)
      {
         // Nothing to delete!
         throw new PersistenceException(
            "There is no current " + getDMOName() + " record. Delete failed",
            341);
      }
      
      // Though it may seem counterintuitive, we need to persist any transient
      // record, so that it is not persisted at the end of this method in
      // response to the reported change.
      try
      {
         flush();
      }
      catch (ValidationException exc)
      {
         reportValidationException(exc);
      }

      // invoke database triggers
      if (triggerDelegate != null)
      {
         // note that ConditionException may be thrown by this method
         triggerDelegate.executeDeleteTriggers(getDMOProxy(), false);
      }
      
      upgradeLock();

      // Null out any foreign key references to the deleted DMO.
      if (foreignNullers == null)
      {
         List<ForeignNuller> l = ForeignNuller.instances(this);
         foreignNullers = Collections.unmodifiableList(l);
      }
      if (!foreignNullers.isEmpty())
      {
         ForeignNuller.breakLinkages(foreignNullers, currentRecord);
      }
      
      // Make sure the copy of currentRecord we are about to archive in
      // a ReversibleDelete is fully initialized (specifically, its lazy,
      // persistent collections, if any).
      initializeLazyExtents();
      
      persistence.delete(currentRecord);
      
      // Notify dirty share context that this record is deleted.
      Serializable id = currentRecord.getId();
      dirtyContext.delete(getEntityName(), id, false, true);
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() + ":  record " + id + " deleted");
         if (LOG.isTraceEnabled())
         {
            LOG.trace(toString());
         }
      }
      
      setReversibleDelete();
      
      // Notify listeners of the change.  This will clear the buffer in
      // stateChanged().  We can't clear the buffer before reporting the
      // change, in case the current record gets evicted upon release, and
      // ChangeBroker doesn't yet know that a flush is necessary due to the
      // delete.
      snapshot = currentRecord;
      reportChange(currentRecord, false, true);
   }
   
   /**
    * Perform a bulk delete, restricted by the given where clause and query
    * substitution parameters.
    * <p>
    * This default implementation is an unsupported operation and raises an
    * exception.  Subclasses which support bulk delete must override this
    * method.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  PersistenceException
    *          if an error occurs performing the bulk delete operation.
    * @throws  UnsupportedOperationException
    *          always in the default implementation.
    */
   protected void delete(String where, Object[] args)
   throws PersistenceException
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Perform a bulk delete of all records in the table associated with this
    * buffer.
    * <p>
    * DELETE FOR was also tested with Progress 4GL with WHERE clause. It is not needed for this
    * method, but behaviour is the same. As result it is noticed that even such statement also
    * waits for whole table even if only one record is locked independently whether this record
    * satisfies WHERE clause or not . So, it seems idea of table-level locking is correct.
    * <p>
    */
   protected void deleteAll()
   {
      // TODO: current table lock level is queried/stored;
      initialize();

      DataModelObject dmo = getDMOProxy();
      final PreselectQuery query = new PreselectQuery(dmo, null, null, "book.bookId asc");

      BlockManager.forEach(BlockManager.TransactionType.FULL, null, new Block()
      {
         public void body()
         {
            query.next();
            try
            {
               delete();
            }
            catch (PersistenceException e)
            {
               ErrorManager.recordOrThrowError(e);
            }
         }
      });

      // Notify all interested parties that a bulk delete occurred.
      // Listening buffers may need to release deleted records.
      RecordChangeEvent.Type eventType = RecordChangeEvent.Type.FULL_BULK_DELETE;
      try
      {
         changeBroker.stateChanged(eventType, this, null, null, null);
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
      }

      // TODO: exclusive table lock is acquired;
      // TODO: original table lock level is restored.
   }

   /**
    * Perform a bulk delete of all records in the table associated with this
    * buffer.
    * <p>
    * DELETE FOR was also tested with Progress 4GL with WHERE clause. It is not needed for this
    * method, but behaviour is the same. As result it is noticed that even such statement also
    * waits for whole table even if only one record is locked independently whether this record
    * satisfies WHERE clause or not . So, it seems idea of table-level locking is correct.
    * <p>
      Below is Savepoint based approach for deleteAll.
      Since we don't currently have simple answers for the Hibernate second-level cache or
      for dealing with any records currently held in buffers, we'd like to first use a simple,
      "converted-style" FOR-EACH loop as the basis for the RecordBuffer.deleteAll
      implementation. This approach leverages the code we already have to deal with these
      issues. If we find this causes memory problems, due to the bulk delete of very large
      tables, we can revisit the savepoint implementation and figure out these remaining issues.
    */
   /*
      protected void deleteAll()
      {
         // TODO: current table lock level is queried/stored;

         initialize();
         try
         {
            flushAll();
         }
         catch (ValidationException exc)
         {
            reportValidationException(exc);
         }

         StringBuilder buf = new StringBuilder("delete from ");
         String dmoImplName = getDMOImplementationName();
         buf.append(dmoImplName);
         String hql = buf.toString();

         try {
            int rowCount = persistence.deleteOrUpdate(hql, null, null);
            if (rowCount > 0)
            {
               changeBroker.forcePendingFlush(this);

               if (TransactionManager.isTransaction())
               {
                  // force the records to be deleted, on db tx rollback
                  SavepointReversible rev = new SavepointReversible(getDMOName(),
                                                                    "BULK DELETE",
                                                                    persistence.bind());
                  bufferManager.addReversible(this, null, rev, false);
               }

               // Notify all interested parties that a bulk delete occurred.
               // Listening buffers may need to release deleted records.
               RecordChangeEvent.Type eventType = RecordChangeEvent.Type.FULL_BULK_DELETE;
               changeBroker.stateChanged(eventType, this, null, null, null);
            }

         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }

         // TODO: exclusive table lock is acquired;
         // TODO: original table lock level is restored.
      }
   */

   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    *
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    *
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where, Object...args)
   {
      aggregate("avg", column, num, where, args);
   }

   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Optional query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where, Object...args)
   {
      aggregate("count", column, num, where, args);
   }

   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    *
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    *
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where, Object...args)
   {
      aggregate("min", column, num, where, args);
   }

   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    *
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    *
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where, Object...args)
   {
      aggregate("max", column, num, where, args);
   }

   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    *
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    *
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where, Object...args)
   {
      aggregate("sum", column, num, where, args);
   }

   /**
    * Reclaim the given list of primary keys for re-use.
    * 
    * @param   keys
    *          A list of primary keys which are available for re-use.  Must
    *          not be <code>null</code>.
    */
   protected void reclaimKeys(List<Serializable> keys)
   {
      checkActive();
      
      IdentityManager identityManager = persistence.getIdentityManager();
      identityManager.reclaimKeys(keys.toArray(new Serializable[0]));
   }

   /**
    * Get the map of DMO property names to getter methods, initializing it
    * first if necessary.
    *
    * @return  Map of DMO getter methods.
    */
   protected Map<String, Method> getGetterMap()
   {
      if (getterMap == null)
      {
         getterMap = PropertyHelper.gettersByProperty(dmoInterface);
      }
      
      return getterMap;
   }
   
   /**
    * Get the map of DMO property names to setter methods, initializing it
    * first if necessary.
    *
    * @return  Map of DMO setter methods.
    */
   protected Map<String, Method> getSetterMap()
   {
      if (setterMap == null)
      {
         setterMap = PropertyHelper.settersByProperty(dmoInterface);
      }
      
      return setterMap;
   }
   
   /**
    * Get the map of DMO property names to sizeXxxx methods, initializing it
    * first if necessary.
    *
    * @return  Map of DMO sizer methods.
    */
   protected Map<String, Method> getSizerMap()
   {
      if (sizerMap == null)
      {
         sizerMap = PropertyHelper.sizersByProperty(dmoInterface);
      }
      
      return sizerMap;
   }
   
   /**
    * Increment the context local use count for the specific instance of the
    * given record.  Used in conjunction with {@link #decrementDMOUseCount} to
    * implement an aggressive DMO eviction policy.  Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * Use counts must be maintained because multiple buffers may hold the
    * same instance of a record.  A record must not be evicted while any
    * buffer still needs it.
    *
    * @param   dmo
    *          DMO instance whose reference count is to be incremented.
    * 
    * @return  Use count for <code>dmo</code> after incrementing.
    *
    * @see     BufferManager#incrementDMOUseCount
    */
   protected int incrementDMOUseCount(Persistable dmo)
   {
      checkActive();
      
      return bufferManager.incrementDMOUseCount(persistence, dmo);
   }
   
   /**
    * Decrement the context local use count for the specific instance of the
    * given record.  Used in conjunction with {@link #incrementDMOUseCount} to
    * implement an aggressive DMO eviction policy.  Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * Use counts must be maintained because multiple buffers may hold the
    * same instance of a record.  A record must not be evicted while any
    * buffer still needs it.
    *
    * @param   dmo
    *          DMO instance whose reference count is to be decremented.
    * 
    * @return  Use count for <code>dmo</code> after decrementing.
    *
    * @throws  PersistenceException
    *          if an error occurs detaching <code>dmo</code> from the
    *          persistence session.
    *
    * @see     BufferManager#decrementDMOUseCount
    */
   protected int decrementDMOUseCount(Persistable dmo)
   throws PersistenceException
   {
      checkActive();
      
      return bufferManager.decrementDMOUseCount(persistence, dmo);
   }
   
   /**
    * Evict the given DMO from the current Hibernate session if it is not
    * referenced by any buffer or other resource.  This method does not
    * modify the DMO's use count, if any.  It is intended to quickly evict
    * DMOs loaded into the session by side effect.
    * 
    * @param   dmo
    *          DMO instance whose reference count is to be decremented.
    * 
    * @return  <code>true</code> if the DMO was evicted;  <code>false</code>
    *          if it was referenced by other resources.
    *
    * @throws  PersistenceException
    *          if an error occurs detaching <code>dmo</code> from the
    *          persistence session.
    */
   protected boolean evictDMOIfUnused(Persistable dmo)
   throws PersistenceException
   {
      checkActive();
      
      return bufferManager.evictDMOIfUnused(persistence, dmo);
   }
   
   /**
    * Report any pending changes in the current record's state to the {@link
    * ChangeBroker} for broadcast to interested listeners.  A change includes
    * any modification to a property or the deletion of the record.  If there
    * is no change to report, this method returns immediately.
    * <p>
    * If this buffer has not previously taken a snapshot of its current
    * record, a deep copy is made of <code>dmo</code> (which represents the
    * post-modification state of this buffer's current record).  Any
    * unreported changes are then rolled back from this copy, so that the
    * copy represents a snapshot of the record at the time it was first
    * stored in the buffer.
    * <p>
    * This seemingly tortuous method of lazy initialization of the snapshot
    * is used because it allows us in certain cases to avoid ever making a
    * deep copy of the buffer's current record.  If that record represents a
    * Hibernate proxy which was never initialized, we avoid the overhead of
    * ever hydrating the proxy, which in tight loops can result in a HUGE
    * time savings.
    * <p>
    * The snapshot is then passed along with <code>dmo</code>, this buffer,
    * and a map of property names to getter method indexes (for extent/array
    * properties), to the change broker for dispatching to listeners.
    *
    * @param   dmo
    *          DMO which represents the post-modification state of the
    *          current record.  In the case of record deletion, this object
    *          represents the state of the record just before deletion.
    * @param   insert
    *          <code>true</code> if change represents a record insertion,
    *          otherwise <code>false</code>.
    * @param   delete
    *          <code>true</code> if change represents a record deletion,
    *          otherwise <code>false</code>.
    *
    * @throws  PersistenceException
    *          if any reflection error occurs taking the snapshot, or if the
    *          change broker forwards an exception from one of its listeners.
    */
   @SuppressWarnings("unchecked")
   protected void reportChange(Persistable dmo,
                               boolean insert,
                               boolean delete)
   throws PersistenceException
   {
      // Sanity check.
      if (!delete                      &&    // is not a delete
          !insert                      &&    // is not an insert
          unreportedChanges.isEmpty())       // is not an update
      {
         return;
      }
      
      try
      {
         Persistable snapshot = (insert ? null : this.snapshot);
         boolean takeSnapshot = (snapshot == null && !insert);
         
         if (takeSnapshot)
         {
            // Bind Hibernate session to Persistence object if necessary.
            if (bindNeeded)
            {
               persistence.bind();
               bindNeeded = false;
            }
            
            // DMO parameter represents the current record post-change (except
            // in the delete case), so we deep copy it, then apply the old
            // values selectively below, to reconstruct the original state.
            // In the delete case, there are no diffs to apply, but we still
            // need the snapshot.
            snapshot = (Persistable) dmo.deepCopy();
         }
         
         // Property/extent map for change event.
         Map<String, List<Integer>> properties =
            new HashMap<String, List<Integer>>();
         
         if (!unreportedChanges.isEmpty())
         {
            try
            {
               Map<String, Method> setterMap = getSetterMap();
               
               Iterator<Map.Entry<String, Object>> iter =
                  unreportedChanges.entrySet().iterator();
               while (iter.hasNext())
               {
                  Map.Entry<String, Object> next = iter.next();
                  String property = next.getKey();
                  Object value = next.getValue();
                  
                  if (value instanceof Map)
                  {
                     Map<Integer, Object[]> map =
                        (Map<Integer, Object[]>) value;
                     List<Integer> indices =
                        new ArrayList<Integer>(map.keySet());
                     properties.put(property, indices);
                     Iterator<Map.Entry<Integer, Object[]>> iter2 =
                        map.entrySet().iterator();
                     while (takeSnapshot && iter2.hasNext())
                     {
                        // Apply old value of property to snapshot.
                        Map.Entry<Integer, Object[]> entry = iter2.next();
                        Integer index = entry.getKey();
                        Object[] diffs = entry.getValue();
                        Object oldVal = diffs[0];
                        Method setter = setterMap.get(property);
                        Object[] args = new Object[] { index, oldVal };
                        setter.invoke(snapshot, args);
                     }
                  }
                  else
                  {
                     properties.put(property, null);
                     if (takeSnapshot)
                     {
                        // Apply old value of property to snapshot.
                        Object[] diffs = (Object[]) value;
                        Object oldVal = diffs[0];
                        Method setter = setterMap.get(property);
                        Object[] args = new Object[] { oldVal };
                        setter.invoke(snapshot, args);
                     }
                  }
               }
            }
            catch (Exception rflExc)
            {
               DBUtils.handleException(getDatabase(), rflExc);
               
               throw new PersistenceException(rflExc);
            }
         }
         
         if (LOG.isTraceEnabled())
         {
            String desc = describe();
            LOG.trace(desc + ":  DMO state changed...");
            LOG.trace(desc + ":  Old state:  " + toString(snapshot));
            LOG.trace(desc + ":  New state:  " + toString());
            LOG.trace(desc + ":  Properties/Extents:  " + properties);
         }
         
         // Notify listeners.
         RecordChangeEvent.Type type = null;
         if (insert)
         {
            type = RecordChangeEvent.Type.INSERT;
         }
         else if (delete)
         {
            type = RecordChangeEvent.Type.DELETE;
         }
         else
         {
            type = RecordChangeEvent.Type.UPDATE;
         }
         changeBroker.stateChanged(type, this, dmo, snapshot, properties);
      }
      finally
      {
         unreportedChanges.clear();
      }
   }
   
   /**
    * Provide a short descriptor of this buffer instance for debugging and
    * logging purposes.  The descriptor includes the buffer's database name,
    * short DMO interface name, its variable name, and the identifier for the
    * current context.
    *
    * @return  Short text descriptor.
    */
   protected String describe()
   {
      StringBuilder buf = new StringBuilder("[");
      buf.append(Utils.describeContext());
      buf.append("-->");
      if (isActive())
      {
         buf.append(getDatabase());
      }
      else
      {
         buf.append("'");
         buf.append(ldbOrAlias);
         buf.append("'");
      }
      buf.append(".");
      buf.append(getDMOName());
      Integer mpid = getMultiplexID();
      if (mpid != null)
      {
         buf.append(":");
         buf.append(mpid);
      }
      buf.append("@");
      buf.append(blockDepth);
      buf.append(" (");
      buf.append(variable);
      buf.append(")]");
      
      return buf.toString();
   }
   
   /**
    * Attempt to upgrade to an exclusive lock on the current record, if we do
    * not already have one.  The current context must already possess a share
    * lock (or greater) for this method to complete successfully.  The current
    * thread will block inside the lock manager if the lock is not immediately
    * available.  When this method returns normally, the caller can assume the
    * lock has been acquired.
    *
    * @throws  PersistenceException
    *          if the current context does not have a share lock (or greater)
    *          on the current record.
    */
   protected void upgradeLock()
   throws PersistenceException
   {
      Serializable id = currentRecord.getId();
      RecordIdentifier ident = new RecordIdentifier(table, id);
      LockType currentType = persistence.queryLock(ident);
      
      if (currentType == LockType.NONE)
      {
         throw new PersistenceException(
            getDMOName() +
            " record has NO-LOCK status, update to field not allowed",
            396);
      }
      
      lockContext.lock(ident, LockType.EXCLUSIVE, this);
   }
   
   /**
    * Retrieve the set of all property names for this buffer's DMO entity,
    * which represent database columns that participate in an index on the
    * entity's backing table, or only the subset which participate in unique
    * indexes.
    * 
    * @param   unique
    *          <code>false</code> to get all indexed properties;
    *          <code>true</code> to get only those that participate in unique
    *          indexes.
    * 
    * @return  Set of indexed property names.  The set may be empty but it
    *          will never be <code>null</code>.
    * 
    * @throws  StopConditionException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   protected Set<String> getIndexedProperties(boolean unique)
   {
      checkActive();
      
      try
      {
         return indexHelper.getIndexedProperties(getEntityName(), unique);
      }
      catch (PersistenceException exc)
      {
         LOG.error("Unable to query index metadata", exc);
         
         throw new StopConditionException(exc);
      }
   }
   
   /**
    * Generate a formatted description of the contents of <code>record</code>.
    * The report includes the name and current value of every data property
    * within the record.
    *
    * @param   record
    *          Record whose contents are to be written to the report.
    *
    * @return  A formatted report of <code>record</code>'s contents.
    */
   protected String toString(Persistable record)
   {
      boolean active = isActive();
      String sep = System.getProperty("line.separator");
      StringBuffer buf = new StringBuffer();
      
      buf.append(active ? getDatabase() : ldbOrAlias);
      buf.append(" ");
      buf.append(getDMOName());
      buf.append(" (");
      buf.append(getDMOAlias());
      buf.append("@");
      buf.append(blockDepth);
      
      if (!active)
      {
         buf.append(") - INACTIVE");
         
         return buf.toString();
      }
      
      buf.append(":");
      buf.append(activeScopeDepth);
      buf.append("#");
      buf.append(openScopeCount);
      buf.append(") [");
      buf.append(table);
      buf.append("]").append(sep);
      buf.append("---------------------------------------------").append(sep);
      
      if (record == null)
      {
         buf.append("   No record");
         
         return buf.toString();
      }
      
      String name = null;
      try
      {
         Map<String, Method> getters = new LinkedHashMap<String, Method>();
         int maxWidth = 0;
         Method[] methods = dmoInterface.getDeclaredMethods();
         
         int len = methods.length;
         for (int i = 0; i < len; i++)
         {
            Method next = methods[i];
            name = next.getName();
            boolean prefixGet = name.startsWith("get");
            boolean prefixIs = name.startsWith("is");
            if ((prefixGet || prefixIs) &&
                next.getParameterTypes().length == 0 &&
                !Collection.class.isAssignableFrom(next.getReturnType()))
            {
               name = name.substring(prefixGet ? 3 : 2);
               name = Introspector.decapitalize(name);
               getters.put(name, next);
               maxWidth = Math.max(maxWidth, name.length());
            }
         }
         
         buf.append("   ");
         StringHelper.leftAlignText("id", maxWidth, buf);
         buf.append(" :  ");
         buf.append(record.getId());
         
         Iterator<String> iter = getters.keySet().iterator();
         while (iter.hasNext())
         {
            buf.append(sep);
            name = iter.next();
            Method next = getters.get(name);
            buf.append("   ");
            StringHelper.leftAlignText(name, maxWidth, buf);
            buf.append(" :  ");
            Object value = next.invoke(record, (Object[]) null);
            if (value instanceof BaseDataType)
            {
               buf.append(((BaseDataType) value).toStringMessage());
            }
            else
            {
               buf.append(value);
            }
         }
      }
      catch (Exception exc)
      {
         DBUtils.handleException(getDatabase(), exc);
         
         if (LOG.isWarnEnabled())
         {
            LOG.warn(describe() + ":  error dumping record contents --> " +
                     getDMOName() + ":" + record.getId() + " at property '" +
                     name + "'",
                     exc);
         }
      }
      
      buf.append(sep);
      
      return buf.toString();
   }
   
   /**
    * Returns the primary key for a new DMO associated with this buffer.
    *
    * @return   The primary key for a new DMO associated with this buffer.
    *
    * @throws   PersistenceException
    *           if there was an error while retrieving the next primary key.
    */
   protected Serializable nextPrimaryKey()
   throws PersistenceException
   {
      checkActive();
      
      return persistence.nextPrimaryKey(table);
   }
   
   /**
    * Initialize the validation helper object.
    * 
    * @return  Validation helper object.
    */
   protected ValidationHelper initValidationHelper()
   {
      return (new ValidationHelper());
   }
   
   /**
    * Initializes the map that contains primary keys of past and present
    * currentRecords and their pinned LockTypes.
    * <p>
    * Add a user defined object to the initial scope, which is a set which
    * will hold the primary keys of records affected by a rollback in that
    * scope.  The same instance of this set is pushed onto every new scope, so
    * this information remains available as scopes are popped off the stack.
    */
   protected void initPinnedLockTypes()
   {
      Set<Serializable> undone = new HashSet<Serializable>(8);
      pinnedLockTypes = new ScopedDictionary<Serializable, LockType>(undone);
   }
   
   /**
    * Push a new scope onto the pinned lock types scoped dictionary.  Add a
    * user defined object to the new scope, which is a set which will hold the
    * primary keys of records affected by a rollback in that scope.  The same
    * instance of this set is pushed onto every new scope, so this information
    * remains available as scopes are popped off the stack.
    */
   protected void pushPinnedLockScope()
   {
      Set<Serializable> undone =
         (Set<Serializable>) pinnedLockTypes.getScope();
      pinnedLockTypes.addScope(undone);
   }
   
   /**
    * Pop the current scope from the pinned lock types scoped dictionary,
    * propagating all entries in this scope up to the next higher scope, if
    * any.  If the current scope is being undone, we mark each record which
    * had its lock type altered in the current scope, so that it will not
    * propagate past the current transaction boundary.  If the current scope
    * is being committed, we unmark any such records, if they are also
    * present in the current scope.
    * 
    * @param   transaction
    *          <code>true</code> if we are at a full transaction boundary,
    *          else <code>false</code>.
    * @param   undo
    *          <code>true</code> if we are processing an undo on the current
    *          scope, else <code>false</code>.
    */
   protected void popPinnedLockScope(boolean transaction, boolean undo)
   {
      Set<Serializable> allKeys =
         pinnedLockTypes.getDictionaryAtScope(0, false).keySet();
      Set<Serializable> undone =
         (Set<Serializable>) pinnedLockTypes.getScope();
      
      // If the current scope was undone, mark all records whose lock types
      // were added in the current scope, so that these changes are not
      // propagated beyond the transaction boundary.
      if (undo)
      {
         undone.addAll(allKeys);
      }
      
      // If we are at a full transaction boundary, we remove all records
      // which have been marked as "undone" from the current scope, before
      // propagating to the parent scope.
      if (transaction)
      {
         allKeys.removeAll(undone);
         undone.clear();
      }
      
      // Delete the current scope, first propagating all entries up to the
      // parent scope.
      pinnedLockTypes.deleteScope(true);
   }
   
   /**
    * Set the lock type which potentially "pins" the communal record lock on
    * the record with the given id (the record was ever or is currently loaded
    * into this buffer).
    *
    * @param   id
    *          Record identifier.
    * @param   pinnedLockType
    *          New pinned lock type.
    *
    * @see     RecordLockContext
    */
   protected void setPinnedLockType(Serializable id, LockType pinnedLockType)
   {
      LockType waitLockType = pinnedLockType.toWaitVariant();
      if (waitLockType == LockType.NONE)
      {
         pinnedLockTypes.removeEntryAtScope(id, 0);
      }
      else
      {
         pinnedLockTypes.addEntry(false, id, waitLockType);
         
         // Since this lock type is being set explicitly in the current
         // scope, remove it from the set of IDs marked as undone.  This will
         // allow proper downgrade/release behavior when/if we transition out
         // of a transaction.
         Set<Serializable> undone =
            (Set<Serializable>) pinnedLockTypes.getScope();
         undone.remove(id);
      }
   }
   
   /**
    * Get the lock type which potentially "pins" the communal record lock on
    * the record with the given identifier (the record was ever or is
    * currently loaded into this buffer)
    *
    * @param   id
    *          Record identifier.
    *
    * @return  Pinned lock type of the specified record.
    *
    * @see     RecordLockContext
    */
   protected LockType getPinnedLockType(Serializable id)
   {
      LockType lockType = pinnedLockTypes.lookup(id);
      
      return (lockType == null ? LockType.NONE : lockType);
   }
   
   /**
    * Indicate whether this buffer is active.  A buffer is active if it has
    * been used by a query or otherwise initialized.  Opening a buffer scope
    * alone is not enough to activate a buffer, but it is a prerequisite to
    * activation;  the buffer must actually be used to be considered active.
    * 
    * @return  <code>true</code> if the buffer has been initialized.
    */
   boolean isActive()
   {
      return (persistence != null);
   }
   
   /**
    * Record or throw a &quot;record not on file&quot; error condition for
    * the current DMO type.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is suppressed.
    */
   void errorNotOnFile()
   {
      StringBuilder buf = new StringBuilder(getDMOAlias());
      buf.append(" record not on file");
      String msg = buf.toString();
      
      if (LOG.isDebugEnabled())
      {
         PersistenceException exc = new PersistenceException(msg, 138);
         LOG.debug(msg, exc);
      }
      
      ErrorManager.recordOrThrowError(138, msg);
   }
   
   /**
    * Get the number of scopes currently open for this buffer.  This is
    * necessary because buffer scopes can be nested.
    *
    * @return  Number of scopes currently open for this buffer.
    */
   int getOpenScopeCount()
   {
      return openScopeCount;
   }
   
   /**
    * Get the depth of the scope at which this buffer was first opened.
    * Nested scopes are ignored;  only the depth of the outermost open scope
    * is reported.
    *
    * @return  Depth at which outermost scope was opened for this buffer.
    */
   int getScopeOpenDepth()
   {
      return activeScopeDepth;
   }
   
   /**
    * Validate and release the current record in this buffer and release its
    * lock, if any, according to the following rules:
    * <ul>
    *   <li>the lock is released at the later of:
    *   <ul>
    *     <li>end of the current transaction (if any);
    *     <li>end of the current buffer scope.
    *   </ul>
    *   <li>the lock is downgraded at the end of the current transaction,
    *       if the current buffer scope is larger than the transaction scope.
    *   <li>if the lock never passed through a transaction, it is released
    *       immediately.
    * </ul>
    *
    * @throws  ErrorConditionException
    *          if validation of the current buffer record (if any) fails.
    */
   void release()
   {
      if (currentRecord != null)
      {
         setCurrentRecord(null, false, false, false);
      }
   }
   
   /**
    * Release the current record and discard the reference record snapshot.
    */
   void reset()
   {
      release();
      snapshot = null;
   }
   
   /**
    * Get the DMO which represents the database record currently held in the
    * buffer.
    *
    * @return  DMO for the current record, or <code>null</code> if the
    *          buffer currently is empty.
    */
   Persistable getCurrentRecord()
   {
      return currentRecord;
   }
   
   /**
    * Reload the current record, if any, from the database.  If
    * <code>lockType</code> is higher than the current lock type, it is
    * upgraded immediately.  If lower, it is downgraded immediately only if
    * outside of a transaction.  If inside a transaction, the lock is marked
    * for downgrade to the specified type upon transaction end.
    *
    * @param   lockType
    *          Requested lock type.
    *
    * @throws  ErrorConditionException
    *          if there is no record loaded into the buffer and the error
    *          manager is not in silent mode.
    */
   void reload(LockType lockType)
   {
      load(currentRecord, lockType, false);
   }
   
   /**
    * Set the DMO implementation instance which backs this buffer.  It is
    * expected that this variant is invoked only by queries storing their
    * retrieved record.
    * <p>
    * Setting the value to <code>null</code> may result in an error or query
    * off-end condition, depending upon the value of <code>errorIfNull</code>
    * and the current silent error mode.
    * <p>
    * It is expected that the calling query specify the sort index which was
    * in use and whether the query ran off-end (and in which direction).  This
    * information is useful to {@link FindQuery find queries} which must
    * determine an appropriate starting point for their search.
    *
    * @param   currentRecord
    *          DMO instance which represents the record currently backing
    *          this buffer.
    * @param   lockType
    *          Lock type (already applied to a non-<code>null</code>
    *          <code>currentRecord</code>), which will be pinned by this
    *          record buffer.
    * @param   errorIfNull
    *          <code>false</code> to raise an end condition if the record
    *          being set is <code>null</code>;  else <code>true</code>
    *          (caller is responsible for handling appropriately in the
    *          latter case).
    * @param   sortIndex
    *          Sort index which was in use at the time the given record was
    *          fetched.
    * @param   offEnd
    *          Enum indicating whether query is off-end, and if so, in which
    *          direction.
    * @param   inverseSorting
    *          Indicates whether <code>offEnd</code> is specified against the
    *          inverse sorting direction.
    * @param   dirtyCopy
    *          Indicates whether <code>currentRecord</code> represents a copy
    *          of a DMO found in the dirty database.
    * 
    * @throws  QueryOffEndException
    *          if <code>currentRecord</code> is <code>null</code> and
    *          <code>errorIfNull</code> is <code>false</code>.
    * 
    * @see     #getOffEnd(SortIndex,boolean)
    */
   void setRecord(Persistable currentRecord,
                  LockType lockType,
                  boolean errorIfNull,
                  SortIndex sortIndex,
                  OffEnd offEnd,
                  boolean inverseSorting,
                  boolean dirtyCopy)
   {
      setCurrentRecord(currentRecord, false, false, dirtyCopy);
      this.sortIndex = sortIndex;
      this.offEnd    = inverseSorting ? offEnd.inverse() : offEnd;
      
      if (currentRecord == null)
      {
         if (!errorIfNull)
         {
            throw new QueryOffEndException("",
                                new QueryOffEndListener[] { offEndListener });
         }
      }
      else
      {
         Serializable id = currentRecord.getId();
         if (lockType != LockType.NONE)
         {
            RecordIdentifier ident = new RecordIdentifier(table, id);
            lockContext.registerBuffer(ident, this);
         }
         setPinnedLockType(id, lockType);
      }
   }
   
   /**
    * Get an indication whether the last current record set by a query:
    * <ol>
    *   <li>was <code>null</code> (meaning the query was off-end), and if so
    *   <li>whether it used the same sort index as the one specified, and if
    *       so
    *   <li>in which direction the query ran off-end.
    * </ol>
    * <p>
    * In the event no query has yet set a record in this buffer, a value of
    * <code>OffEnd.FRONT</code> will be returned.  In the event the last query
    * neither ran off-end, nor used the given sort index, a value of
    * <code>OffEnd.NONE</code> will be returned.
    * 
    * @param   sortIndex
    *          Sort index to compare with the last sort index used to
    *          populate (or clear) the buffer.
    * @param   inverseSorting
    *          Indicates whether the off-end is requested by a query having
    *          the inverse sorting direction.
    * 
    * @return  See above.
    * 
    * @see     #setRecord(Persistable,
    *                     LockType,
    *                     boolean,
    *                     SortIndex,
    *                     OffEnd,
    *                     boolean,
    *                     boolean)
    */
   OffEnd getOffEnd(SortIndex sortIndex, boolean inverseSorting)
   {
      if (this.sortIndex == null || this.sortIndex.equals(sortIndex))
      {
         return inverseSorting ? offEnd.inverse() : offEnd;
      }
      
      return OffEnd.NONE;
   }

   /** 
    * Get the {@link QueryOffEndListener} instance for this record buffer.
    * 
    * @return the {@link QueryOffEndListener} instance for this record buffer.
    */
   QueryOffEndListener getQueryOffEndListener()
   {
      return offEndListener;
   }
   
   /**
    * Push a temporary record context onto the buffer for purposes such as
    * client-side sorting.  Such a context must be pushed before {@link
    * #setTempRecord(Persistable) setting a temporary record} is possible.
    * A context allows one or more temporary records to be set for a specific
    * operation, such as the execution of client-side where clause
    * expressions by a query, and for such operations to be nested.  For
    * example, a client-side where expression may eventually execute a query
    * which then itself may execute a client-side where expression.  Each
    * level of client-side where clause expression must execute within its
    * own temporary record context, such that the state of the outer where
    * clause expression is not lost when the inner expression is finished.
    * When client code is finished with a temporary record context, it should
    * {@link #popTempContext() pop} the current context.
    * <p>
    * Pushes and pops of the temporary record context must be balanced.  This
    * is the responsibility of the caller.
    *
    * @see     WhereExpression
    */
   void pushTempContext()
   {
      if (tempRecords == null)
      {
         // It is rare that we would need a depth greater than 1, so make the
         // default size 1.
         tempRecords = new ArrayList<Persistable>(1);
      }
      
      tempRecords.add(null);
   }
   
   /**
    * Pop the current, temporary record context from this buffer.
    * <p>
    * As a convenience, the currently active temporary record for this buffer
    * is set to <code>null</code> first.
    * 
    * @throws  PersistenceException
    *          if setting the current, temporary record to <code>null</code>
    *          triggers a failure to evict the temporary record from the
    *          active Hibernate session.
    * 
    * @see     #pushTempContext()
    */
   void popTempContext()
   throws PersistenceException
   {
      setTempRecord(null);
      tempRecords.remove(tempRecords.size() - 1);
   }
   
   /**
    * Set a temporary record into the buffer for purposes such as client-side
    * sorting.  Setting a temporary record has no impact on locking or
    * validation.  It is essentially a read-only record which temporarily
    * backs the buffer proxy, so that non-updating client code can run using
    * the buffer.
    * <p>
    * The default value of the temporary record is <code>null</code>.  It is
    * only set to a non-<code>null</code> reference in order to do some work
    * with the buffer.  When the work is complete, it is critical that the
    * temporary record be set back to <code>null</code>.
    * <p>
    * Temporary records increment the context-wide DMO use count while they
    * are active and decrement it when they are released.  This may result
    * in a temporary record being evicted from the persistence session when
    * it is replaced.
    *
    * @param   tempRecord
    *          Record which will temporarily back the buffer.
    *
    * @throws  PersistenceException
    *          if there is an error evicting a temporary record from the
    *          persistence session after it is used.
    */
   void setTempRecord(Persistable tempRecord)
   throws PersistenceException
   {
      Persistable old = getTempRecord();
      
      if (old != null)
      {
         decrementDMOUseCount(old);
      }

      tempRecords.set(tempRecords.size() - 1, tempRecord);
      
      if (tempRecord != null)
      {
         incrementDMOUseCount(tempRecord);
      }
   }
   
   /**
    * Called by the <code>BufferManager</code> when a new block scope has
    * been entered.  Called for all blocks.
    *
    * @param   blockDepth
    *          1-based depth of the block being entered, relative to the
    *          outermost block.
    */
   void enterBlock(int blockDepth)
   {
      this.blockDepth = blockDepth;
      
      if (TransactionManager.isTransaction())
      {
         pushPinnedLockScope();
      }
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() + ":  entered block " + blockDepth);
      }
   }
   
   /**
    * Called by the <code>BufferManager</code> when the current block scope
    * is about to end.  Called for all blocks.
    *
    * @param   blockDepth
    *          1-based depth of the block being exited, relative to the
    *          outermost block.
    */
   void exitBlock(int blockDepth)
   {
      if (LOG.isDebugEnabled())
      {
         LOG.debug(describe() + ":  exited block " + blockDepth);
      }
      
      if (TransactionManager.isTransaction())
      {
         // We always pass false for transaction parameter, because by the
         // time this method is called, a commit or rollback for the full
         // transaction will already have been processed;  processing it
         // again here would be redundant.
         popPinnedLockScope(false, false);
      }
      
      this.blockDepth = blockDepth - 1;
   }
   
   /**
    * Get the combined DMO and Buffer grouping interface associated with this
    * buffer.
    * 
    * @return  DMO and Buffer grouping interface.
    */
   Class<?> getDMOBufInterface()
   {
      return dmoBufInterface;
   }
   
   /**
    * Get the DMO interface associated with this buffer.
    *
    * @return  DMO interface.
    */
   Class<?> getDMOInterface()
   {
      return dmoInterface;
   }
   
   /**
    * Report the DMO implementation class associated with this buffer.
    *
    * @return  DMO implementation class.
    *
    * @throws  PersistenceException
    *          if no implementation class is found for the given database and
    *          interface.
    */
   Class<?> getDMOImplementationClass()
   {
      checkActive();
      
      return dmoClass;
   }
   
   /**
    * Get the unqualified name of the DMO's interface.
    *
    * @return  DMO interface short name.
    */
   String getDMOName()
   {
      return Utils.getUnqualifiedName(dmoInterface);
   }
   
   /**
    * Get the unqualified name of the DMO implementation class.
    *
    * @return  DMO short class name. 
    */
   String getDMOImplementationName()
   {
      return Utils.getUnqualifiedName(getDMOImplementationClass());
   }
   
   /**
    * Get the variable name associated with this buffer instance.
    *
    * @return  Variable name associated with the buffer.
    */
   String getDMOAlias()
   {
      return variable;
   }
   
   /**
    * Get the persistence service object associated with this buffer.  This
    * object is the buffer's interface to the physical database which backs
    * it.
    *
    * @return  Persistence service object for this buffer's database.
    */
   Persistence getPersistence()
   {
      checkActive();
      
      return persistence;
   }
   
   /**
    * Get the physical database name associated with this buffer.
    *
    * @return  Physical database name.
    */
   Database getDatabase()
   {
      checkActive();
      
      return persistence.getDatabase();
   }
   
   /**
    * Convenience method to get the <code>P2JDialect</code> in use for this
    * record buffer.
    * 
    * @return  Database dialect.
    */
   P2JDialect getDialect()
   {
      checkActive();
      
      return (P2JDialect) persistence.getDialect();
   }
   
   /**
    * Get the schema name of the database associated with this buffer.
    *
    * @return  Database schema name.
    */
   String getSchema()
   {
      return DatabaseManager.getSchema(getDatabase());
   }
   
   /**
    * Get the logical database name associated with this buffer.
    *
    * @return  Logical database name.
    */
   String getLogicalDatabase()
   {
      checkActive();
      
      return logicalDatabase;
   }
   
   /**
    * Get the name of the table associated with this buffer.
    *
    * @return  Table name.
    */
   String getTable()
   {
      checkActive();
      
      return table;
   }
   
   /**
    * Get the dirty share context used by this buffer.
    * 
    * @return  Dirty share context for this buffer's database.
    */
   DirtyShareContext getDirtyContext()
   {
      checkActive();
      
      return dirtyContext;
   }
   
   /**
    * Enter into a safe mode where current record is <code>null</code>, but
    * calls to DMO getter methods return unknown value instead of raising an
    * error condition.  This is intended to support outer join.
    */
   void setUnknownMode()
   {
      unknownMode = true;
   }
   
   /**
    * Get the record ID of the current buffer's current record, if any.  P2J
    * implementation of the RECID function.
    * <p>
    * This method will return a 32-bit recid which represents the internal
    * row ID of the given record.  If the actual row ID (which is implemented
    * as a wider integer) cannot fit due to overflow, an error condition will
    * be raised.
    *
    * @return  Record ID of the underlying data record. If no record currently
    *          backs the buffer associated with <code>dmo</code>, a recid
    *          initialized as unknown value is returned.
    * 
    * @throws  ErrorConditionException
    *          if the actual record ID is too wide to fit within a 32-bit
    *          integer.
    */
   public recid recordID()
   {
      if (isAvailable())
      {
         Persistable record = getCurrentRecord();
         Long id = record.getId();
         long value = id.longValue();
         if (value < Integer.MIN_VALUE || value > Integer.MAX_VALUE)
         {
            throw new ErrorConditionException(
               "Row id of " +
               value +
               " is too wide to fit within a 32-bit record ID");
         }
         
         return (new recid(value));
      }
      else
      {
         return (new recid());
      }
   }
   
   /**
    * Get the row ID of the current buffer's current record, if any.  P2J
    * implementation of the ROWID function.
    *
    * @return  Row ID of the underlying data record.  If no record currently
    *          backs the buffer associated with <code>dmo</code>, a
    *          <code>rowid</code> initialized as unknown value is returned.
    */
   public rowid rowID()
   {
      if (isAvailable())
      {
         Persistable record = getCurrentRecord();
         Long id = record.getId();
         
         return (new rowid(id));
      }
      else
      {
         return (new rowid());
      }
   }
   
   /**
    * Instantiate a data record to back the specified DMO and to be added to
    * to the database upon transaction commit.  The record's data members are
    * initialized to their default values.  The created record becomes the
    * buffer's current record.  A share lock is automatically applied to the
    * new record.  Foreign keys, if any, are synchronized using the default
    * values of legacy key properties.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   void create()
   {
      initialize();
      
      Serializable id = null;
      RecordIdentifier ident = null;
      
      try
      {
         if (isTransient())
         {
            try
            {
               flush();
            }
            catch (ValidationException exc) 
            {
               // if the current referenced record does not pass validation,
               // it means it was never flushed when it was created in the 
               // first place. so, before allowing the validation exception
               // to be processed, we must release this not-valid record
               setCurrentRecord(null, true, false, false);
               
               throw exc;
            }
         }
         
         // Create new record with default properties.
         Persistable record = instantiateDMO();
         
         // Generate and assign a unique key.
         id = nextPrimaryKey();
         record.setId((Long) id);
         
         // Acquire a default share lock on the new record.
         ident = new RecordIdentifier(table, id);
         lockContext.lock(ident, LockType.SHARE, this);
         lockContext.registerBuffer(ident, this);
         
         // Set the new record into the buffer.
         setCurrentRecord(record, false, true, false);
         
         // Enable this create to be rolled back.
         setReversibleCreate();
         
         // execute create triggers
         if (triggerDelegate != null)
         {
            triggerDelegate.executeCreateTriggers(getDMOProxy(), false);
         }
         
         // Synchronize all foreign keys based on the default legacy property
         // values.
         if (dirtyLegacyKeyProps != null)
         {
            dirtyLegacyKeyProps.addAll(legacyForeignKeys);
            synchForeignRelations(false);
         }
         
         // Remember that this record has not been persisted yet.
         createScope = blockDepth;
         
         if (!registeredForFlush)
         {
            flushQueue.addListener(this);
            registeredForFlush = true;
         }
         
         if (dirtyLegacyKeyProps != null)
         {
            dirtyLegacyKeyProps.clear();
         }
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug(describe() + ":  new record created");
         }
         if (LOG.isTraceEnabled())
         {
            LOG.trace(toString());
         }
         
         // save the current record into the snapshot
         snapshot = (Persistable) currentRecord.deepCopy();
         
         // Null out ident to indicate we completed successfully.
         ident = null;
      }
      catch (ValidationException exc)
      {
         reportValidationException(exc);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException("Error creating record", exc);
      }
      catch (InstantiationException exc)
      {
         throw new RuntimeException("Error creating record", exc);
      }
      catch (InvocationTargetException exc)
      {
         DBUtils.handleException(getDatabase(), exc);
         
         throw new RuntimeException("Error creating record", exc);
      }
      finally
      {
         // If ident is not null, we are not exiting normally;  have to unlock
         // record to avoid leak.
         if (ident != null)
         {
            try
            {
               lockContext.lock(ident, LockType.NONE, this);
            }
            catch (LockUnavailableException exc)
            {
            }
         }
      }
   }
   
   /**
    * Validate the record buffer associated with <code>dmo</code>.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   void validate()
   {
      if (!isAvailable())
      {
         return;
      }
      
      try
      {
         Persistable record = getCurrentRecord();
         validationHelper.validateFull(record, true);
      }
      catch (ValidationException exc)
      {
         reportValidationException(exc);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Indicate whether a backing data record currently is available. This
    * will only be the case if one has been created with {@link #create} or
    * retrieved in a query.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   boolean isAvailable()
   {
      return (currentRecord != null);
   }

   /**
    * Checks if the record in the buffer has changed in database since the
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   boolean isRecordChanged()
   {
      return recordChanged;
   }

   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   boolean wasLocked()
   {
      return locked;
   }
   
   /**
    * Set the flag indicating whether the last record retrieval attempt
    * failed because another user had the target record locked.
    *
    * @param   locked
    *          New value of the flag.
    */
   void setLocked(boolean locked)
   {
      this.locked = locked;
   }
   
   /**
    * Indicate whether the buffer currently contains a record, and whether
    * that record was newly created (as opposed to having been read from the
    * database).
    * 
    * @return  <code>true</code> if record is newly created, else
    *          <code>false</code>.
    */
   boolean isNewlyCreated()
   {
      return (currentRecord != null && newlyCreated);
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   boolean wasAmbiguous()
   {
      return ambiguous;
   }
   
   /**
    * Set the flag indicating whether the last record retrieval attempt
    * failed because the query criteria was ambiguous.
    *
    * @param   ambiguous
    *          New value of the flag.
    */
   void setAmbiguous(boolean ambiguous)
   {
      this.ambiguous = ambiguous;
   }
   
   /**
    * Disable the buffer's default behavior of releasing its current record
    * when the transaction manager invokes its {@link #iterate()} method.
    * This is necessary for the outer buffers in a compound query.
    */
   void disableReleaseOnIterate()
   {
      iterateReleaseOverrideScope = blockDepth;
   }
   
   /**
    * Indicate whether the record currently in the buffer represents a copy of
    * a DMO from the dirty database.  This happens in the case where another
    * context inserts a new record, but has not committed it yet, and this
    * context finds that record via dirty checking.
    * 
    * @return  <code>true</code> if current record is a dirty copy, else
    *          <code>false</code>.
    */
   boolean isDirtyCopy()
   {
      return dirtyCopy;
   }
   
   /**
    * Indicate whether the buffer currently contains a record, and whether
    * that record is transient.  A record is transient if it has been newly
    * created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if record is new, else <code>false</code>.
    */
   boolean isTransient()
   {
      return (currentRecord != null && createScope >= 0);
   }
   
   /**
    * Mark the buffer's current record as no longer transient by resetting it
    * create scope to its default value.
    */
   void markPersisted()
   {
      createScope = -1;
   }
   
   /**
    * Get the DMO proxy which references this buffer instance.
    *
    * @return  DMO proxy object.
    */
   BufferReference getDMOProxy()
   {
      return dmoProxy;
   }
   
   /**
    * Add all property names in the given collection to the set of legacy key
    * properties being monitored for foreign association synchronization
    * purposes.  Any change made to one or more of these properties should
    * trigger a foreign key synchronization attempt.
    *
    * @param   properties
    *          A collection of DMO property names.
    *
    * @see     AssociationSyncher
    * @see     LocalSyncher
    * @see     InverseSyncher
    */
   void addLegacyForeignKeys(Collection<String> properties)
   {
      if (legacyForeignKeys != null)
      {
         legacyForeignKeys.addAll(properties);
      }
   }
   
   /**
    * Get a map which defines the intersection of dirty properties and those
    * properties which participate in any indexes in the backing table.
    * 
    * @return  A map whose keys are property names and whose values represent
    *          dirty updates.  The map may be empty, but will not be
    *          <code>null</code>.
    */
   Map<String, Object> getDirtyProperties()
   {
      if (dirtyProps.isEmpty())
      {
         return Collections.emptyMap();
      }
      
      Map<String, Object> map = new HashMap<String, Object>(dirtyProps);
      map.keySet().retainAll(indexedProperties);
      
      return map;
   }
   
   /**
    * Indicate whether the given property participates in any index.
    * 
    * @param   property
    *          Property name.
    * 
    * @return  <code>true</code> if <code>property</code> participates in an
    *          index, else <code>false</code>.
    */
   boolean isPropertyIndexed(String property)
   {
      checkActive();
      
      return indexedProperties.contains(property);
   }
   
   /**
    * Get the buffer type hash key associated with this buffer.
    * 
    * @return  <code>BufferType</code> key associated with this buffer, or
    *          <code>null</code> if the buffer has not yet been initialized. 
    */
   BufferType getBufferType()
   {
      if (!isActive())
      {
         return null;
      }
      
      return bufferType;
   }
   
   /**
    * Attempt to flush all buffers which are backed by the same table in the
    * same database as this buffer.  This will call {@link #flush()} on all
    * such buffers (including this one) which have registered for such events
    * with the {@link BufferFlushQueue}.  Calls to each buffer's
    * <code>flush()</code> method are made in the order the buffers were
    * registered with the queue.
    * 
    * @throws  ValidationException
    *          if validation on any buffer fails.
    * @throws  ErrorConditionException
    *          if there is an error accessing the database while persisting
    *          a new record in any buffer.
    */
   void flushAll()
   throws ValidationException
   {
      flushQueue.flushAll(this);
   }
   
   /**
    * Validate the current record and attach it to the current persistence
    * session if it has been newly created and not yet saved.  Also try to
    * synchronize any foreign key related DMOs which may be out of sync.
    *
    * @throws  ValidationException
    *          if validation on the buffer fails.
    * @throws  ErrorConditionException
    *          if there is an error accessing the database while persisting
    *          the record.
    */
   void flush()
   throws ValidationException
   {
      if (!isTransient())
      {
         return;
      }
      
      boolean commit = false;
      try
      {
         // Validate remaining indexes/properties in the record.
         validationHelper.validateRemaining();
         
         // Perform inverse synchronization
         synchForeignRelations(true);
         
         // Persist the record.
         commit = persistence.beginTransaction();
         persistence.save(currentRecord, currentRecord.getId());
         markPersisted();
         
         // Share uncommitted change to the record, if not previously done.
         if (validationHelper.wasShared())
         {
            dirtyContext.insert(this, currentRecord, null, null, true);
         }
         
         // Enable this create to be rolled back, for undoable buffers.
         if (isUndoable())
         {
            reversibleCreate.markPersisted();
         }
         
         reportChange(currentRecord, true, false);
         
         if (LOG.isDebugEnabled())
         {
            String desc = describe();
            LOG.debug(desc + ":  flushed buffer");
            if (LOG.isTraceEnabled())
            {
               LOG.trace(desc + ":  " + toString());
               LOG.trace(desc + ":  location:", new Throwable());
            }
         }
      }
      catch (ValidationException exc)
      {
         commit = false;
         
         throw exc;
      }
      catch (PersistenceException exc)
      {
         if (commit)
         {
            commit = false;
            try
            {
               persistence.rollback();
            }
            catch (PersistenceException exc2)
            {
            }
         }
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         if (createScope < 0)
         {
            reversibleCreate = null;
         }
         
         if (commit)
         {
            try
            {
               persistence.commit();
            }
            catch (PersistenceException exc)
            {
               ErrorManager.recordOrThrowError(exc);
            }
         }
      }
   }
   
   /**
    * Get current state of 'pending rollback processed' flag.
    * 
    * @return <code>true</code> if pending rollback was processed. 
    */
   boolean isPendingRollbackProcessed()
   {
      return pendingRollbackProcessed;
   }

   /**
    * Reset 'pending rollback processed' flag. 
    */
   void resetPendingRollbackProcessed()
   {
      pendingRollbackProcessed = false;
   }

   /**
    * Worker method which aggregates the values in the table backing this buffer.
    *
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*") for COUNT.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    *
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   private void aggregate(String aggregate, String column, NumberType num, String where, Object...args)
   {
      disableSilentError();

      initialize();

      Persistence persistence = getPersistence();
      String dmoImplName = getDMOImplementationName();
      org.hibernate.type.Type[] types = (args == null) ? null : DBUtils.makeTypeArray(args);

      StringBuilder buf = new StringBuilder();
      buf.append("select ");
      buf.append(aggregate);
      buf.append("(");
      buf.append(column);
      buf.append(") from ");
      buf.append(dmoImplName);
      if (where != null && where.length() > 0)
      {
         buf.append(" where ");
         buf.append(where);
      }
      try
      {
         List list = persistence.list(buf.toString(), args, types, 1, 0, true);
         final Undoable result;
         if (list == null)
         {
            result = null;
         }
         else
         {
            Object value = list.get(0);
            if (value instanceof Number)
            {
               result = new decimal((Number)value);
            }
            else if (value instanceof Undoable)
            {
               result = (NumberType)value;
            }
            else
            {
               throw new ErrorConditionException("Wrong result type");
            }
         }
         num.assign(result);
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
      }
   }

   /**
    * Verify that the record buffer has been initialized.  If not, raise an
    * exception.
    * 
    * @throws  IllegalStateException
    *          if the buffer has not been initialized.
    */
   private void checkActive()
   {
      if (!isActive())
      {
         throw new IllegalStateException("Buffer " +
                                         variable +
                                         " is not active");
      }
   }
   
   /**
    * Merge the reversibles (or re-roll actions, for no-undo temp-tables) from
    * the current scope with the list from the parent scope.
    */
   private void mergeReversibles()
   {
      boolean trace = LOG.isTraceEnabled();
      boolean undoable = isUndoable();
      
      // Move record-level, reversible actions to parent scope.  These entries
      // will be added last to the list.
      Set<Reversible> r0 = bufferManager.getReversibles(this, 0, false);
      Map<Serializable, List<Reversible>> m0 = 
         bufferManager.getReversiblesMap(this, 0, false);

      // Merge property-level, reversible updates into parent scope.
      Map<Serializable, Reversible> u0 = null, u1 = null;
      u0 = bufferManager.getUndoables(this, 0, false);
      if (u0 != null && !u0.isEmpty())
      {
         u1 = bufferManager.getUndoables(this, 1, true);
      }
      
      if (trace)
      {
         String s = (r0 != null ? r0.toString() : "(none)");
         LOG.trace(describe() +
                   ":  moving reversibles into parent scope:  " +
                   s);
      }
      if (r0 != null && !r0.isEmpty())
      {
         Set<Reversible> r1 = bufferManager.getReversibles(this, 1, true);
        
         Iterator<Reversible> iter = r0.iterator();
         while (iter.hasNext())
         {
            Reversible next = iter.next();
            boolean add = true;

            if (next instanceof ReversibleUpdate)
            {
               Serializable key = next.getId();
               boolean uniqueConstraints = 
                  ((ReversibleUpdate) next).uniqueConstraints;
               
               if (!uniqueConstraints && 
                   u0 != null && u0.containsKey(key) &&
                   u1 != null && u1.containsKey(key))
               {
                  // must skip all ReversibleUpdate's which affect non-unique
                  // properties; only one of those reversibles is kept, the 
                  // one in the "upper" (r1) set; this optimisez the rollback
                  // of non-unique fields, as only the last change will be 
                  // kept.
                  add = false;
               }
            }
            
            if (add)
            {
               // if previous scope didn't record any info related to this 
               // reversible or this reversible affects properties which are 
               // part of a unique clause, we link it to the set. 
               r1.add(next);
            }

            iter.remove();
         }
      }
      
      if (m0 != null && !m0.isEmpty())
      {
         // as a second map of reversibles is kept, we must merge it too.
         Map<Serializable, List<Reversible>> m1 = 
            bufferManager.getReversiblesMap(this, 1, true);
         
         Iterator<Serializable> itrList = m0.keySet().iterator();
         while (itrList.hasNext())
         {
            Serializable key = itrList.next();
            List<Reversible> list0 = m0.get(key);
            List<Reversible> list1 = m1.get(key);
            
            if (list1 == null)
            {
               // upper level doesn't know about this DMO, add the list 
               // to the map
               m1.put(key, list0);
            }
            else
            {
               // the two lists must be merged on the upper level
               list1.addAll(list0);
            }
            
            itrList.remove();
         }
      }
      
      if (trace)
      {
         String s = (u0 != null ? u0.keySet().toString() : "(none)");
         LOG.trace(describe() +
                   ":  merging undoable updates into parent scope:  " +
                   s);
      }
      if (u0 != null && u1 != null)
      {
         Iterator<Map.Entry<Serializable, Reversible>> iter =
            u0.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<Serializable, Reversible> next = iter.next();
            Serializable key = next.getKey();
            ReversibleUpdate src = (ReversibleUpdate) next.getValue();
            ReversibleUpdate dst = (ReversibleUpdate) u1.get(key);

            if (dst != null)
            {
               if (undoable)
               {
                  dst.nonDestructiveMerge(src);
               }
               else
               {
                  // for no-undo temp-tables, this implementation is used
                  // to re-roll the changes
                  dst.rollbackMap.putAll(src.rollbackMap);
               }
            }
            else
            {
               u1.put(key, src);
            }

            iter.remove();
         }
      }
   }
   
   /**
    * Revert all dirty properties of the current record to their previous
    * state and clear the {@link #dirtyProps} map.
    * <p>
    * If there is no current record or no pending changes to the current
    * record, the method returns immediately.
    * 
    * @throws  PersistenceException
    *          if there is an error using reflection to set DMO properties.
    */
   private void undoDirty()
   throws PersistenceException
   {
      if (currentRecord == null || dirtyProps.isEmpty())
      {
         return;
      }
      
      try
      {
         Iterator<Map.Entry<String, Object>> iter =
            dirtyProps.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<String, Object> next = iter.next();
            String property = next.getKey();
            Method setter = getSetterMap().get(property);
            Object value = next.getValue();
            if (value instanceof Map)
            {
               Iterator<Map.Entry<Integer, Object[]>> iter2 =
                  ((Map<Integer, Object[]>) value).entrySet().iterator();
               while (iter2.hasNext())
               {
                  Map.Entry<Integer, Object[]> entry = iter2.next();
                  Integer index = entry.getKey();
                  Object[] diffs = entry.getValue();
                  Object oldValue = diffs[0];
                  setter.invoke(currentRecord, index, oldValue);
               }
            }
            else
            {
               Object[] diffs = (Object[]) value;
               Object oldValue = diffs[0];
               setter.invoke(currentRecord, oldValue);
            }
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException(
            "Error rolling back dirty changes",
            exc);
      }
      finally
      {
         dirtyProps.clear();
      }
   }
   
   /**
    * Process one or more DMO property changes by creating or updating a
    * <code>ReversibleUpdate</code> object and reporting the change to the
    * {@link ChangeBroker} for further distribution to registered listeners.
    *
    * @throws  PersistenceException
    *          if a problem occurs reporting the change to registered
    *          listeners.
    */
   @SuppressWarnings("unchecked")
   private void processChange()
   throws PersistenceException
   {
      if (currentRecord != null && unreportedChanges.isEmpty())
      {
         return;
      }
      
      boolean undoable = isUndoable();

      // If we are in a transaction, remember this property change so it can
      // be undone at rollback if necessary.
      if (TransactionManager.isTransaction())
      {
         Serializable id = currentRecord.getId();
         
         // Try to find a ReversibleUpdate for this DMO in the current scope.
         Map<Serializable, Reversible> undoables = null;
         undoables = bufferManager.getUndoables(this, 0, true);
         
         ReversibleUpdate ru = (ReversibleUpdate) undoables.get(id);
         boolean uniqueConstraints = false;         
         
         if (ru != null)
         {
            uniqueConstraints = 
               anyUniqueConstraint(unreportedChanges.keySet());
         }
         
         // Create a new one if not found or the update affects fields which
         // are part in a unique clause. For the unique case, we must ensure 
         // the proper order of actions is kept, no matter how many times an
         // update action is performed for a certain DMO.
         if (ru == null || uniqueConstraints)
         {
            boolean set = ru == null && !uniqueConstraints;
            
            ru = new ReversibleUpdate();
            ru.uniqueConstraints = uniqueConstraints;

            if (set)
            {
               // this map will contain only changes to properties which are
               // not part of a unique constraint; they are allowed to be
               // merged
               undoables.put(id, ru);
            }

            setReversibleEntry(ru);
         }
         
         // Add an undo entry for each change.
         Iterator<Map.Entry<String, Object>> iter =
            unreportedChanges.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<String, Object> next = iter.next();
            String property = next.getKey();
            Object value = next.getValue();
            if (value instanceof Map)
            {
               Iterator<Map.Entry<Integer, Object[]>> iter2 =
                  ((Map<Integer, Object[]>) value).entrySet().iterator();
               while (iter2.hasNext())
               {
                  Map.Entry<Integer, Object[]> entry = iter2.next();
                  Integer index = entry.getKey();
                  Object[] diffs = entry.getValue();
                  Object oldValue = diffs[0];
                  Object newValue = diffs[1];
                  ru.addUndoEntry(property, 
                                  index, 
                                  undoable ? oldValue : newValue);
               }
            }
            else
            {
               Object[] diffs = (Object[]) value;
               Object oldValue = diffs[0];
               Object newValue = diffs[1];
               ru.addUndoEntry(property, 
                               null, 
                               undoable ? oldValue : newValue);
            }
         }
      }
      
      reportChange(currentRecord, false, false);
   }

   /**
    * Check if any of the given properties is part of a unique constraint.
    * 
    * @param    properties
    *           A set of properties to be checked if any is part of an unique
    *           constraint.
    * @return   <code>true</code> if any of the specified property is part
    *           of an unique constraint.
    */
   private boolean anyUniqueConstraint(Set<String> properties)
   {
      String schema = getSchema();
      // the check is destructive, so a copy is used
      Set<String> copy = new HashSet<String>(properties);

      Iterator<Set<String>> iter = 
         DMOIndex.uniqueConstraints(schema, dmoInterface);
      while (iter.hasNext())
      {
         Set<String> set = iter.next();
         
         if (copy.removeAll(set))
         {
            // one of the passed properties matched, so an unique constraint
            // is involved
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Get the currently active temporary record.
    * 
    * @return  Current temporary record set into this buffer.
    * 
    * @see     #setTempRecord(Persistable)
    */
   private Persistable getTempRecord()
   {
      if (tempRecords == null || tempRecords.isEmpty())
      {
         return null;
      }
      
      return tempRecords.get(tempRecords.size() - 1);
   }
   
   /**
    * Load the specified record, if any, from the database.  If
    * <code>lockType</code> is higher than the current lock type, it is
    * upgraded immediately.  If lower, it is downgraded immediately only if
    * outside of a transaction.  If inside a transaction, the lock is marked
    * for downgrade to the specified type upon transaction end.
    * 
    * @param   target
    *          DMO to be reloaded.  If <code>null</code>, an exception will be
    *          raised.
    * @param   lockType
    *          Requested lock type.
    * @param   undo
    *          <code>true</code> if currently processing an undo or
    *          <code>false</code> if currently processing a reload from database
    *
    * @throws  ErrorConditionException
    *          if there is no record loaded into the buffer and the error
    *          manager is not in silent mode.
    */
   private void load(Persistable target, LockType lockType, boolean undo)
   {
      if (target == currentRecord && undo && isTemporary())
      {
         // A temporary record cannot have been changed by another session,
         // so there is no need to reload it in an undo situation if it is
         // already loaded as the current record.
         return;
      }
      
      LockType pinned = null;
      Serializable id = null;
      
      try
      {
         if (target == null)
         {
            errorNotOnFile();
            
            return;
         }
         
         id = target.getId();
         
         pinned = getPinnedLockType(id);
         if (!lockType.equals(pinned))
         {
            setPinnedLockType(id, lockType);
         }
         
         Persistable dmo = null;
         boolean dirtyCopy = false;
         
         if (LockType.NONE.equals(lockType))
         {
            dmo = dirtyContext.getDirtyDMO(getEntityName(), id);
         }
         
         if (dmo != null)
         {
            dirtyCopy = true;
         }
         else if (target == currentRecord && isTransient())
         {
            dmo = target;
         }
         else
         {
            // This method is called from 2 places:
            // 1. RecordBuffer.LightweightUndoable.load() - undo = true
            // 2. RecordBuffer.reload()                   - undo = false
            // in the second case, we must force the dmo is up-to-date (last param must be true)
            dmo = persistence.load(this, id, lockType, true, !undo);
         }
         
         setCurrentRecord(dmo, undo, newlyCreated, dirtyCopy);
      }
      catch (PersistenceException exc)
      {
         if (id != null && pinned != null)
         {
            setPinnedLockType(id, pinned);
         }
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Trigger the initialization of persistent collections which represent
    * extent fields by invoking a getter method for each such collection.
    * This will force Hibernate to load the collection if it has not already
    * been done.
    */
   private void initializeLazyExtents()
   {
      if (currentRecord == null || getSizerMap().isEmpty())
      {
         return;
      }
      
      try
      {
         if (lazyExtentTriggers == null)
         {
            lazyExtentTriggers =
               PropertyHelper.lazyExtentTriggersForInterface(dmoInterface,
                                                             currentRecord);
         }
         
         Iterator<Method> iter = lazyExtentTriggers.iterator();
         while (iter.hasNext())
         {
            // Execute the extent field getter method with a safe index of 0
            // (we know any such property must have at minimum 2 entries).
            Method getter = iter.next();
            getter.invoke(currentRecord, 0);
         }
      }
      catch (Exception exc)
      {
         // RuntimeException is acceptable here, because this would indicate a
         // programming error, so it should not happen in practice.
         throw new RuntimeException("Error initializing lazy extents", exc);
      }
   }
   
   /**
    * Attempt to synchronize the foreign keys of any records linked to this
    * buffer's current record, based upon changes made to relevant properties
    * in the current record.
    *
    * @param inverse
    *        If <code>true</code> then inverse synchronization is performed.
    *        Otherwise only synchronization triggered by changes to the DMO at
    *        the "local" end of the association is performed.
    *
    * @see  InverseSyncher
    * @see  LocalSyncher
    */
   private void synchForeignRelations(boolean inverse)
   {
      try
      {
         if (assocSynchers == null || dirtyLegacyKeyProps.isEmpty())
         {
            return;
         }
         
         Iterator<AssociationSyncher> iter = assocSynchers.iterator();
         while (iter.hasNext())
         {
            AssociationSyncher next = iter.next();
            if (next instanceof InverseSyncher
                ? inverse && isNewlyCreated()
                : !inverse)
            {
               next.attemptSynch(dirtyLegacyKeyProps);
            }
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Set the DMO proxy which references this buffer instance.
    *
    * @param   dmoProxy
    *          DMO proxy object.
    */
   private void setDMOProxy(BufferReference dmoProxy)
   {
      this.dmoProxy = dmoProxy;
   }
   
   /**
    * Set the DMO implementation instance which backs this buffer, or release
    * the record currently stored in the buffer.
    *
    * @param   currentRecord
    *          DMO instance which represents the record currently backing
    *          this buffer.  If <code>null</code>, the record (and possibly
    *          its corresponding lock) is released.
    * @param   undo
    *          <code>true</code> if currently processing an undo, else
    *          <code>false</code>.
    * @param   newlyCreated
    *          <code>true</code> if the record being stored in the buffer is
    *          newly created.
    * @param   dirtyCopy
    *          Indicates whether <code>currentRecord</code> represents a copy
    *          of a DMO found in the dirty database.
    *
    * @throws  ErrorConditionException
    *          if there is an error transitioning a lock when releasing the
    *          current record;  if there is an error validating or saving a
    *          transient record during flush.
    */
   private void setCurrentRecord(Persistable currentRecord,
                                 boolean undo,
                                 boolean newlyCreated,
                                 boolean dirtyCopy)
   {
      boolean sameRecord = areDMOsSame(this.currentRecord, currentRecord);

      if (!sameRecord)
      {
         resetPendingRollbackProcessed();
      }
      
      if (currentRecord != null)
      {
         sameRecord = sameRecord && (this.dirtyCopy == dirtyCopy);
         
         if (sameRecord)
         {
            if (!undo)
            {
               this.newlyCreated = newlyCreated;
               snapshot = null;
            }

            bufferManager.notifyRecordWasLoaded(this);
            return;
         }
      }
      
      this.dirtyCopy = dirtyCopy;
      
      try
      {
         // Certain processing should not take place if we are resetting the
         // current record due to an undo.
         if (!undo)
         {
            // Flush any previously created record to the database before we
            // release/replace it.
            flush();
            
            // If a newly created record is being replaced via a query, do
            // not carry the createScope setting forward.  This only matters
            // if the flush above was not clean.
            if (!newlyCreated && !sameRecord)
            {
               markPersisted();
            }
         }
      }
      catch (ValidationException exc)
      {
         reportValidationException(exc);
      }
      finally
      {
         if (!sameRecord)
         {
            validationHelper.reset(currentRecord, newlyCreated);
            dirtyProps.clear();
         }
      }
      
      if (this.currentRecord != null)
      {
         // Store placeholder record if releasing current record and we
         // haven't done so previously.
         if (snapshot == null && (currentRecord == null || newlyCreated))
         {
            snapshot = this.currentRecord;
         }
         
         // Transition lock when releasing current record.
         if (currentRecord == null || !sameRecord)
         {
            try
            {
               // Try to release or downgrade the lock, if any.
               Serializable id = this.currentRecord.getId();
               LockType pinned = getPinnedLockType(id);
               if (TransactionManager.isTransaction())
               {
                  if (pinned.isExclusive())
                  {
                     // Downgrade pinned lock type.  This won't actually
                     // downgrade the lock, it just lets the lock context
                     // know that this buffer is done with its exclusive
                     // lock on this record.
                     setPinnedLockType(id, LockType.SHARE);
                  }
               }
               else
               {
                  // Try to release the lock.
                  RecordIdentifier ident = new RecordIdentifier(table, id);
                  lockContext.lock(ident, LockType.NONE, this);
               }
            }
            catch (LockUnavailableException exc)
            {
               // Safe to ignore, will not be thrown by lock release.
            }
            finally
            {
               try
               {
                  decrementDMOUseCount(this.currentRecord);
               }
               catch (PersistenceException exc)
               {
                  ErrorManager.recordOrThrowError(exc);
               }
               
               // make sure the current record is updated, if any exceptions were thrown
               this.currentRecord = currentRecord;
               // invalidate CURRENT-CHANGED flag. it will be re-set at the end of 
               // armCurrentChanged/updateCurrentChanged bracketing if the case
               this.recordChanged = false;

               bufferManager.notifyRecordWasLoaded(this);
            }
         }
      }
      
      // Update current record.
      this.currentRecord = currentRecord;
      this.recordChanged = false; // invalidate CURRENT-CHANGED flag (see above)
      this.newlyCreated = newlyCreated;
      
      bufferManager.notifyRecordWasLoaded(this);
      
      if (currentRecord != null && !sameRecord)
      {
         incrementDMOUseCount(currentRecord);
      }
      
      // If not null and not currently processing an undo, reset certain
      // state.
      if (currentRecord != null && !undo)
      {
         locked = false;
         unknownMode = false;
         
         // Creating a new record does not change our placeholder snapshot.
         if (!newlyCreated)
         {
            snapshot = null;
         }
         
         if (LOG.isTraceEnabled())
         {
            String desc = describe();
            LOG.trace(desc + ":  currentRecord = " + toString());
            LOG.trace(desc + ":  snapshot = " + toString(snapshot));
         }
      }
   }
   
   /**
    * Prepares for an immediate reload(). The current value of the record is saved temporarily
    * in order to be available to compare to the new value that will be obtained in reload().
    * This method is only called from within P2JQuery.current() implementation in order to
    * prepare the buffer to recognise if the field has changed externally. After a call to 
    * this method and a reload() of the buffer, updateCurrentChanged() must be called to ensure
    * the proper <code>recordChanged</code> settings and free of the temporary compare buffer.
    */
   public void armCurrentChanged()
   {
      this.shadowCopyRecord = (this.currentRecord == null) ? null :
            (Persistable) this.currentRecord.deepCopy();
   }

   /**
    * Updates the <code>recordChanged</code> flag.
    * This must be called right after an <code>armCurrentChanged()</code>followed by a
    * <code>reload()</code>.
    * This checks the current value stored in <code>currentRecord</code> against the value
    * temporarily saved and updates the <code>recordChanged</code> accordingly. After that, the
    * temp data is freed.
    */
   public void updateCurrentChanged()
   {
      if (this.shadowCopyRecord == null)
      {
         // something is wrong
         System.err.println("armCurrentChanged() should be called before updateCurrentChanged()");
         this.recordChanged = false;
         return;
      }

      if (this.currentRecord == null)
      {
         // the record must have been deleted
         this.recordChanged = true;
         this.shadowCopyRecord = null;
         return;
      }

      this.recordChanged = false;
      final Map<String, Method> getters = getGetterMap();
      final Map<String, Method> sizers = getSizerMap();
      outer: for (String field : getters.keySet())
      {
         Method m = getters.get(field);
         // change detected by checking bit-by-bit of each field
         if (m.getParameterTypes().length != 0)
         {
            // this is a getter for a item of an EXTENT field
            try
            {
               Object n1 = sizers.get(field).invoke(this.currentRecord);
               Object n2 = sizers.get(field).invoke(this.shadowCopyRecord);
               int n = ((integer) n1).intValue();
               if (n != ((integer) n2).intValue())
               {
                  // the extent sizes should match
                  this.recordChanged = true;
                  break outer;
               }

               for (int k = 0; k < n; k++)
               {
                  Object f1 = m.invoke(this.currentRecord, k);
                  Object f2 = m.invoke(this.shadowCopyRecord, k);
                  if (!f1.equals(f2))
                  {
                     this.recordChanged = true;
                     // if one element of the extent field is different then the whole
                     // record is marked as changed
                     break outer;
                  }
               }
            }
            catch (Exception e)
            {
               // normally should not fail
               System.err.println("Failed to evaluate " + m.toGenericString());
               this.recordChanged = true;
               break outer;
            }
         }
         else
         {
            try
            {
               Object f1 = m.invoke(this.currentRecord);
               Object f2 = m.invoke(this.shadowCopyRecord);
               if (!f1.equals(f2))
               {
                  this.recordChanged = true;
                  break;
               }
            }
            catch (Exception e)
            {
               // normally should not fail
               System.err.println("Failed to evaluate " + m.toGenericString());
               this.recordChanged = true;
               break outer; // no need to check the other fields
            }
         }
      }

      this.shadowCopyRecord = null;
   }


   /**
    * Indicate whether a commit is pending for the most recent changes made
    * to the current record.
    *
    * @return  <code>true</code> if a commit is pending, else
    *          <code>false</code>.
    */
   private boolean isCommitPending()
   {
      return commitPending;
   }
   
   /**
    * Set the commit pending state.
    *
    * @param   commitPending
    *          <code>true</code> if we are currently in batch assign mode and
    *          the changes made while in this mode should be committed to the
    *          database immediately upon completing this mode;
    *          <code>false</code> if the commit pending state should be reset
    *          such that no commit should occur.
    */
   private void setCommitPending(boolean commitPending)
   {
      this.commitPending = commitPending;
   }
   
   /**
    * Create and set the current reversible action for the given record.
    * This variant produces a <code>ReversibleCreate</code> instance.
    * <p>
    * If this buffer is not undoable, this method is a no-op.
    */
   private void setReversibleCreate()
   {
      if (!isUndoable())
      {
         setReversibleEntry(new ReversibleDelete());
      }
      else
      {
         reversibleCreate = new ReversibleCreate();
         setReversibleEntry(reversibleCreate);
      }
   }
   
   /**
    * Create and set the current reversible action for the given record.
    * This variant produces a <code>ReversibleDelete</code> instance.
    * <p>
    * If this buffer is not undoable, this method is a no-op.
    */
   private void setReversibleDelete()
   {
      if (!isUndoable())
      {
         ReversibleCreate rev = new ReversibleCreate();
         // in case of undoable tables, the record must be marked as 
         // persisted to be able to delete it
         rev.markPersisted();
         setReversibleEntry(rev);
      }
      else
      {
         setReversibleEntry(new ReversibleDelete());         
      }
   }
   
   /**
    * Set the current reversible action which is associated with the current
    * record. If there are existing actions, this new action is added to 
    * the chain, so all actions will be available for rollback.
    *
    * @param   item
    *          New, "current" reversible action.
    */
   private void setReversibleEntry(Reversible item)
   {
      // If we are in a transaction, remember this record change so it can be
      // undone at rollback if necessary.
      if (TransactionManager.isTransaction())
      {
         Serializable key = item.getId();

         bufferManager.addReversible(this, key, item);

         if (!isUndoable() && LOG.isDebugEnabled())
         {
            LOG.debug(describe() + ":  adding reversible [" + item + "]");
         }
      }
   }
   
   /**
    * Record or throw a &quot;record not available&quot; error condition for
    * the current DMO type.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is suppressed.
    */
   private void errorNotAvailable()
   {
      String msg = "No " + getDMOAlias() + " record is available";
      
      if (LOG.isDebugEnabled())
      {
         PersistenceException exc = new PersistenceException(msg, 91);
         LOG.debug(msg, exc);
      }
      
      // execute "no record is available" triggers
      if (triggerDelegate != null)
      {
         triggerDelegate.executeNoRecordAvailableTriggers(getDMOProxy());
      }

      ErrorManager.recordOrThrowError(91, msg);
   }
   
   /**
    * Helper class for validation processing.  Three types of validation are
    * supported:
    * <ul>
    *   <li>full validation:  all fields and constraints in the given DMO
    *       which require processing are validated;
    *   <li>dirty validation:  only those fields and constraints which have
    *       been touched by DMO property updates are validated;
    *   <li>remaining validation:  all fields and constraints which have not
    *       otherwise been validated by dirty validation are validated.
    * </ul>
    * <p>
    * This class also ensures uncommitted updates to permanent tables are
    * shared across contexts via the dirty share manager.
    */
   protected class ValidationHelper
   {
      /** Map of unique index names to ordered lists of property names */
      private final Map<String, List<String>> uniqueIndexMap;
      
      /** Map of non-unique index names to ordered lists of property names */
      private final Map<String, List<String>> nonUniqueIndexMap;
      
      /** Object which validates the current record */
      private final DMOValidator validator;
      
      /** Unique indexes still requiring validation for a transient record */
      private Set<String> uncheckedUniqueIndexes = Collections.emptySet();
      
      /** Non-nullable props requiring validation for a transient record */
      private Set<String> uncheckedNonNullables = Collections.emptySet();
      
      /** Properties made dirty since current record was loaded */
      private Set<String> cumulativeDirtyProps = null;
      
      /** Has the current record been modified since it was loaded? */
      private boolean touched = false;
      
      /** Has the record's state change yet been shared across sessions? */
      private boolean shared = false;
      
      /**
       * Default constructor.
       */
      ValidationHelper()
      {
         validator = DMOValidator.getInstance(RecordBuffer.this);
         try
         {
            String entity = getEntityName();
            uniqueIndexMap = indexHelper.getIndexMapForEntity(entity, true);
            nonUniqueIndexMap = indexHelper.getIndexMapForEntity(entity,
                                                                 false);
         }
         catch (PersistenceException exc)
         {
            // Fatal error.
            throw new StopConditionException(exc);
         }
      }
      
      /**
       * Cross reference dirty DMO properties with indexes containing them,
       * creating a list of indexes which have been updated.
       * 
       * @param   entity
       *          DMO type name.
       * @param   unique
       *          If <code>true</code>, only unique indexes are checked.  If
       *          <code>false</code>, only non-unique indexes are checked.
       * @param   fullMatch
       *          If <code>true</code> every property in an index must be
       *          dirty for that index to be considered dirty.  If
       *          <code>false</code>, any dirty property in an index is enough
       *          to consider that index dirty.
       * 
       * @return  A list of names of indexes considered dirty, in ascending
       *          alphabetical order.
       */
      protected List<String> listDirtyIndexes(String entity,
                                              boolean unique,
                                              boolean fullMatch)
      {
         Set<String> dirtyKeys = dirtyProps.keySet();
         List<String> indexList = null;
         
         if (!dirtyKeys.isEmpty())
         {
            Map<String, List<String>> map = (unique
                                             ? uniqueIndexMap
                                             : nonUniqueIndexMap);
            Iterator<Map.Entry<String, List<String>>> iter =
               map.entrySet().iterator();
            while (iter.hasNext())
            {
               Map.Entry<String, List<String>> entry = iter.next();
               List<String> indexProps = entry.getValue();
               boolean hit = false;
               if (fullMatch)
               {
                  Iterator<String> propIter = indexProps.iterator();
                  while (propIter.hasNext())
                  {
                     String property = propIter.next();
                     if (dirtyKeys.contains(property))
                     {
                        hit = true;
                     }
                     else if (!cumulativeDirtyProps.contains(property))
                     {
                        hit = false;
                        
                        break;
                     }
                  }
               }
               else
               {
                  Iterator<String> propIter = indexProps.iterator();
                  while (!hit && propIter.hasNext())
                  {
                     String property = propIter.next();
                     if (dirtyKeys.contains(property))
                     {
                        hit = true;
                     }
                  }
               }
               
               if (hit)
               {
                  if (indexList == null)
                  {
                     indexList = new ArrayList<String>();
                  }
                  String name = entry.getKey();
                  indexList.add(name);
               }
            }
         }
         
         if (indexList == null)
         {
            indexList = Collections.emptyList();
         }
         else
         {
            Collections.sort(indexList);
         }
         
         return indexList;
      }
      
      /**
       * Reset the state variables which track whether the given record has
       * been modified (any property), and which unique indexes and
       * non-nullable properties remain to be validated.
       * 
       * @param   dmo
       *          DMO being tracked for validation.  If <code>null</code>
       *          the index and property variables will be set to empty sets.
       * @param   newlyCreated
       *          <code>true</code> if the DMO is newly created, otherwise
       *          <code>false</code>.
       */
      void reset(Persistable dmo, boolean newlyCreated)
      {
         touched = newlyCreated;
         shared = newlyCreated;
         
         if (dmo != null)
         {
            // Begin tracking unique indexes that require validation.
            try
            {
               String entity = getEntityName();
               List<String> unique = indexHelper.getIndexesForEntity(entity,
                                                                     true);
               uncheckedUniqueIndexes = new TreeSet<String>(unique);
            }
            catch (PersistenceException exc)
            {
               ErrorManager.recordOrThrowError(exc);
            }
            
            // Begin tracking non-nullable properties that require validation.
            uncheckedNonNullables = validator.getNonNullables();
            
            // Begin tracking cumulative set of properties made dirty.
            cumulativeDirtyProps = new HashSet<String>();
         }
         else
         {
            Set<String> empty = Collections.emptySet();
            uncheckedUniqueIndexes = empty;
            uncheckedNonNullables = empty;
            cumulativeDirtyProps = empty;
         }
      }
      
      /**
       * Change the state of the currently tracked DMO to dirty.
       */
      void setTouched()
      {
         touched = true;
      }
      
      /**
       * Indicate whether the current record has had its state changes shared
       * with other sessions via the dirty share manager, since it was first
       * created.
       * 
       * @return  <code>true</code> if the record's changes have been shared,
       *          else <code>false</code>.
       */
      boolean wasShared()
      {
         return shared;
      }
      
      /**
       * Perform a full validation (all unique constraints and non-nullable
       * properties) of the given record.
       * 
       * @param   dmo
       *          DMO to be validated.
       * @param   share
       *          Whether or not to share information about dirty properties
       *          with other contexts via the dirty share manager.
       * 
       * @throws  ValidationException
       *          if the DMO state fails any unique or non-nullable constraint
       *          check.
       * @throws  PersistenceException
       *          if there is any error gathering index metadata or a database
       *          error performing validation.
       */
      void validateFull(Persistable dmo, boolean share)
      throws ValidationException,
             PersistenceException
      {
         if (dmo == null)
         {
            return;
         }
         
         Serializable id = dmo.getId();
         String entity = getEntityName();
         boolean isTransient = isTransient();
         boolean isCurrentRecord = areDMOsSame(dmo, currentRecord);
         boolean insert = (isCurrentRecord &&
                           isTransient &&
                           !dirtyContext.isLocalInsert(entity, id));
         List<String> nonUnique = Collections.emptyList();
         ValidationData vData =
            new ValidationData(uniqueProperties,
                               uniqueIndexMap.keySet(),
                               nonUnique,
                               insert,
                               isCurrentRecord || (share && !isTemporary()),
                               isTransient);
         
         if (isCurrentRecord)
         {
            validate(vData, true);
         }
         else
         {
            validator.validate(RecordBuffer.this, dmo, vData);
         }
      }
      
      /**
       * Perform validation of any unique constraints and non-nullable
       * properties of the buffer's current record, which have not already
       * been validated since this record was loaded or created.
       * 
       * @throws  ValidationException
       *          if the DMO state fails any unique or non-nullable constraint
       *          check.
       * @throws  PersistenceException
       *          if there is a database error performing validation.
       */
      void validateRemaining()
      throws ValidationException,
             PersistenceException
      {
         if (currentRecord == null || !touched)
         {
            return;
         }
         
         Serializable id = currentRecord.getId();
         boolean isTransient = isTransient();
         boolean insert = (isTransient &&
                           !dirtyContext.isLocalInsert(getEntityName(), id));
         Set<String> nonUnique = nonUniqueIndexMap.keySet();
         ValidationData vData =
            new ValidationData(uncheckedNonNullables,
                               uncheckedUniqueIndexes,
                               nonUnique,
                               insert,
                               !isTemporary(),
                               isTransient);
         
         validate(vData, true);
      }
      
      /**
       * Perform validation of those unique constraints and non-nullable
       * properties of the buffer's current record, which are dirty due to
       * updates to the DMO.
       * <p>
       * If validation fails, dirty properties are reverted to their previous
       * state.
       * 
       * @throws  ValidationException
       *          if the DMO state fails any unique or non-nullable constraint
       *          check.
       * @throws  PersistenceException
       *          if there is a database error performing validation.
       */
      void validateDirty()
      throws ValidationException,
             PersistenceException
      {
         if (currentRecord == null || !touched)
         {
            return;
         }
         
         boolean share = !isTemporary();
         ValidationData vData = createValidationData(share);
         
         try
         {
            validate(vData, false);
            
            // Flush a transient record to the database if no more validation
            // is necessary.
            if (isTransient() &&
                uncheckedUniqueIndexes.isEmpty() &&
                uncheckedNonNullables.isEmpty())
            {
               touched = false;
               flush();
            }
         }
         catch (ValidationException exc)
         {
            undoDirty();
            
            throw exc;
         }
      }
      
      /**
       * Prevalidate a property which is about to be stored into a DMO using
       * the given setter method.  The existing properties in the DMO are not
       * validated.  The given property is validated iff it is also marked as
       * dirty within the <code>dirtyProps</code> map.
       * <p>
       * If this method returns normally, validation was successful.
       *
       * @param   property
       *          DMO property which requires validation.  May not be
       *          <code>null</code>.
       * @param   value
       *          Value about to be applied to the given property.
       * 
       * @return  <code>true</code> if all unique indexes and non-nullable
       *          properties have been validated once the given property has
       *          been validated, else <code>false</code>.
       *
       * @throws  ValidationException
       *          if validation on the buffer fails.
       * @throws  PersistenceException
       *          if any error occurs querying the database during validation.
       */
      boolean prevalidate(String property, BaseDataType value)
      throws ValidationException,
             PersistenceException
      {
         if (property == null)
         {
            throw new NullPointerException("property may not be null");
         }
         
         assert (currentRecord != null);
         assert (dirtyProps.containsKey(property));
         
         boolean share = !isTemporary();
         ValidationData vData = createValidationData(share);
         validator.validate(RecordBuffer.this, vData, property, value);
         
         // Update housekeeping collections if validation was a success.
         uncheckedUniqueIndexes.removeAll(vData.getDirtyUnique());
         uncheckedNonNullables.remove(property);
         updateSharedFlag(vData);
         
         boolean complete = (uncheckedUniqueIndexes.isEmpty() &&
                             uncheckedNonNullables.isEmpty());
         
         if (complete)
         {
            cumulativeDirtyProps.addAll(dirtyProps.keySet());
            dirtyProps.clear();
            touched = false;
         }
         else
         {
            dirtyProps.remove(property);
            cumulativeDirtyProps.add(property);
         }
         
         return complete;
      }
      
      /**
       * Validate the record currently stored in this buffer.  The parameters
       * of validation are defined by the <code>vData</code> object passed to
       * this method.  Updates which affect any indexes will be shared via the
       * dirty share manager.
       * 
       * @param   vData
       *          Validation data which determines which properties and
       *          indexes are validated and shared.
       * @param   clear
       *          Flag indicating whether the {@link
       *          RecordBuffer#unreportedChanges}
       *          list should be cleared on a validation error.
       *
       * @throws  ValidationException
       *          if validation on the buffer fails.
       * @throws  PersistenceException
       *          if any error occurs querying the database during validation.
       */
      private void validate(ValidationData vData, boolean clear)
      throws ValidationException,
             PersistenceException
      {
         try
         {
            validator.validate(RecordBuffer.this, currentRecord, vData);
            
            // Update housekeeping collections if validation was a success.
            uncheckedUniqueIndexes.removeAll(vData.getDirtyUnique());
            uncheckedNonNullables.removeAll(vData.getDirtyProperties());
            cumulativeDirtyProps.addAll(dirtyProps.keySet());
            updateSharedFlag(vData);
            
            dirtyProps.clear();
         }
         catch (ValidationException exc)
         {
            if (clear)
            {
               unreportedChanges.clear();
            }
            
            // Transaction manager will trigger rollback of DMO changes.
            throw exc;
         }
      }
      
      /**
       * Gather validation parameters, including dirty unique indexes, dirty
       * non-unique indexes, dirty property names, whether to share updates
       * via the dirty share manager, and whether a dirty insert is required.
       * 
       * @param   share
       *          Whether to share updates via the dirty share manager.
       *
       * @return  Validation parameters.
       */
      private ValidationData createValidationData(boolean share)
      {
         String entity = getEntityName();
         Serializable id = currentRecord.getId();
         boolean fullMatch = isTransient();
         Set<String> dirtyKeys = dirtyProps.keySet();
         List<String> uniq = listDirtyIndexes(entity, true, fullMatch);
         List<String> nonUniq = listDirtyIndexes(entity, false, fullMatch);
         boolean insert = (share &&
                           fullMatch &&
                           !dirtyContext.isLocalInsert(entity, id));
         ValidationData vData = new ValidationData(dirtyKeys,
                                                   uniq,
                                                   nonUniq,
                                                   insert,
                                                   share,
                                                   fullMatch);
         
         return vData;
      }
      
      /**
       * Update the shared flag based on the validation data just used to
       * validate the current record.
       * 
       * @param   vData
       *          Validation data which determines which properties and
       *          indexes are validated and shared.
       */
      private void updateSharedFlag(ValidationData vData)
      {
         if (shared)
         {
            return;
         }
         
         shared = (vData.isShareDirty() &&
                  (!vData.getDirtyUnique().isEmpty() ||
                   !vData.getDirtyNonUnique().isEmpty()));
      }
   }
   
   /**
    * Handler for methods invoked on the DMO proxy provided to client code.
    * Methods handled here include all methods of the DMO interface used to
    * create the proxy, as well as the <code>buffer()</code> method which
    * permits classes in this package to retrieve the record buffer instance
    * itself from the proxy. The former methods are meant for use by client
    * code, while the latter allows persistence support classes, such as
    * query objects, to access the buffer for implementation and housekeeping
    * purposes.
    */
   private final class Handler
   implements InvocationHandler,
              SessionListener
   {
      /** Hibernate session */
      private Session session = null;
      
      /**
       * Default constructor.
       */
      Handler()
      {
      }
      
      /**
       * Respond to a session closing event by remembering that a session
       * bind may be necessary before the next invocation of a proxied method
       * on a DMO.  This prevents <code>LazyInitializationException</code>s
       * from being thrown if Hibernate itself has proxied the backing DMO
       * and needs a session to retrieve the backing data.
       *
       * @param   event
       *          Session event type.
       *
       * @return  Always <code>false</code>, indicating this object should
       *          not be deregistered as a session listener after this method
       *          returns.
       * 
       * @throws  PersistenceException
       *          not thrown.
       */
      public boolean sessionEvent(SessionListener.Event event)
      throws PersistenceException
      {
         switch (event)
         {
            case SESSION_CLOSING_NORMALLY:
            case SESSION_CLOSING_WITH_ERROR:
               bindNeeded = true;
               session = null;
               break;
               
            default:
               break;
         }
         
         return false;
      }
      
      /**
       * Invoked when this session listener is removed from the list of
       * registered session listeners.
       * <p>
       * This implementation does nothing.
       */
      public void deregisteredSessionListener()
      {
      }
      
      /**
       * This method acts as an interceptor which is called every time a
       * method of the proxy object is invoked. It allows us to perform pre-
       * and post-processing for these methods, and to delegate the regular
       * work of the method to the backing implementation DMO.
       * <p>
       * Certain methods are <i>reserved</i>, and are not forwarded to the
       * backing DMO;  instead, they are invoked on the enclosing instance of
       * the <code>RecordBuffer</code> class.  These are any methods declared
       * by the <code>BufferReference</code> interface or by the
       * <code>Object</code> class.
       *
       * @param   proxy
       *          Proxy object upon which the intercepted method is invoked.
       * @param   method
       *          Method to be invoked through reflection.
       * @param   args
       *          Arguments, if any, to <code>method</code>.
       *
       * @throws  Throwable
       *          if an exception/error occurs upon the invocation target,
       *          or if any problem occurs during pre- or post-processing.
       */
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         Object result = null;
         
         try
         {
            // Handle special (reserved) methods.
            Class<?> declClass = method.getDeclaringClass();
            if (declClass.equals(BufferReference.class) ||
                declClass.equals(Object.class))
            {
               return method.invoke(RecordBuffer.this, args);
            }
            
            // Target method is from a DMO interface.
            
            initialize();
            
            // Bind Hibernate session to Persistence object if necessary.
            if (session == null)
            {
               session = persistence.bind();
               bindNeeded = false;
            }
            
            // We may be in temporary record mode to allow client-side where
            // clause processing.
            Persistable tempRecord = getTempRecord();
            
            // If a temporary record is set, use it to back the invocation.
            if (tempRecord != null)
            {
               return method.invoke(tempRecord, args);
            }
            
            // If method is a setter, property will be non-null.
            String property = propsBySetter.get(method);
            
            // Ensure backing DMO is available.
            if (currentRecord == null)
            {
               boolean validMethod = getters.contains(method);
               
               // Outer join processing will set unknown mode, because in an
               // outer join situation, it can be valid to have no record in
               // the buffer and still request a property value.  Return
               // unknown value in this case.
               if (unknownMode && validMethod)
               {
                  return FieldReference.unknownValue(method);
               }
               else
               {
                  // Throw an exception if not in silent mode.
                  if (property != null)
                  {
                     // Attempt to set field on empty buffer results in
                     // "unable to update" error.
                     ErrorManager.recordOrThrowError(
                        142,
                        "Unable to update " + getTable() + " Field");
                  }
                  else
                  {
                     // Attempt to get field from empty buffer results in
                     // "record not available" error.
                     errorNotAvailable();
                  }
                  
                  // Silent error mode;  return unknown value if possible.
                  return (validMethod
                          ? FieldReference.unknownValue(method)
                          : null);
               }
            }
            
            // Is property an extent field?
            boolean extent = false;
            if (args != null)
            {
               extent = (args.length == (property == null ? 1 : 2));
            }
            
            // Handle unknown value and out of bounds conditions for the
            // element index value.
            if (extent)
            {
               RangeCode rc = checkRange(args, method, property);
               switch (rc)
               {
                  case NORMAL:
                     break;
                     
                  case OUT_OF_BOUNDS:
                     int idx =
                        args[0] instanceof int64
                        ? ((int64) args[0]).intValue() // handles integer case, too
                        : ((Integer) args[0]).intValue();
                     String msg =
                        "Array subscript " + (++idx) + " is out of range";
                     ErrorManager.recordOrThrowError(26, msg);
                     return (property != null
                             ? null
                             : FieldReference.unknownValue(method));
                     
                  case UNKNOWN_VALUE:
                     return (property != null ? null : args[0]);
               }
            }
            
            // If the target property represents an extent field and the
            // current record is not transient, the associated collection may
            // have been detached from its session by our aggressive eviction
            // policy.  Ensure that it is properly associated with the current
            // Hibernate session, since it may need to be lazily initialized
            // and we don't want that to trigger an exception.
            if (extent && createScope < 0 && !session.contains(currentRecord))
            {
               try
               {
                  session.lock(currentRecord, LockMode.NONE);
               }
               catch (HibernateException exc)
               {
                  throw new PersistenceException(exc);
               }
            }
            
            // Only used for setter invocation (see below).
            boolean changed = false;
            boolean processChange = false;
            boolean commitNow = false;
            boolean needsFlush = false;
            
            // Special handling for setter methods.
            if (property != null)
            {
               // We don't want to upgrade the lock, validate, or assign if
               // an error is pending.
               if (ErrorManager.isPendingError())
               {
                  return null;
               }
               
               // changeCodes[0] = did setter change value?
               // changeCodes[1] = must change be processed/reported?
               // changeCodes[2] = is validation on this (transient) DMO done?
               boolean[] changeCodes = new boolean[] { false, false, false };
               
               if (extent)
               {
                  Integer idx = args[0] instanceof int64
                                ? ((int64) args[0]).intValue() // handles integer case, too
                                : (Integer) args[0];
                  detectChange(idx, 
                               property,
                               (BaseDataType) args[1],
                               changeCodes);
               }
               else if (getSizerMap().containsKey(property))
               {
                  // This is a bulk setter method which sets multiple values
                  // for an extent field.  It requires special handling.
                  Method sizer = sizerMap.get(property);
                  Object o = sizer.invoke(currentRecord, (Object[]) null);
                  int size = ((int64) o).intValue();
                  
                  // Bulk setters come in 2 flavors: array and scalar.
                  if (args[0].getClass().isArray())
                  {
                     // Array version.
                     size = Math.min(size, Array.getLength(args[0]));
                     Object[] array = (Object[]) args[0];
                     for (int i = 0; i < size; i++)
                     {
                        detectChange(new Integer(i),
                                     property,
                                     (BaseDataType) array[i],
                                     changeCodes);
                     }
                  }
                  else
                  {
                     // Scalar version.
                     for (int i = 0; i < size; i++)
                     {
                        detectChange(new Integer(i),
                                     property,
                                     (BaseDataType) args[0],
                                     changeCodes);
                     }
                  }
               }
               else
               {
                  // Simple property.
                  detectChange(null,
                               property,
                               (BaseDataType) args[0],
                               changeCodes);
               }
               
               changed = changeCodes[0];
               processChange = changeCodes[1];
               
               // Possibly upgrade our lock, but only if data has changed.
               if (changed)
               {
                  try
                  {
                     // Try to get an exclusive lock.
                     upgradeLock();
                  }
                  catch (PersistenceException exc)
                  {
                     ErrorManager.recordOrThrowError(
                        142,
                        "Unable to update " + getTable() + " Field",
                        exc);
                     
                     return null;
                  }
               }
               
               needsFlush = changeCodes[2];
               
               // Validate the impending change, unless in batch assign
               // mode.  If in batch mode, add the current buffer to the
               // set of those pending validation, deferring validation
               // until the assignment batch is closed.
               if (!bufferManager.isBatchAssignMode())
               {
                  // Not batch mode;  validate and update immediately, but
                  // only if we are not performing a bulk copy.
                  
                  // Do we need to commit immediately?
                  if (isAutoCommit() && (!isTransient() || needsFlush))
                  {
                     boolean commit = (!isCommitPending() &&
                                       persistence.beginTransaction());
                     if (bulkCopy)
                     {
                        if (commit)
                        {
                           setCommitPending(true);
                        }
                     }
                     else
                     {
                        commitNow = commit;
                     }
                  }
               }
               else
               {
                  // Batch mode;  defer validation.
                  bufferManager.addDirtyBatchBuffer(RecordBuffer.this);
                  
                  // Do we need to commit when batch mode ends?
                  if (isAutoCommit() &&
                      (!isTransient() || needsFlush) &&
                      !isCommitPending() &&
                      persistence.beginTransaction())
                  {
                     setCommitPending(true);
                  }
               }
            }

            boolean invokeTriggers = changed &&
                                     !extent &&
                                     !bufferManager.isBatchAssignMode() && 
                                     triggerDelegate != null;

            BaseDataType oldVal = null;
            if (invokeTriggers)
            {
               // Get the existing value.
               Method getter = getGetterMap().get(property);
               oldVal = (BaseDataType) getter.invoke(currentRecord);
            }

            // (Finally) invoke method on underlying DMO.
            result = method.invoke(currentRecord, args);

            // invoke database triggers
            if (invokeTriggers)
            {
               try
               {
                  triggerDelegate.executeAssignTriggers(getDMOProxy(),
                                                        property,
                                                        oldVal.duplicate(),
                                                        false);
               }
               catch (ConditionException exc)
               {
                  unreportedChanges.remove(property);
                  method.invoke(currentRecord, oldVal);
                  throw exc;
               }
            }
            
            // Update set of dirty legacy key properties for this buffer, if
            // necessary, and synchronize foreign keys (if not in batch
            // assign mode).
            boolean clearProps = false;
            if (changed &&
                legacyForeignKeys != null &&
                legacyForeignKeys.contains(property))
            {
               dirtyLegacyKeyProps.add(property);
               if (!bufferManager.isBatchAssignMode())
               {
                  // Not batch mode;  synchronize foreign keys immediately.
                  synchForeignRelations(false);
                  clearProps = true;
               }
            }
            
            // Flush to database if needed.  This will trigger a validation of
            // any remaining, unvalidated changes.
            if (needsFlush)
            {
               flush();
            }
            
            if (clearProps)
            {
               dirtyLegacyKeyProps.clear();
            }
            
            // Broadcast notification if state changed.
            if (processChange)
            {
               processChange();
            }
            
            // Commit now if necessary.
            if (commitNow)
            {
               persistence.commit();
            }
         }
         catch (ValidationException exc)
         {
            // TODO:  sometimes this message is not reported on validation
            // failure in Progress;  figure out when it should appear and when not.
            
            // When a field validation error occurs (Eg. 32 bit Integer overflow) the exception 
            // thrown is an ErrorConditionException which is the target for an 
            // InvocationTargetException so it is catched below. 
            ErrorManager.recordOrThrowError(
               142,
               "Unable to update " + getTable() + " Field",
               exc);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         catch (InvocationTargetException exc)
         {
            // look for ErrorConditionException
            Throwable ex2 = exc;
            while (ex2 != null && !(ex2 instanceof ConditionException))
            {
               ex2 = ex2.getCause();
            }
            
            if (ex2 != null)
            {
               ErrorManager.recordOrThrowError(142, "Unable to update " + getTable() + " Field");
            }
            else
            {
               DBUtils.handleException(getDatabase(), exc);
               
               Throwable cause = exc.getTargetException();
               
               if (LOG.isErrorEnabled())
               {
                  String errorText = describe() + ":  error invoking proxied DMO method";
                  LOG.error(errorText, exc);
               }
               
               // Must be the result of a programming error.  Don't wrap in a
               // legacy-style exception type.
               throw cause;
            }
         }
         
         return result;
      }
      
      /**
       * Register this object with the Persistence service object as a session
       * listener.  It is not deregistered explicitly, but rather when this
       * object is garbage collected.  This is necessary because we can't be
       * sure what scope we will be in when this method is invoked, since it
       * is only invoked when the buffer is first initialized.
       */
      void registerSessionListener()
      {
         persistence.registerSessionListener(this, SessionListener.Scope.NONE);
      }
      
      /**
       * Given a getter or setter method for a converted extent field and the
       * arguments for that method, check the extent index argument for
       * validity.  Test for unknown value, as well as out of bounds values.
       * <p>
       * The index argument (at position 0 in the <code>args</code> array) can
       * be either a <code>Number</code> or a <code>NumberType</code>.  If it
       * is not null/unknown, overwrite it in the array with its value as an
       * <code>Integer</code>.
       * 
       * @param   args
       *          Original arguments to getter/setter method.
       * @param   method
       *          Getter/setter method.
       * @param   property
       *          DMO property name associated with <code>method</code>.
       * 
       * @return  The appropriate range code to indicate whether the index
       *          value is unknown, out of bounds, or normal.
       * 
       * @throws  Exception
       *          if there is a reflection error invoking a sizer method when
       *          performing out of bounds check.
       */
      private RangeCode checkRange(Object[] args,
                                   Method method,
                                   String property)
      throws Exception
      {
         Object index = args[0];
         int i = -1;
         
         if (index == null)
         {
            return RangeCode.UNKNOWN_VALUE;
         }
         else if (index instanceof NumberType)
         {
            // Check for unknown value.
            NumberType num = (NumberType) index;
            if (num.isUnknown())
            {
               return RangeCode.UNKNOWN_VALUE;
            }
            
            i = num.intValue();
         }
         else if (index instanceof Number)
         {
            i = ((Number) index).intValue();
         }
         else
         {
            throw new IllegalArgumentException(
               "Unexpected index type for DMO property index");
         }
         
         // Check minimum end of range.
         if (i < 0)
         {
            return RangeCode.OUT_OF_BOUNDS;
         }
         
         if (property == null)
         {
            property = propsByGetter.get(method);
         }
         
         Integer max = extentsByProperty.get(property);
         if (max == null)
         {
            Method sizer = getSizerMap().get(property);
            int64 extent = (int64) sizer.invoke(currentRecord);
            max = new Integer(extent.intValue());
            extentsByProperty.put(property, max);
         }
         
         // Check maximum end of range.
         if (i >= max.intValue())
         {
            return RangeCode.OUT_OF_BOUNDS;
         }
         
         return RangeCode.NORMAL;
      }
      
      /**
       * Process a setter method about to be invoked by the invocation
       * handler.  Processing in this sense entails marking properties as
       * dirty and detecting which values have changed for purposes of
       * reporting change via the {@link ChangeBroker}.
       * <p>
       * This method handles both indexed and non-indexed properties.  The
       * <code>index</code> parameter must be non-<code>null</code> and
       * non-negative for indexed properties, and should be <code>null</code>
       * for non-indexed properties.
       *
       * @param   index
       *          Non-negative, zero-based index value to be used for an
       *          indexed property.  Should be <code>null</code> for a
       *          non-indexed property.
       * @param   property
       *          Name of property affected by the setter invocation.
       * @param   newVal
       *          New value to be set into the property.
       * @param   changeCodes
       *          Array of boolean values which are updated by this method
       *          to provide feedback to the caller as to the change which
       *          will occur as a result of the setter method invocation:
       *          <ul>
       *             <li><code>changeCodes[0]</code>:  indicates that either
       *                 the new value being set is different than the
       *                 existing property value, or that the boolean at this
       *                 position in the array was already <code>true</code>
       *                 at the invocation of this method.
       *             <li><code>changeCodes[1]</code>:  indicates that either
       *                 this method determined that a change occurred which
       *                 must be processed immediately by the record buffer
       *                 and reported to the change broker, or that the
       *                 boolean at this position in the array was already
       *                 <code>true</code> at the invocation of this method.
       *             <li><code>changeCodes[2]</code>:  indicates that change
       *                 was prevalidated and no further validation is
       *                 necessary for the current (transient) record before
       *                 it can be flushed to the database.
       *          </ul>
       *          Since this method may be called in a loop for a single,
       *          bulk setter method, these values will not be changed once
       *          they are set to <code>true</code>, such that any change
       *          detected within the loop is properly detected by the caller
       *          at the loop's end.
       */
      private void detectChange(Integer index,
                                String property,
                                BaseDataType newVal,
                                boolean[] changeCodes)
      throws IllegalAccessException,
             InvocationTargetException,
             PersistenceException,
             ValidationException
      {
         boolean changed = false;
         boolean processChange = false;
         boolean validationDone = false;
         
         boolean extent = (index != null);

         newVal = (BaseDataType) newVal.deepCopy();

         // Get the old value and compare the new value against it to
         // determine whether the value has changed as a result of this
         // setter invocation.
         
         // Get the existing value.
         Method getter = getGetterMap().get(property);
         Object[] tmpArgs = (extent ? new Object[] { index } : null);
         Object oldVal = getter.invoke(currentRecord, tmpArgs);
         
         // Force case-sensitive comparison if values are character
         // instances.  This will ensure we notice a change if only the
         // the case changed.
         if (oldVal instanceof character)
         {
            character c1 = (character) oldVal;
            character c2 = (character) newVal;
            changed = !c1.equalsCaseSensitively(c2);
         }
         
         // Determine whether the value changed.
         if (changed ||
             (newVal == null && oldVal != newVal) ||
             (newVal != null && !newVal.equals(oldVal)))
         {
            changed = true;
            
            // Record information to be submitted to the ChangeBroker.
            // We record the old value and the new value.
            Object[] diffs = new Object[] { oldVal, newVal };
            if (extent)
            {
               Map<Integer, Object[]> map =
                  (Map<Integer, Object[]>) unreportedChanges.get(property);
               if (map == null)
               {
                  map = new HashMap<Integer, Object[]>();
                  unreportedChanges.put(property, map);
               }
               map.put(index, diffs);
            }
            else
            {
               unreportedChanges.put(property, diffs);
            }
         }
         
         if (changed || isTransient())
         {
            validationHelper.setTouched();
            
            // Track the fact that this property is being updated and remember
            // the old value in case validation fails and we need to roll
            // back.  Note that for purposes of tracking dirty fields for
            // transient records, we don't care if the value actually changed,
            // only that it was set.
            if (extent)
            {
               Map<Integer, Object[]> map =
                  (Map<Integer, Object[]>) dirtyProps.get(property);
               if (map == null)
               {
                  map = new HashMap<Integer, Object[]>();
                  dirtyProps.put(property, map);
               }
               map.put(index, new Object[] { oldVal, newVal });
            }
            else
            {
               dirtyProps.put(property, new Object[] { oldVal, newVal });
            }
            
            // Prevalidate if necessary.
            if (!bufferManager.isBatchAssignMode())
            {
               if (!bulkCopy)
               {
                  validationDone =
                     validationHelper.prevalidate(property, newVal);
               }
               
               processChange = true;
            }
         }
         
         // Once a change code is true, do not reset it to false, as this
         // method is sometimes called in a loop, and we want to know if any
         // invocation in the loop has triggered a change.
         changeCodes[0] = (changeCodes[0] || changed);
         changeCodes[1] = (changeCodes[1] || processChange);
         changeCodes[2] = (changeCodes[2] || validationDone);
      }
   }

   /**
    * A <code>Reversible</code> version which affects more than one record.
    */
   protected abstract class AbstractBulkReversible
   extends Reversible
   {
      /**
       * This method checks if this bulk reversible touched the record with
       * the given key.
       * 
       * @param   key
       *          The record to be checked.
       *          
       * @return  <code>true</code> if the record was touched by this bulk
       *          reversible.
       */
      protected abstract boolean containsRecord(Serializable key);
   }
   
   /**
    * Concrete implementation of <code>Reversible</code> which on rollback
    * reverses a record create action by deleting its DMO in the underlying
    * persistence session.
    */
   private final class ReversibleCreate
   extends AbstractReversible
   {
      /** Flag indicating associated record was persisted to the database */
      private boolean persisted = false;
      
      /**
       * Constructor.
       */
      ReversibleCreate()
      {
         super(RecordBuffer.this, currentRecord);
      }
      
      /**
       * Get a string which briefly describes the action which this object
       * will reverse upon rollback.
       * 
       * @return  Action string.
       */
      protected String getAction()
      {
         return "Creation";
      }
      
      /**
       * Reverse the creation of the underlying database record by deleting
       * it from the underlying persistence session, which deletes it from
       * the database.  Ensure the buffer has no current record set.
       *
       * @param   rollForward
       *          If <code>true</code>, this method is being invoked in order
       *          to re-roll forward a non-undoable change, after a
       *          database-level rollback.  This indicates the modification is
       *          made within a new transaction invoked after the
       *          database-level rollback.
       *          If <code>false</code>, this method is being invoked in order
       *          to roll back an undoable change within the context of an
       *          application-level, full transaction or sub-transaction
       *          rollback.
       * @param   transaction
       *          <code>true</code> if this rollback occurs at a full
       *          transaction boundary, else <code>false</code>.
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when deleting the record.
       */
      protected void rollbackWorker(boolean rollForward, boolean transaction)
      throws PersistenceException
      {
         try
         {
            if (persisted)
            {
               Persistable dmo = getDMO();
               
               // execute create triggers, as this is a rollback
               if (triggerDelegate != null)
               {
                  triggerDelegate.executeCreateTriggers((DataModelObject) dmo, 
                                                        true);
               }

               persistence.delete(dmo);
               
               RecordIdentifier ident = new RecordIdentifier(getEntityName(),
                                                             dmo.getId());
               dirtyContext.rollbackInsert(ident, false);
               
               reportChange(dmo, false, true);
               
               if (LOG.isDebugEnabled())
               {
                  StringBuilder buf = new StringBuilder(" CREATE ROLLBACK: ");
                  buf.append(dmo);
                  if (LOG.isTraceEnabled())
                  {
                     buf.append(";  details:  ");
                     buf.append(RecordBuffer.this.toString(dmo));
                  }
                  LOG.debug(describe() + buf.toString());
               }
            }
         }
         finally
         {
            // Null out current record in any case.
            setCurrentRecord(null, true, false, false);
         }
      }
      
      /**
       * Invoked to mark the action as truly reversible.  This is necessary
       * because the object is created when a new record is first created,
       * but that record may never actually be persisted to the database.
       * Once it is, this method is invoked, such that a rollback will delete
       * the DMO from the database.
       */
      void markPersisted()
      {
         persisted = true;
      }
   }
   
   /**
    * Concrete implementation of <code>Reversible</code> which on rollback
    * reverses a record delete action by reattaching its DMO to the underlying
    * persistence session.
    */
   private final class ReversibleDelete
   extends Reversible
   {
      /** Deleted DMO */
      private Persistable dmo = null;
      
      /**
       * Constructor.
       */
      ReversibleDelete()
      {
         setDMO(currentRecord);
      }
      
      /**
       * Get a string which briefly describes the action which this object
       * will reverse upon rollback.
       * 
       * @return  Action string.
       */
      protected String getAction()
      {
         return "Deletion";
      }
      
      /**
       * Indicate whether this <code>Reversible</code> implementation stores its
       * DMO locally, as opposed to storing an identifier and having to retrieve
       * the DMO from the backing database, or not making a DMO available at all.
       * 
       * @return  <code>true</code> if the DMO is stored locally, else
       *          <code>false</code>.
       */
      protected boolean storesDMO()
      {
         return true;
      }
      
      /**
       * Get the DMO which was deleted.
       * 
       * @return  DMO in the state it was in at the time of deletion.
       */
      protected Persistable getDMO()
      {
         // the current record in the buffer must be returned only if this
         // is an undoable table; else, use the one in the "dmo" field

         if (isUndoable() &&
             dmo != null && currentRecord != null &&
             dmo.getId().equals(currentRecord.getId()))
         {
            // If the current record has extent fields, make sure they are
            // initialized.  This should only happen in the reversible
            // rollback case (otherwise, the record to be restored would not
            // be the current record).
            initializeLazyExtents();
            
            return currentRecord;
         }
         
         return dmo;
      }
      
      /**
       * Store the deleted DMO.
       * 
       * @param   dmo
       *          Deleted DMO.
       */
      protected void setDMO(Persistable dmo)
      {
         this.dmo = dmo;
      }
      
      /**
       * Get the unqualified interface name of the DMO being managed by this
       * reversible.
       * 
       * @return  DMO interface name.
       */
      protected String getDMOName()
      {
         return RecordBuffer.this.getDMOName();
      }
      
      /**
       * Get the ID of the underlying DMO.
       *
       * @return  DMO ID.
       */
      protected Serializable getId()
      {
         return (dmo == null ? null : dmo.getId());
      }
      
      /**
       * Reverse the deletion of the underlying database record by reattaching
       * it to the underlying persistence session.
       *
       * @param   rollForward
       *          If <code>true</code>, this method is being invoked in order
       *          to re-roll forward a non-undoable change, after a
       *          database-level rollback.  This indicates the modification is
       *          made within a new transaction invoked after the
       *          database-level rollback.
       *          If <code>false</code>, this method is being invoked in order
       *          to roll back an undoable change within the context of an
       *          application-level, full transaction or sub-transaction
       *          rollback.
       * @param   transaction
       *          <code>true</code> if this rollback occurs at a full
       *          transaction boundary, else <code>false</code>.
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when detaching the record.
       */
      protected void rollback(boolean rollForward, boolean transaction)
      throws PersistenceException
      {
         Persistable dmo = getDMO();
       
         // at this point, before validation is performed, there might be
         // Update operations which weren't flushed (as there are other 
         // pending operations). flush is needed so validation is done using
         // the latest versions
         persistence.maybeFlush();
         
         try
         {
            validationHelper.validateFull(dmo, false);
         }
         catch (ValidationException exc)
         {
            throw new PersistenceException(exc);
         }
         
         Serializable id = dmo.getId();
         Persistable merged = persistence.merge(dmo);
         setCurrentRecord(merged, true, false, false);
         
         // execute delete triggers, as this is a rollback
         if (triggerDelegate != null)
         {
            triggerDelegate.executeDeleteTriggers((DataModelObject) dmo, true);
         }

         // Rollback the delete with the dirty share context.
         dirtyContext.delete(getEntityName(), id, true, false);
         
         // Re-insert the record in the dirty database if necessary.
         if (isNewlyCreated())
         {
            // TODO:  this will re-insert the DMO in its current state, which
            // may differ from the state it was in when originally inserted.
            dirtyContext.insert(RecordBuffer.this, dmo, null, null, false);
         }
         
         if (LOG.isDebugEnabled())
         {
            StringBuilder buf = new StringBuilder(" DELETE ROLLBACK: ");
            buf.append(dmo);
            if (LOG.isTraceEnabled())
            {
               buf.append(";  details:  ");
               buf.append(RecordBuffer.this.toString(dmo));
            }
            LOG.debug(describe() + buf.toString());
         }
         
         reportChange(dmo, true, false);
      }
   }
   
   /**
    * Concrete implementation of <code>Reversible</code> which on rollback
    * reverses one or more updates made to a record's properties.
    */
   private final class ReversibleUpdate
   extends AbstractReversible
   {
      /** Map of property key to original value */
      private final Map<Key, Object> rollbackMap =
         new HashMap<Key, Object>(2);
      
      /** 
        * Flag indicating if this update touches any property which is part
        * of an unique constraint.
        */
      private boolean uniqueConstraints = false;
      
      /**
       * Constructor.
       */
      ReversibleUpdate()
      {
         super(RecordBuffer.this, currentRecord);
      }
      
      /**
       * Overridden implementation which adds a dump of the rollback map of
       * undo entries to the parent's display.
       *
       * @return  String which describes internal state of this object.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder(super.toString());
         buf.append(" ");
         buf.append(rollbackMap);
         
         return buf.toString();
      }
      
      /**
       * Get a string which briefly describes the action which this object
       * will reverse upon rollback.
       * 
       * @return  Action string.
       */
      protected String getAction()
      {
         return "Update";
      }
      
      /**
       * Add an undo entry to the rollback map.  The key to the map entry is
       * an object composed of the property name and a zero-based index of an
       * array element (<code>null</code> for simple properties).  The value
       * of the map entry is the original value to which the property will be
       * rolled back if necessary.
       * <p>
       * If an entry already exists for the given key, it is <b>not</b>
       * replaced.  Thus, multiple changes to the same property are collapsed
       * such that a rollback actually rolls back to the original valud of
       * the property, not to an intermediate state.
       *
       * @param   property
       *          Name of updated property.
       * @param   index
       *          Zero-based index for an indexed property;  should be
       *          <code>null</code> for a simple property.
       * @param   oldValue
       *          Value to which the property should be rolled back.
       */
      void addUndoEntry(String property, Integer index, Object oldValue)
      {
         Key key = new Key(property, index);
         if (!rollbackMap.containsKey(key))
         {
            rollbackMap.put(key, oldValue);
            if (LOG.isTraceEnabled())
            {
               LOG.trace(describe() +
                         ":  added undo entry {" +
                         getId() +
                         ":" +
                         property +
                         (index != null ? ("[" + index + "]") : "") +
                         ":  " +
                         oldValue +
                         "}");
            }
         }
      }
      
      /**
       * Merge all undo entries from a source instance of this class to this
       * instance.  If this instance already contains an entry for a key
       * found in <code>source</code>'s rollback map, the entry within
       * <code>source</code> is ignored.
       *
       * @param   source
       *          Source object from which to merge undo entries.
       */
      void nonDestructiveMerge(ReversibleUpdate source)
      {
         boolean trace = LOG.isTraceEnabled();
         Iterator<Map.Entry<Key, Object>> iter =
            source.rollbackMap.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<Key, Object> next = iter.next();
            Key key = next.getKey();
            if (!rollbackMap.containsKey(key))
            {
               Object oldValue = next.getValue();
               rollbackMap.put(key, oldValue);
               if (trace)
               {
                  LOG.trace(describe() +
                            ":  merged undo entry {" +
                            getId() +
                            ":" +
                            key.property +
                            (key.index != null
                             ? ("[" + key.index + "]")
                             : "") +
                            ":  " +
                            oldValue +
                            "}");
               }
            }
         }
      }
      
      /**
       * Reverse all updates made to the target DMO by invoking setter
       * methods to set back all original property values known to this
       * object.
       *
       * @param   rollForward
       *          If <code>true</code>, this method is being invoked in order
       *          to re-roll forward a non-undoable change, after a
       *          database-level rollback.  This indicates the modification is
       *          made within a new transaction invoked after the
       *          database-level rollback.
       *          If <code>false</code>, this method is being invoked in order
       *          to roll back an undoable change within the context of an
       *          application-level, full transaction or sub-transaction
       *          rollback.
       * @param   transaction
       *          <code>true</code> if this rollback occurs at a full
       *          transaction boundary, else <code>false</code>.
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when rolling back record updates.
       */
      protected void rollbackWorker(boolean rollForward, boolean transaction)
      throws PersistenceException
      {
         Map<String, Method> setterMap = getSetterMap();
         Map<String, Method> getterMap = getGetterMap();
         String property = null;
         boolean debug = LOG.isDebugEnabled();
         boolean trace = LOG.isTraceEnabled();
         String desc = null;
         
         try
         {
            Persistable dmo = getDMO();
            
            if (debug)
            {
               desc = describe();
               StringBuilder buf = new StringBuilder(desc);
               buf.append(" UPDATE ROLLBACK: ");
               buf.append(dmo);
               buf.append(";  affected properties:  ");
               buf.append(rollbackMap.keySet());
               
               if (trace)
               {
                  buf.append(";  pre-rollback:  ");
                  buf.append(RecordBuffer.this.toString(dmo));
                  LOG.trace(buf.toString());
               }
               else
               {
                  LOG.debug(buf.toString());
               }
            }
            
            List<String> properties = (isNewlyCreated()
                                       ? null
                                       : new ArrayList<String>());
            List<BaseDataType> oldValues = new ArrayList<BaseDataType>();
            
            Iterator<Map.Entry<Key, Object>> iter =
               rollbackMap.entrySet().iterator();
            while (iter.hasNext())
            {
               Map.Entry<Key, Object> next = iter.next();
               Key key = next.getKey();
               Object oldValue = next.getValue();
               property = key.property;
               
               // Collect property names and their associated rollback values
               // for indexed properties.  Extent fields are not eligible.
               if (properties != null &&
                   key.index == null &&
                   indexedProperties.contains(key.property))
               {
                  properties.add(key.property);
                  oldValues.add((BaseDataType) oldValue);
               }
               
               Method getter = getterMap.get(property);
               Object[] getArgs = (key.index == null 
                                   ? null
                                   : new Object[] { key.index });
               Object currentValue = getter.invoke(dmo, getArgs);
               Object[] diffs = new Object[] { currentValue, oldValue };
               
               Method setter = setterMap.get(property);
               Object[] args = (key.index == null
                                ? new Object[] { oldValue }
                                : new Object[] { key.index, oldValue });
               setter.invoke(dmo, args);
               
               // because when rolling back property changes we bypass the
               // proxy handler, we need to log these changes and after that
               // report the changes to the change broker
               if (key.index == null)
               {
                  unreportedChanges.put(property, diffs);
               }
               else
               {
                  Map<Integer, Object[]> map =
                     (Map<Integer, Object[]>) unreportedChanges.get(property);
                  if (map == null)
                  {
                     map = new HashMap<Integer, Object[]>();
                     unreportedChanges.put(property, map);
                  }
                  map.put(key.index, diffs);
               }
            }
            
            if (triggerDelegate != null)
            {
               // execute the ASSIGN triggers, as this is a rollback
               iter = rollbackMap.entrySet().iterator();
               while (iter.hasNext())
               {
                  Map.Entry<Key, Object> next = iter.next();
                  Key key = next.getKey();
                  property = key.property;
                  
                  Method getter = getterMap.get(property);
                  Object[] getArgs = (key.index == null 
                                      ? null
                                      : new Object[] { key.index });
                  BaseDataType currentValue = 
                     (BaseDataType) getter.invoke(dmo, getArgs);
                  currentValue = currentValue.duplicate();
                  
                  triggerDelegate.executeAssignTriggers((DataModelObject) dmo,
                                                        property,
                                                        currentValue,
                                                        true);
               }
            }

            // Update dirty share manager, but only if it was updated with the
            // original changes.
            if (properties != null && !properties.isEmpty())
            {
               int size = properties.size();
               String[] props = new String[size];
               BaseDataType[] bdts = new BaseDataType[size];
               dirtyContext.update(getEntityName(),
                                   dmo.getId(),
                                   properties.toArray(props),
                                   oldValues.toArray(bdts),
                                   true);
            }
            
            reportChange(dmo, false, false);
            
            if (trace)
            {
               StringBuilder buf = new StringBuilder(desc);
               buf.append(" UPDATE ROLLBACK: ");
               buf.append(dmo);
               buf.append(";  post-rollback:  ");
               buf.append(RecordBuffer.this.toString(dmo));
               LOG.trace(buf.toString());
            }
         }
         catch (Exception exc)
         {
            DBUtils.handleException(getDatabase(), exc);
            
            throw new PersistenceException(
               "Error rolling back DMO update to '" + property + "'",
               exc);
         }
      }
      
      /**
       * Custom map key for the enclosing class' rollback map.  They key
       * consists of a unique combination of property name and index.
       */
      private class Key
      {
         /** Property name */
         private final String property;
         
         /** Extent field index */
         private final Integer index;
         
         /**
          * Constructor.
          *
          * @param   property
          *          Property name.
          * @param   index
          *          Extent field index.
          */
         Key(String property, Integer index)
         {
            this.property = property;
            this.index = index;
         }
         
         /**
          * Equality test.  Two keys are considered equal if their property
          * names are identical and their index values are identical (null
          * index values are considered equal).
          *
          * @param   o
          *          Another key to compare with this instance.
          *
          * @return  <code>true</code> if this key equals the given key, else
          *          <code>false</code>.
          */
         public boolean equals(Object o)
         {
            Key that = (Key) o;
            
            if ((index == null || that.index == null) && index != that.index)
            {
               return false;
            }
            
            return (property.equals(that.property) &&
                    (index == null || index.equals(that.index)));
         }
         
         /**
          * Generate a hash code for a given key, such that all hash codes
          * created are consistent with this object's implementation of the
          * <code>equals</code> method.
          *
          * @return  Hash code for this key.
          */
         public int hashCode()
         {
            int result = 17;
            result = 37 * result + property.hashCode();
            if (index != null)
            {
               result = 37 * result + index.hashCode();
            }
            
            return result;
         }
         
         /**
          * Provide a string representation of this key, suitable for
          * debugging.
          *
          * @return  String representation of this key.
          */
         public String toString()
         {
            StringBuilder buf = new StringBuilder(property);
            if (index != null)
            {
               buf.append("[").append(index).append("]");
            }
            
            return buf.toString();
         }
      }
   }
   
   /**
    * A lightweight implementation of the <code>Undoable</code> interface
    * which restores to the record buffer the current record and snapshot
    * which were active at the time of backup.
    * <p>
    * An instance of this class is registered for undo support with the
    * <code>TransactionManager</code> during record buffer construction.
    */
   private final class LightweightUndoable
   implements Undoable
   {
      /** DMO which is active in record buffer at the time of backup */
      private Persistable activeRecord = null;
      
      /** Snapshot which is active in record buffer at the time of backup */
      private Persistable activeSnapshot = null;
      
      /** Lock type pinned by the active record at the time of backup */
      private LockType pinnedLockType = null;
      
      /**
       * Store a reference to the record currently active in the record
       * buffer in a new instance of this class and return it.
       *
       * @return  An instance of this class which contains the undoable state
       *          of the record buffer at the moment it is called.
       */
      public Undoable deepCopy()
      {
         LightweightUndoable copy = new LightweightUndoable();
         
         copy.activeRecord = currentRecord;
         copy.activeSnapshot = snapshot;
         
         copy.pinnedLockType = (currentRecord != null
                                ? getPinnedLockType(currentRecord.getId())
                                : LockType.NONE);
         
         return copy;
      }
      
      /**
       * Restore to the record buffer the current record and current
       * reversible which were set at the time the record buffer was backed
       * up for undo support.
       *
       * @param   from
       *          An instance of this class containing the backup state to be
       *          restored.
       */
      public void assign(Undoable from)
      {
         if (offEndRollback)
         {
            return;
         }
         
         if (!isActive())
         {
            // If the buffer is not yet active, then the state will not have
            // changed, and there is nothing to do.
            return;
         }
         
         LightweightUndoable copy = (LightweightUndoable) from;
         
         // Reset current record and snapshot to what they were before backup.
         Persistable active = copy.activeRecord;
         if (active != null)
         {
            if (isUndoable())
            {
               if (active != currentRecord ||
                   LockType.NONE.equals(copy.pinnedLockType))
               {
                  load(active, copy.pinnedLockType, true);
               }
            }
            else
            {
               // for no-undo tables, the record may not exist in the table,
               // in case validation failed.
               setCurrentRecord(active, true, false, false);
            }
         }
         else
         {
            setCurrentRecord(null, true, false, false);
         }
         snapshot = copy.activeSnapshot;
      }
   }
}
